#!/usr/bin/python2
# -*- coding: utf-8 -*-
'''
Convert different file formats to graphml.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import sys        
import argparse
import networkx as nx
from itertools import islice

def splitLine(s, nr=1):
    '''Strip whitespace characters from s, raises ValueError if s does not
    contain nr elements. Return list of elements from s.''' 
    s = s.split()
    if len(s) != nr:
        raise ValueError("Line {} does not contain {} element(s) of items."
                         .format(s,nr))
    return s
    

def convertMaxToGraph(netFile):
    '''Takes a file in DIMACS-max format and returns a DiGraph single-source
    labeled s, single sink labeled t.
    '''
    N = nx.DiGraph()
    with open(netFile, 'r') as f:
        for line in f:
            items = line.split()
            if items[0] == 'p':
                assert len(items) == 4
                n = int(items[2])
                m = int(items[3])
            elif items[0] == 'n':
                assert len(items) == 3
                if items[2] == 's':
                    s = items[1]
                elif items[2] == 't':
                    t = items[1]
                else:
                    raise ValueError('Unspecified entry in line {}'.
                                     format(line))
            elif items[0] == 'a':
                assert len(items) == 4 
                N.add_edge(items[1], items[2], upper = int(items[3]))
                assert int(items[3]) != float('inf')
    assert s and t
    nx.relabel_nodes(N, {s: 's'}, copy=False)
    nx.relabel_nodes(N, {t: 't'}, copy=False)
    assert len(N) == n and len(N.edges()) == m
    return N


def convertMinToGraph(netFile):
    '''Takes a file in DIMACS-min format and returns a DiGraph,
    discarding the costs of arcs. '''
    N = nx.DiGraph()
    with open(netFile, 'r') as f:
        countDoubleArcs = 0
        for line in f:
            items = line.split()
            if items[0] == 'p':
                assert len(items) == 4
                n = int(items[2])
                m = int(items[3])
            elif items[0] == 'n':
                assert len(items) == 3
                N.add_node(items[1], supply = int(items[2]))
            elif items[0] == 'a':
                assert len(items) == 6 
                if N.has_edge(items[1], items[2]):
                    countDoubleArcs += 1
                    N[items[1]][items[2]]['lower'] += int(items[3])
                    N[items[1]][items[2]]['upper'] += int(items[4])
                else:
                    N.add_edge(items[1], items[2], lower = int(items[3]),
                               upper = int(items[4]))

    #assert len(N) == n and len(N.edges()) + countDoubleArcs == m
    return N


def convertRmfToGraph(netFile):
    '''Takes a file in rmf format (special ZIB-max format) and returns a DiGraph,
    contains no arc costs.'''
    N = nx.DiGraph()
    with open(netFile, 'r') as f:
        lines = f.readlines()
        n = int(splitLine(lines[0])[0])
        m = int(splitLine(lines[1])[0])
        s = splitLine(lines[2])[0]
        t = splitLine(lines[3])[0]
        for line in islice(lines,4,None):
            items = splitLine(line, 3)
            N.add_edge(items[0], items[1], upper = int(float(items[2])))
            assert int(float(items[2])) != float('inf')
    assert s and t
    nx.relabel_nodes(N, {s: 's'}, copy=False)
    nx.relabel_nodes(N, {t: 't'}, copy=False)
    assert len(N) == n and len(N.edges()) == m
    return N


def main(netFile, outputFile):
    '''Takes DIMACS-max or -min, or rmf format and prints a graphml file.
    '''
    if netFile.endswith('max'):
        N = convertMaxToGraph(netFile)
    elif netFile.endswith('min'):
        N = convertMinToGraph(netFile)
    elif netFile.endswith('rmf'):
        N = convertRmfToGraph(netFile)
    else:
        sys.exit("File format not supported")

    nx.write_graphml(N, outputFile)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('networkFile', help='old network file')
    parser.add_argument('graphmlFile', help='new graphml file', nargs='?',
                        default='test.graphml')    
    args = parser.parse_args()
    main(args.networkFile, args.graphmlFile)    








