#!/usr/bin/python2
# -*- coding: utf-8 -*-
'''
Make a feasible instance infeasible.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import sys        
import argparse
import networkx as nx
import random
from math import floor, ceil
from itertools import islice

import netmic.network_utility as netUtil
import netmic.graphml_checker as check


# assumes that neither supplies nor bounds exist
def undirectedToRandomlyDirectedGraph(undirN):
    '''Return a randomly directed copy of dirN.'''
    N = nx.DiGraph()
    for (v,w) in undirN.edges_iter():
        if random.choice([True, False]):
            N.add_edge(v, w)
        else:
            N.add_edge(w, v)
    return N


def balanceSupplyByAddition(N, maxB=float('inf')):
    '''Substract or add maximal possible supply with respect to maxB from random
    nodes until network is balanced.
    '''
    unBalancedSum = netUtil.sumBInS(N, N.nodes())
    while unBalancedSum > 0:
        v = random.choice(N.nodes())
        changeB = min(maxB + N.node[v]['supply'], unBalancedSum)
        N.node[v]['supply'] -= changeB
        unBalancedSum -= changeB
    while unBalancedSum < 0:
        v = random.choice(N.nodes())
        changeB = max(-maxB + N.node[v]['supply'], unBalancedSum)
        N.node[v]['supply'] -= changeB
        unBalancedSum -= changeB

        
def balanceSupply(N, maxB, decimal=0):
    '''Balance supply in N, by multiplying all except one supplies with a factor,
    the remainder is added to the fixed one.
    '''
     # check if N is already balanced
    sumTotal = netUtil.sumBInS(N, N.nodes_iter())
    if sumTotal == 0:  
        return

    # Node '0' is used to balance possible remainders, so sum up supply for all
    # nodes except '0'.
    unBalancedSum = netUtil.sumBInS(N, islice(N.nodes(),1,None))
    # If the sum is zero, the factor used below would be inf.
    # Hence, we use the the 'backup' function.
    if unBalancedSum == 0:
        balanceSupplyByAddition(N, maxB)
        return
    
    factor = float(maxB) / unBalancedSum
    if abs(factor) >= 1:
        balanceSupplyByAddition(N, maxB)
        return

    newSum = 0    
    if unBalancedSum > 0:
        for v in islice(N.nodes(),1,None):
            if not decimal:
                N.node[v]['supply'] = int(floor(factor * N.node[v]['supply']))
            else:
                N.node[v]['supply'] = round(factor * N.node[v]['supply'], decimal)
            newSum += N.node[v]['supply']
    else:
        for v in islice(N.nodes(),1,None):
            if not decimal:
                N.node[v]['supply'] = int(ceil(-factor * N.node[v]['supply']))
            else:
                N.node[v]['supply'] = round(-factor * N.node[v]['supply'], decimal)
            newSum += N.node[v]['supply']        
    N.node[N.nodes()[0]]['supply'] = round(-newSum, decimal)

def addArbitrarySuppliesAndBoundsFloat(N, maxB, maxU, decimal):
    '''Set the supply of all nodes to arbitrary values in (-maxB, maxB), and the
    upper bounds of all arcs to arbitrary values in (0, maxU). All values floats.
    '''
    for v in N.nodes_iter(data=True):
        v[1]['supply'] = round(random.uniform(-maxB, maxB), decimal)
    balanceSupply(N, maxB, decimal)
    assert round(netUtil.sumBInS(N, N.nodes()), 9) == 0
    
    for a in N.edges_iter(data=True):
        a[2]['upper'] = round(random.uniform(0, maxU), decimal)

                     
def addArbitrarySuppliesAndBounds(N, maxB, maxU):
    '''Set the supply of all nodes to arbitrary values in (-maxB, maxB), and the
    upper bounds of all arcs to arbitrary values in (0, maxU). All values ints.
    '''
    for v in N.nodes_iter(data=True):
        v[1]['supply'] = random.randint(-maxB, maxB)
    balanceSupply(N, maxB)
    assert netUtil.sumBInS(N, N.nodes_iter()) == 0
    
    for a in N.edges_iter(data=True):
        a[2]['upper'] = random.randint(0, maxU)


def addMaxFlowPercentage(N, p, decimal=0):
    '''Set the supply and demand of node "s" and "t", resp.,
    to (1+p)*max-flow value.
    '''
    assert N.has_node('s') and N.has_node('t')
    if not decimal:
        flow = int(ceil(nx.maximum_flow_value(N, 's', 't', "upper") * (1.0 + p)))
    else:
        flow = round(nx.maximum_flow_value(N, 's', 't', "upper") * (1.0 + p),
                     decimal)
    if flow == 0:
        flow = 1
    N.node['s']['supply'] =  flow
    N.node['t']['supply'] = -flow


def computeInfeasibleBFactor(N, ssN):
    '''Return a factor c, such that c*supply-vector will render N infeasible.'''
    # delete the upper bounds for all arcs (s,v), (w,t); save the sum of original
    # supply values
    bPlus = 0.0
    for v in ssN.successors('s'):
        bPlus += ssN.edge['s'][v]['upper']
        del ssN.edge['s'][v]['upper']
    for v in ssN.predecessors('t'):
        del ssN.edge[v]['t']['upper']

    # Compute max-flow without given supplies
    try:
        flow, x = nx.maximum_flow(ssN, 's', 't', 'upper')
    # the problem might be unbounded, then there are infinite capacities on arcs
    # other than (s,v),(w,t). Set these to bPlus (also in original network) and
    # compute max-flow
    except nx.NetworkXUnbounded:   
        for v,w in ssN.edges_iter():
            if 'upper' not in ssN.edge[v][w]:
                ssN.edge[v][w]['upper'] = bPlus
                N.edge[v][w]['upper'] = bPlus
        flow, x = nx.maximum_flow(ssN, 's', 't', 'upper')

    # factor is the maximum ratio between supply in max-flow and original supply
    return max(float(x['s'][v])/N.node[v]['supply']
               for v in netUtil.listOfSupplyNodes_iter(N))
        
def scaleSupplies(N, alpha, decimal):
    '''Multiply all supplies with alpha and round to decimal places.'''
    if not decimal:
        for v in N:
            N.node[v]['supply'] = int(round(alpha*N.node[v]['supply'], 0))
    else:
        for v in N:
            N.node[v]['supply'] = round(alpha*N.node[v]['supply'], decimal)
    balanceSupplyByAddition(N)

def scaledInfeasibleNetwork(N, decimal):
    '''If N is feasible, scale supplies with a factor, so that N is infeasible
    afterwards (requires $b\ne 0$).'''
    ssN = netUtil.setUpSingleSourceSinkNetwork(N)
    netUtil.removeInfiniteUpperBoundsForMaxFlow(ssN)
    flow = nx.maximum_flow_value(ssN, 's', 't', 'upper')

    if netUtil.sumBPlus(N) > flow:
        return 

    infFactor = computeInfeasibleBFactor(N, ssN)
    scaleSupplies(N, infFactor, decimal)
    

def supplyFromUpperBoundsSingleSource(N):
    '''Return the sum of upper bounds on arcs from "s".'''
    assert 's' in N.nodes()
    return sum(N.edge['s'][w]['upper'] for w in N.edge['s'])

    
def setNrBoundsToZeroSTCase(N, nrBounds):
    '''Set nr upper bounds to 0.
    Take randomly chosen $S\subset V$ and set random bound on outgoing arc to 0,
    until GH-ineq. is violated. Take next S, until nr is reached.
    Return True if at least one of the chosen S violates GH-ineq.
    Note: Network can be infeasible even if False is returned.
    Version for st-networks, where only node 's' has a positive supply.'''
    if nrBounds > len(N.edges()):
        raise ValueError("More bounds ({}) requested than there are arcs({}).".
                         format(nrBounds, N.edges()))

    flowValue = supplyFromUpperBoundsSingleSource(N)
    N.node['s']['supply'] = flowValue
    N.node['t']['supply'] = -flowValue

    # auxiliary nodeList to choose from, such that status of 's' and 't' can be
    # predetermined
    nodeList = N.nodes()
    del nodeList[nodeList.index('s')]
    del nodeList[nodeList.index('t')]
    success = False
    # abort criterion, if too many sets S are unsuccessfully repeated
    nrShotsS = 2**(len(nodeList))

    while nrBounds > 0 and nrShotsS > 0:
        S = random.sample(nodeList,random.randint(0,len(nodeList)))
        S.append('s')
        nrShotsS -= 1
        delta = netUtil.deltaPlus(N, S)
        U = netUtil.sumUInA(N, delta)
        # set bounds in delta+(S) to zero as long as S is feasible 
        while flowValue <= U and nrBounds > 0:
            v,w = random.choice(delta)
            del delta[delta.index((v,w))]   # do not repeat arc
            if N.edge[v][w]['upper'] > 0:
                U -= N.edge[v][w]['upper']
                N.edge[v][w]['upper'] = 0
                nrBounds -= 1
        if flowValue > U:
            success = True
    return success

def setNrBoundsToZero(N, nrBounds):
    '''Set nr upper bounds to 0.
    Take randomly chosen $S\subset V$ and set random bound on outgoing arc to 0,
    until GH-ineq. is violated. Take next S, until nr is reached.
    Return True if at least one of the chosen S violates GH-ineq.
    Note: Network can be infeasible even if False is returned.
    General version, where more than one node has a positive supply;
    might have lower bounds.'''
    nrBoundsInN = len(N.edges()) + len([a for a in N.edges_iter() if 'lower' in
                                        N.edge[a[0]][a[1]] and
                                        N.edge[a[0]][a[1]]['lower'] != 0])
    if nrBounds > nrBoundsInN:
        raise ValueError("More bounds ({}) requested than there are bounds({}).".
                         format(nrBounds, nrBoundsInN))

    success = False
    # abort criterion, if too many sets S are unsuccessfully repeated
    nrShotsS = 2**(len(N.nodes()))

    while nrBounds > 0 and nrShotsS > 0:
        B = 0
        while B == 0: # innodes cannot be made infeasible in this way
            S = random.sample( N.nodes(),random.randint(1, len(N.nodes())-1) )
            B = netUtil.sumBInS(N, S)
            nrShotsS -= 1
        if B > 0: # supply case, upper bounds from deltaPlus
            setUpper = netUtil.deltaPlus(N, S)
            setLower = netUtil.deltaMinusIfLowerBound(N, S)
        else:      # demand case, upper bounds from deltaMinus
            B = -1*B
            setUpper = netUtil.deltaMinus(N, S)
            setLower = netUtil.deltaPlusIfLowerBound(N, S)
            
        U = netUtil.sumUInA(N, setUpper)
        L = netUtil.sumLInA(N, setLower)

        # set upper bound to lower bound and vice versa, as long as S is feasible 
        while B <= U - L and nrBounds > 0:
            if len(setUpper) + len(setLower) == 0:  # no more bounds left
                break
            if random.randint(1,len(setUpper)+len(setLower)) <= len(setUpper):
            # choose from upper bounds
                v,w = random.choice(setUpper)
                del setUpper[setUpper.index((v,w))]   # do not repeat arc
                if 'lower' in N.edge[v][w] and N.edge[v][w]['lower'] != 0: 
                    if N.edge[v][w]['upper'] > N.edge[v][w]['lower']:
                        U -= N.edge[v][w]['upper'] - N.edge[v][w]['lower'] 
                        N.edge[v][w]['upper'] = N.edge[v][w]['lower']
                        nrBounds -= 1
                elif N.edge[v][w]['upper'] > 0:
                    U -= N.edge[v][w]['upper'] 
                    N.edge[v][w]['upper'] = 0
                    nrBounds -= 1
            else:  # choose from lower bounds
                v,w = random.choice(setLower)
                del setLower[setLower.index((v,w))] # do not repeat arc                    
                if N.edge[v][w]['lower'] <  N.edge[v][w]['upper']:
                    L += N.edge[v][w]['upper'] - N.edge[v][w]['lower']
                    N.edge[v][w]['lower'] = N.edge[v][w]['upper']
                    nrBounds -= 1
        if B > U - L:
            success = True
    return success

def halveBounds(N, deciPlaces):
    '''Set u = u/2 as long as N is feasible. Then double supply.'''
    while netUtil.feasibilityCheck(N):
        for _,_,d in N.edges_iter(data=True):
            if not deciPlaces:
                d['upper'] = d['upper']/2
            else:
                d['upper'] = round(d['upper']/2., deciPlaces)
    for _,d in N.nodes_iter(data=True):
        d['supply'] *= 2


def setNrBoundsRandom(N, nrBounds):
    '''Set upper bounds to 0 on a random arc, nrBounds times, and multiply supply
    with 2. If network is still infeasible afterwards, repeat.
    If it takes too long, aboard and return False.'''
    if nrBounds > N.number_of_edges():
        raise ValueError("More bounds ({}) requested than there are edges({}).".
                         format(nrBounds, N.number_of_edges()))

    shots = N.number_of_edges()/nrBounds
    while shots > 0:
        shots -= 1
        for v,d in N.nodes_iter(data=True):
            d['supply'] *= 10
        CA = random.sample(N.edges(), nrBounds)
        for v,w in CA:
            N.edge[v][w]['upper'] = 0
        if not netUtil.feasibilityCheck(N):
            return True
    return False

    
def main_max(args):
    '''Handle networks resulting from max-format: upper bounds, no supplies,
    asserts node names "s" and "t".
    case p: Set supply of s and t, resp., to (1+p)*max-flow value.
    case maxBounds: Set maxBounds upper bounds to 0'''
    random.seed(args.seed)
    N = nx.read_graphml(args.networkFile)
    assert N.is_directed()
    nx.relabel_nodes(N, {args.s: "s"}, copy=False)
    nx.relabel_nodes(N, {args.t: "t"}, copy=False)

    if args.p:
        addMaxFlowPercentage(N, args.p, args.deciPlaces)
    else:
        success = setNrBoundsToZeroSTCase(N, args.nrBounds)
        if not success and netUtil.feasibilityCheck(N):
            sys.exit('Given no. of bounds ({}) was not enough for infeasibility.'.
                     format(args.nrBounds))
    assert not netUtil.feasibilityCheck(N)

    nx.relabel_nodes(N, {"s": args.s}, copy=False)
    nx.relabel_nodes(N, {"t": args.t}, copy=False)
    nx.write_graphml(N, args.graphmlFile)
    
    
def main_graph(args):
    '''Handle networks resulting from graph-format: undirected graph, no bounds,
    no supplies.'''
    random.seed(args.seed)
    N = nx.read_graphml(args.networkFile)
    if not N.is_directed():
        N = undirectedToRandomlyDirectedGraph(N)
    if not args.deciPlaces:
        addArbitrarySuppliesAndBounds(N, args.bMax, args.uMax)
    else:
        addArbitrarySuppliesAndBoundsFloat(N, args.bMax, args.uMax,
                                           args.deciPlaces)
    if args.halve:
        halveBounds(N, args.deciPlaces)
    elif args.scale:
        if len(netUtil.listOfSupplyNodes(N)) == 0:
               sys.exit('Given network has zero supply (or no supply, max-'
               'format?). Scaling won\'t work.')
        scaledInfeasibleNetwork(N, args.deciPlaces)
    elif args.nrBounds:
        success = setNrBoundsToZero(N, args.nrBounds)
        if not success and netUtil.feasibilityCheck(N):
            sys.exit('Given no. of bounds ({}) was not enough for infeasibility.'.
                     format(args.nrBounds))
    else:
        success = setNrBoundsRandom(N, args.nrRandom)
        if not success:
            sys.exit('Try again with different seed.')
     
    assert not netUtil.feasibilityCheck(N)
    nx.write_graphml(N, args.graphmlFile)

    
def main_min(args):
    '''Handle networks resulting from min-format: directed graph, supplies,
    upper and possibly lower bounds.'''
    random.seed(args.seed)

    # we need a simple graph with default values in this case
    dataChecker = check.graphmlConsistency(nx.read_graphml(args.networkFile))
    N = dataChecker.newN

    if args.scale:
        if len(netUtil.listOfSupplyNodes(N)) == 0:
               sys.exit('Given network has zero supply (or no supply, max-'
               'format?). Scaling won\'t work.')
        scaledInfeasibleNetwork(N, args.deciPlaces)
    elif args.nrBounds:
        success = setNrBoundsToZero(N, args.nrBounds)
        if not success and netUtil.feasibilityCheck(N):
            sys.exit('Given no. of bounds ({}) was not enough for infeasibility.'.
                     format(args.nrBounds))
    else:
        success = setNrBoundsRandom(N, args.nrRandom)
        if not success:
            sys.exit('Try again with different seed.')
     
    assert not netUtil.feasibilityCheck(N)
    nx.write_graphml(N, args.graphmlFile)

    
###############################################################################        
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Takes a network and returns an'
                                     ' infeasible network, both in graphml '
                                     'format.')
    parser.add_argument('networkFile', help='old graphml file')
    parser.add_argument('graphmlFile', help='new graphml file', nargs='?',
                        default='new.graphml')
    parser.add_argument('--seed', help='seed for random numbers', type=int,
                        default=5)
    parser.add_argument('-d', '--deciPlaces', help="number of decimal places in "
                        "data, 0 for int values",  type=int, default=0)
    
    subparsers = parser.add_subparsers()

    # max parser
    parser_max = subparsers.add_parser('max', help="assumes data for max-flow-"
                                       "problems, with single source and sink")
    parser_max.set_defaults(func=main_max)
    group_max = parser_max.add_mutually_exclusive_group(required=True)
    group_max.add_argument('-a', '--maxNumberBounds', help="MIAC will be smaller"
                           " than a", type=int, dest='nrBounds')
    group_max.add_argument('-p', '--additionalFlowPercentage', help="supply will"
                           " be (1+p) * max-flow", type=float, dest='p')
    parser_max.add_argument('-s', '--source', help="source node", default='s',
                              dest='s')
    parser_max.add_argument('-t', '--sink', help="sink node", default='t',
                              dest='t')

    # graph parser
    parser_graph = subparsers.add_parser('graph', help="assumes pure graph data,"
                                         " without supplies or bounds")
    parser_graph.add_argument('-b', '--upperBoundSupply', help="largest absolute"
                              " value for supply", type=int, default=1000,
                              dest='bMax')
    parser_graph.add_argument('-u', '--upperBoundBound', help="largest absolute "
                              "value for flow bounds", type=int, default=50,
                              dest='uMax')
    group_graph = parser_graph.add_mutually_exclusive_group(required=True)
    group_graph.add_argument('-v', '--halve', help="halve u until infeasible",
                             action='store_true', dest='halve')
    group_graph.add_argument('-s', '--scale', help="scale b with a factor",
                             action='store_true', dest='scale')
    group_graph.add_argument('-a', '--maxNumberBounds', help="set (at most) a "
                             " bounds to zero", type=int, dest='nrBounds')
    group_graph.add_argument('-r', '--random', help="set (at least) r "
                             "bounds to zero", type=int, dest='nrRandom')
    parser_graph.set_defaults(func=main_graph)

    # min parser
    parser_min = subparsers.add_parser('min', help="assumes data for min-cost-"
                                       "flows, with supplies and bounds")
    parser_min.set_defaults(func=main_min)
    group_min = parser_min.add_mutually_exclusive_group(required=True)
    group_min.add_argument('-s', '--scale', help="scale b with a factor",
                           action='store_true', dest='scale')
    group_min.add_argument('-a', '--maxNumberBounds', help="set (at most) a "
                           "bounds to zero", type=int, dest='nrBounds')
    group_min.add_argument('-r', '--random', help="set (at least) r "
                           "bounds to zero", type=int, dest='nrRandom')
 
    args = parser.parse_args()
    args.func(args)


