#!/usr/bin/python2
# -*- coding: utf-8 -*-
'''
Produce random graphs; sparse, dense or arbitrary.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import os
import argparse
import networkx as nx
import random
import errno


def deleteRandomEdges(G, nrDelEdges):
    '''Delete nrDeleteEdges many edges from G if they do not connect leaves,
    asserts at least n-1 + nrDelEdges edges in G.'''
    if len(G.edges()) < len(G.nodes())-1 + nrDelEdges:
        raise IndexError('Graph with {} nodes has not enough edges ({}) to '
                         'delete {} edges.'.format(len(G.nodes()), len(G.edges()),
                                                   nrDelEdges))
    while nrDelEdges > 0:
        v,w = random.choice(G.edges())
        if nx.degree(G, v) > 1 and nx.degree(G, w) > 1:
            G.remove_edge(v,w)
            nrDelEdges -= 1    

def makeConnected(G):
    '''Add edges between components of G, delete same number of arbitrary edges.'''
    m = len(G.edges())
    if m < len(G.nodes())-1:
        raise IndexError('Graph has not enough edges to be connected.')

    components = list(nx.connected_components(G))
    while len(components) > 1:
        for i in xrange(len(components) -1):
            G.add_edge(components[i][0], components[i+1][0])
        deleteRandomEdges(G, len(components) -1)
        components = list(nx.connected_components(G))
    assert m == len(G.edges())

    
def randomDenseGraph(n):
    '''Return an undirected dense graph.
    Number of edges is a percentage p of the complete graph,
    where p is normal distributed with $\mu = 0.85, \sigma = 0.035$.'''
    G = nx.complete_graph(n)
    p = 5.
    while p > 1.0:
        p = random.gauss(0.85, 0.035)
    deleteRandomEdges(G, n*(n-1)/2 - int(p * n*(n-1)/2))
    makeConnected(G)  
    return G

    
def randomSpanningTree(G):
    '''Add a random spanning tree to G, by mimicking Kruskals algorithm.'''
    components = list(nx.connected_components(G))
    while len(components) > 1:
        comp1, comp2 = random.sample(components, 2)
        G.add_edge(random.choice(comp1), random.choice(comp2))
        components = list(nx.connected_components(G))
    assert len(G.edges()) == len(G.nodes())-1

    
def addRandomEdges(G, nrAddEdges):
    '''Add nrAddEdges random edges to G.'''
    m = len(G.edges()) + nrAddEdges
    n = len(G.nodes())
    if m > n*(n-1)/2:
        raise IndexError('Cannot add {} edges to graph with {} nodes and {} '
                         'edges.'.format(nrAddEdges, n, len(G.edges())))
    while nrAddEdges > 0:
        v,w = random.sample(G.nodes(), 2)
        if not G.has_edge(v,w):
            G.add_edge(v,w)
            nrAddEdges -= 1
    assert m == len(G.edges()) or nrAddEdges < 0

    
def randomSparseGraph(n):
    '''Return an undirected sparse, but connected graph.
    If n is large enough, number of edges is a percentage p of the complete graph,
    where p is normal distributed with $\mu = 0.06, \sigma = 0.015$;
    otherwise, at least a spanning tree exists.'''
    G = nx.Graph()
    G.add_nodes_from(range(n))
    randomSpanningTree(G)  # make sure graph is connected
    p = -5.
    while p < 0.:
        p = random.gauss(0.06, 0.015)
    addRandomEdges(G, int(n*(n-1)/2*p) - n+1)
    return G

    
def main_dense(args):
    '''Print random, undirected, connected, dense graphs to graphml-files.'''
    newDir = args.newDir + "dense" 
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

    random.seed(args.seed)
    for i in xrange(args.g):
        G = randomDenseGraph(args.n)
        file_name = 'dense_' + str(args.n) + '_' + str(i) + '.graphml'
        nx.write_graphml(G, os.path.join(newDir, file_name))

def main_sparse(args):
    '''Print random, undirected, connected, sparse graphs to graphml-files.'''
    newDir = args.newDir + "sparse" 
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

    random.seed(args.seed)
    for i in xrange(args.g):
        G = randomSparseGraph(args.n)
        file_name = 'sparse_' + str(args.n) + '_' + str(i) + '.graphml'
        nx.write_graphml(G, os.path.join(newDir, file_name))
        
def main_arbi(args):
    '''Print random, undirected, connected graphs to graphml-files.'''
    newDir = args.newDir + "random" 
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

    random.seed(args.seed)
    for i in xrange(args.g):
        p = random.uniform(0.2, 0.7)
        m = max(int(p*args.n*(args.n-1)/2), args.n-1)
        G = nx.gnm_random_graph(args.n, m, seed=args.seed)
        makeConnected(G)
        file_name = 'arbitrary_' + str(args.n) + '_' + str(i) + '.graphml'
        nx.write_graphml(G, os.path.join(newDir, file_name))

        
###############################################################################        
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Print (an) undirected, random '
                                     'graph(s) in graphml format.')
    parser.add_argument('n', help='number of nodes', type=int)
    parser.add_argument('g', help='number of produced graphs', type=int)
    parser.add_argument('-newDir', help='direction for the new data-folder',
                        nargs='?', default='')
    parser.add_argument('--seed', help='seed for random numbers', nargs='?',
                        type=int, default=42)
    
    subparsers = parser.add_subparsers()

    # "dense" parser
    parser_dense = subparsers.add_parser('dense', help="Graphs will be dense.")
    parser_dense.set_defaults(func=main_dense)
    # "sparse" parser
    parser_sparse = subparsers.add_parser('sparse', help="Graphs will be sparse.")
    parser_sparse.set_defaults(func=main_sparse)
    # "arbitrary" parser
    parser_arbi = subparsers.add_parser('arbi', help="Graphs will be neither "
                                        "sparse, nor dense.")
    parser_arbi.set_defaults(func=main_arbi)

    args = parser.parse_args()
    args.func(args)

