#!/usr/bin/python2
# -*- coding: utf-8 -*-
'''
Produce connected graphs of special classes.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import sys        
import os
import errno
import argparse
import networkx as nx
from networkx.algorithms import bipartite
import random
import warnings
from generator.make_random_graphs import deleteRandomEdges


def main_regular(args):
    '''Print random regular graphs to graphml-file.'''
    if args.g > 1:
        warnings.warn("Check for repeating graphs, esp. if n and d are close!")
        
    newDir = args.newDir + "regular" 
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    for i in range(args.g):
        G = nx.random_regular_graph(args.d, args.n)
        file_name = ('regular_' + str(args.n) + '_d' + str(args.d) + '_' + str(i)
                     + '.graphml')
        if not nx.is_connected(G):
            warnings.warn("Did not write " + file_name + ", graph was "
                          "disconnected.")
        else:
            nx.write_graphml(G, os.path.join(newDir, file_name))        

def main_wheel(args):
    '''Print wheel graph to graphml-file.'''
    newDir = args.newDir + "wheel"
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    G = nx.wheel_graph(args.n)
    file_name = 'wheel_' + str(args.n) + '.graphml'
    nx.write_graphml(G, os.path.join(newDir, file_name))
      
def main_star(args):
    '''Print star graph to graphml-file.'''
    newDir = args.newDir + "star"
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    G = nx.star_graph(args.n)
    file_name = 'star_' + str(args.n+1) + '.graphml'
    nx.write_graphml(G, os.path.join(newDir, file_name))
    
def main_hypercube(args):
    '''Print hypercube graph to graphml-file.'''
    if args.n > 15:
        choice = raw_input("This will produce {} nodes. Do you want to continue "
                           "(y,Y)? ".format(2**args.n))
        if choice != 'y' and choice != 'Y':
            sys.exit("Good choice!")
            
    newDir = args.newDir + "hypercube"
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    G = nx.hypercube_graph(args.n)
    file_name = 'hypercube_' + str(2**args.n) + '.graphml'
    nx.write_graphml(G, os.path.join(newDir, file_name))

def main_grid(args):
    '''Print 2d grid graph to graphml-file.'''
    newDir = args.newDir + "grid"
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    G = nx.grid_2d_graph(args.n, args.n2)
    file_name = 'grid_' + str(args.n*args.n2) + '_' + str(args.n) + '.graphml'
    nx.write_graphml(G, os.path.join(newDir, file_name))
    
def main_circ_ladder(args):
    '''Print circular ladder graph to graphml-file.'''
    newDir = args.newDir + "circ_ladder"
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
        G = nx.circular_ladder_graph(args.n)
        file_name = 'circ_ladder_' + str(2*args.n) + '.graphml'
        nx.write_graphml(G, os.path.join(newDir, file_name))

def main_ladder(args):
    '''Print ladder graph to graphml-file.'''
    newDir = args.newDir + "ladder"
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    G = nx.ladder_graph(args.n)
    file_name = 'ladder_' + str(2*args.n) + '.graphml'
    nx.write_graphml(G, os.path.join(newDir, file_name))
    
def makeBipartiteConnected(G):
    '''Add edges between components of G, delete same number of arbitrary edges.'''
    assert bipartite.is_bipartite(G)
    m = len(G.edges())
    if m < len(G.nodes())-1:
        raise IndexError('Graph has not enough edges to be connected.')
        
    while not nx.is_connected(G):
        components = list(sorted(nx.connected_components(G), key=len,
                                 reverse=True))
        assert len(components[0]) > 1
        # connect all components to components[0], since this must contain nodes
        # from both sets
        for v in (components[i][0] for i in xrange(1, len(components))):
            j = 0
            while G.node[components[0][j]]['bipartite'] == G.node[v]['bipartite']:
                assert G.has_node(components[0][j+1])
                j += 1
            G.add_edge(v, components[0][j])
        deleteRandomEdges(G, len(components) -1)
    assert m == len(G.edges())

def main_bipartite(args):
    '''Print random bipartite graphs to graphml-file.'''
    newDir = args.newDir + "bipartite" 
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    random.seed(args.seed)
    for i in range(args.g):
        m = random.randint(args.n+args.n2, args.n*args.n2)
        G = nx.bipartite_gnmk_random_graph(args.n, args.n2, m)
        makeBipartiteConnected(G)
        file_name = ('bipartite_' + str(args.n + args.n2) + '_' + str(args.n) 
                     + '_' + str(i) + '.graphml')
        nx.write_graphml(G, os.path.join(newDir, file_name))        

def main_barbell(args):
    '''Print random barbell graph to graphml-file.'''
    newDir = args.newDir + "barbell" 
    try:
        os.makedirs(newDir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    G = nx.barbell_graph(args.n, args.n2)
    file_name = ('barbell_' + str(2*args.n + args.n2) + '_' + str(args.n)
                 + '.graphml')
    nx.write_graphml(G, os.path.join(newDir, file_name))

     
###############################################################################        
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Print (an) undirected graph(s)'
                                     ' in graphml format.')
    parser.add_argument('n', help='number of nodes', type=int)
    
    parser.add_argument('-newDir', help='direction for the new data-folder',
                        nargs='?', default='')
    parser.add_argument('--seed', help='seed for random numbers', nargs='?',
                        type=int, default=42)
    
    subparsers = parser.add_subparsers()

    parser_barbell = subparsers.add_parser('barbell', help="Barbell graph, two "
                                           "K_n connected with P_n2, in total "
                                           "2*n+n2 nodes")
    parser_barbell.add_argument('n2', help="number of nodes on path", type=int)
    parser_barbell.set_defaults(func=main_barbell)

    parser_bipartite = subparsers.add_parser('bipartite', help="Bipartite graph"
                                             " first node set with n nodes, "
                                             "second with n2 nodes")
    parser_bipartite.add_argument('n2', help='number of nodes in the second set',
                                 type=int)
    parser_bipartite.add_argument('g', help='number of produced graphs', type=int)
    parser_bipartite.set_defaults(func=main_bipartite)
    
    parser_ladder = subparsers.add_parser('ladder', help="Ladder graph, two paths"
                                          " with n nodes, connected by rungs")
    parser_ladder.set_defaults(func=main_ladder)

    parser_circ_ladd = subparsers.add_parser('circ_ladder', help="Circular ladder"
                                             " graph, two n cycles with rungs "
                                             "between nodes i and n+i")
    parser_circ_ladd.set_defaults(func=main_circ_ladder)

    parser_grid = subparsers.add_parser('grid', help="Grid graph, with n*n2 nodes")
    parser_grid.add_argument('n2', help='second dimension of the grid', type=int)
    parser_grid.set_defaults(func=main_grid)

    parser_cube = subparsers.add_parser('hypercube', help="Hypercube graph with "
                                        "2**n nodes")
    parser_cube.set_defaults(func=main_hypercube)

    parser_star = subparsers.add_parser('star', help="Star graph with n+1 nodes")
    parser_star.set_defaults(func=main_star)

    parser_wheel = subparsers.add_parser('wheel', help="Wheel graph")
    parser_wheel.set_defaults(func=main_wheel)

    parser_regular = subparsers.add_parser('regular', help="Regular graph with "
                                           "n nodes and degree d, where n*d must"
                                           " be even")
    parser_regular.add_argument('d', help='degree of the nodes', type=int)
    parser_regular.add_argument('g', help='number of produced graphs', type=int)
    parser_regular.set_defaults(func=main_regular)

    args = parser.parse_args()
    args.func(args)


