#!/usr/bin/python2
# -*- coding: utf-8 -*-

import networkx as nx
import unittest
import random
import netmic.network_utility as netUtil

from generator.data_infeasibilitizer import *

class auxGraphFunctionTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=5)
        self.G.add_node(2, supply=-3)
        self.G.add_node(3, supply=0)
        self.G.add_node(4, supply=-3)
        self.G.add_node(5, supply=1)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(3, 2, upper=2)
        self.G.add_edge(1, 4, upper=4)
        self.G.add_edge(5, 2, upper=1.3)

    def testHalveBounds(self):
        halveBounds(self.G, 0)
        self.assertEqual(self.G[1][4]['upper'], 2)
        self.assertEqual(self.G[5][2]['upper'], 0.65)
        self.assertEqual(netUtil.sumBPlus(self.G), 12)

class randomlyDirectedTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.Graph()
        self.G.add_nodes_from(range(5))
        self.G.add_edge(1, 2)
        self.G.add_edge(2, 4)
        self.G.add_edge(3, 2)
        self.G.add_edge(0, 2)
        self.G.add_edge(1, 4)

    def testRandomlyDirected(self):
        auxG = undirectedToRandomlyDirectedGraph(self.G)
        self.assertEqual(len(auxG.edges()), 5)
        self.assertTrue(nx.is_weakly_connected(auxG))
        
class arbitraryDataTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_nodes_from(range(5))
        self.G.add_edge(1, 2)
        self.G.add_edge(2, 4)
        self.G.add_edge(3, 2)
        self.G.add_edge(0, 2)
        self.G.add_edge(1, 4)

    def testAddArbitrarySuppliesAndBounds(self):
        addArbitrarySuppliesAndBounds(self.G, 10, 1)
        sumBPlus = sum(v[1]['supply'] for v in self.G.nodes(data=True)
                        if v[1]['supply'] > 0)
        self.assertTrue(sumBPlus <= 30)
        sumU = sum(a[2]['upper'] for a in self.G.edges(data=True))
        self.assertTrue(sumU <= 5)

    def testAddArbitrarySuppliesAndBoundsFloat(self):
        random.seed(5)
        addArbitrarySuppliesAndBoundsFloat(self.G, 9.9, 1.1, 2)
        self.assertEqual(self.G.node[0]['supply'], -9.9)
        self.assertEqual(self.G.node[1]['supply'], 1.96)
        sumU = sum(a[2]['upper'] for a in self.G.edges(data=True))
        self.assertAlmostEqual(sumU, 3.3 )

class balanceSupplyTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=-1)
        self.G.add_node(2, supply=4)
        self.G.add_node(3, supply=3)
        self.G.add_node(4, supply=-2)
        self.G.add_node(5, supply=3)

    def testBalanceSupply(self):
        balanceSupply(self.G, 5)
        self.assertEqual(self.G.node[1]['supply'], -2)
        self.assertEqual(netUtil.sumBInS(self.G, self.G.nodes_iter()), 0)

    def testBalanceSupplyWithSmallFactor(self):
        self.G.node[2]['supply'] = 1
        balanceSupply(self.G, 5)
        sumBPlus = sum(v[1]['supply'] for v in self.G.nodes_iter(data=True)
                        if v[1]['supply'] > 0)
        self.assertTrue(sumBPlus <= 10)


    def testBalanceSupplyByAddition(self):
        balanceSupplyByAddition(self.G, 5)
        self.assertEqual(netUtil.sumBInS(self.G, self.G.nodes_iter()), 0)
        sumBPlus = sum(v[1]['supply'] for v in self.G.nodes_iter(data=True)
                        if v[1]['supply'] > 0)
        self.assertTrue(sumBPlus <= 10)

class balanceSupplyFloat(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=-1.1)
        self.G.add_node(2, supply=4.23)
        self.G.add_node(3, supply=3.7)
        self.G.add_node(4, supply=-2.9)
        self.G.add_node(5, supply=3.0)

    def testBalanceSupply(self):
        balanceSupply(self.G, 5.1, 2)
        self.assertEqual(self.G.node[1]['supply'], -5.11)
        self.assertEqual(self.G.node[2]['supply'], 2.69)
        self.assertEqual(self.G.node[3]['supply'], 2.35)
        self.assertEqual(self.G.node[4]['supply'], -1.84)
        self.assertEqual(self.G.node[5]['supply'], 1.91)

    def testBalanceSupplyWithSmallFactor(self):
        self.G.node[2]['supply'] = 1.543
        balanceSupply(self.G, 5.1, 3)
        self.assertEqual(self.G.node[1]['supply'], -5.101)
        self.assertEqual(self.G.node[2]['supply'], 1.473)
        self.assertEqual(self.G.node[3]['supply'], 3.532)
        self.assertEqual(self.G.node[4]['supply'], -2.768)
        self.assertEqual(self.G.node[5]['supply'], 2.864)
        
class maxFlowPercentageTest(unittest.TestCase):
    def setUp(self): 
        self.G = nx.DiGraph()
        self.G.add_node('s')
        self.G.add_node('t')
        self.G.add_node(3)
        self.G.add_node(4)
        self.G.add_edge('s', 3, upper=2)
        self.G.add_edge(3, 't', upper=2)
        self.G.add_edge('s', 4, upper=4)

    def testMaxNetworkInfeasible(self):
        addMaxFlowPercentage(self.G, 0.2)
        self.assertEqual(self.G.node['s']['supply'], 3)

    def testMaxNetworkInfeasibleFloat(self):
        self.G.add_edge(4, 't', upper=1.22)
        addMaxFlowPercentage(self.G, 0.2, 3)
        self.assertEqual(self.G.node['s']['supply'], 3.864)
        
class maxFlowNrBoundsTest(unittest.TestCase):
    def setUp(self): 
        self.G = nx.DiGraph()
        self.G.add_node('s')
        self.G.add_node('t')
        self.G.add_node(1)
        self.G.add_node(2)
        self.G.add_node(3)
        self.G.add_node(4)
        self.G.add_edge('s', 1, upper=1)
        self.G.add_edge('s', 2, upper=3)
        self.G.add_edge(1, 2, upper=1)
        self.G.add_edge(1, 3, upper=1)
        self.G.add_edge(2, 4, upper=3)
        self.G.add_edge(4, 1, upper=2)
        self.G.add_edge(3, 't', upper=3)
        self.G.add_edge(4, 't', upper=1)

    def testSupplyFromUpperBoundsSingleSource(self):
        self.assertEqual(supplyFromUpperBoundsSingleSource(self.G), 4)
        
    def testSetNrBoundsTooLarge(self):
        with self.assertRaises(ValueError):
            setNrBoundsToZeroSTCase(self.G, 9)

    def testSetNrBoundsToZeroSmall(self):
        random.seed(5)
        self.assertFalse(setNrBoundsToZeroSTCase(self.G, 2))
        self.assertEqual(sum(self.G.edge[v][w]['upper'] for (v,w) in
                             self.G.edges_iter()), 13)
        self.assertEqual(self.G.edge['s'][1]['upper'], 0)

    def testSetNrBoundsToZeroSmall2(self):
        random.seed(3)
        self.assertTrue(setNrBoundsToZeroSTCase(self.G, 3))
        self.assertEqual(sum(self.G.edge[v][w]['upper'] for (v,w) in
                             self.G.edges_iter()), 8)
        self.assertEqual(self.G.edge[3]['t']['upper'], 0)        

class minTest(unittest.TestCase):
    def setUp(self): 
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=5)
        self.G.add_node(2, supply=-3)
        self.G.add_node(3, supply=0)
        self.G.add_node(4, supply=-3)
        self.G.add_node(5, supply=1)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(3, 2, upper=2)
        self.G.add_edge(1, 4, upper=4)
        self.G.add_edge(5, 2, upper=1.3)

    def testSetNrBoundsRandom(self):
        random.seed(42)
        self.G.node[3]['supply'] = -2
        self.G.node[2]['supply'] = -1
        setNrBoundsRandom(self.G, 1)
        self.assertEqual(self.G.edge[3][2]['upper'], 0)
        self.assertEqual(netUtil.sumBPlus(self.G), 60)

    def testComputeInfeasibleBFactor(self):
        ssN = netUtil.setUpSingleSourceSinkNetwork(self.G)
        self.assertAlmostEqual(computeInfeasibleBFactor(self.G,ssN), 1.3)

    def testScaleSupplies(self):
        scaleSupplies(self.G, 0.333333333, 0)
        self.assertEqual(self.G.node[1]['supply'], 2)
        self.assertEqual(self.G.node[2]['supply'], -1)
        self.assertEqual(self.G.node[3]['supply'], 0)

    def testScaleSuppliesFloat(self):
        random.seed(5)
        self.G.node[1]['supply'] = 5.1
        self.G.node[4]['supply'] = -3.1
        scaleSupplies(self.G, 1.234, 2)
        self.assertEqual(self.G.node[1]['supply'], 6.29)
        self.assertEqual(self.G.node[2]['supply'], -3.7)
        self.assertEqual(self.G.node[4]['supply'], -3.82)

    def testScaledInfeasibleNetwork(self):
        self.G.add_edge(5, 4, upper=1.1)
        scaledInfeasibleNetwork(self.G, 0)
        self.assertEqual(self.G.node[1]['supply'], 12)
        self.assertEqual(self.G.node[2]['supply'], -7)
        self.assertEqual(self.G.node[5]['supply'], 2)

    def testScaledInfeasibleNetworkFloat(self):
        scaledInfeasibleNetwork(self.G, 2)
        self.assertEqual(self.G.node[1]['supply'], 6.5)
        self.assertEqual(self.G.node[2]['supply'], -3.9)
        self.assertEqual(self.G.node[5]['supply'], 1.3)

    def testSetNrBoundsTooLarge(self):
        with self.assertRaises(ValueError):
            setNrBoundsToZero(self.G, 5)

    def testSetNrBoundsToZero(self):
        random.seed(5)
        self.assertTrue(setNrBoundsToZero(self.G, 3))
        self.assertEqual(sum(self.G.edge[v][w]['upper'] for (v,w) in
                             self.G.edges()), 1.3)

    def testSetNrBoundsToZero2(self):
        self.G.edge[1][4]['lower'] = 1.3
        self.G.edge[5][2]['lower'] = 1
        random.seed(3)
        self.assertTrue(setNrBoundsToZero(self.G, 4))
        self.assertEqual(sum(self.G.edge[v][w]['upper'] for (v,w) in
                             self.G.edges_iter()), 2.3)
 


if __name__ == '__main__':
    unittest.main()
