# -*- coding: utf-8 -*-
'''
Auxiliary function concerning solutions for flow networks.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
from itertools import product, ifilter, ifilterfalse
import math
from collections import deque
import netmic.network_utility as netUtil

def completeSolution(N, x):
    '''Completes a solution x with 0 values for every a not present in x.'''
    for v in ifilterfalse(lambda v: v in x,  N):
        x[v] = dict()
    for a in ifilterfalse(lambda a: a[1] in x[a[0]], N.edges_iter()):
        x[a[0]][a[1]] = 0
            
def xBoundFeasible(N, x):
    '''Return true if 0 <= x <= u.'''
    for a in N.edges_iter():
        (v, w) = a
        if round(x[v][w], 6) < 0:
            return False
        elif round(x[v][w] - N.edge[v][w]["upper"], 6) > 0:
            return False
    return True

def flowExcessAtNode(N, x, v):
    '''Return $b_v - x_{(v,w)} + x_{(w,v)}$.'''
    dv = N.node[v]["supply"]
    for a in N.out_edges_iter(v):
        dv -= x[v][a[1]]
    for a in N.in_edges_iter(v):
        dv += x[a[0]][v]
    return dv

def xBFlowFeasible(N, x):
    '''Return true if flow conservation is met in x.'''
    for v in N:
        if round(flowExcessAtNode(N, x, v), 4) != 0:
            return False
    return True

def unsaturatedSets(N, x):
    '''Return list of two node-lists with unsaturated supply/demand by x.'''
    S = []
    T = []
    for v in N:
        dv = flowExcessAtNode(N, x, v)
        if round(dv, 6) > 0:
            S.append(v)
        elif round(dv, 6) < 0:
            T.append(v)
    return [S,T]

def emptySolutionDict(N):
    '''Return a solution dict which contains an empty dict for every node.'''
    return dict([(u, {}) for u in N])

def coverFromSolution(N, x):
    '''Return a cover made of all bounds violated by x.'''
    C = emptySolutionDict(N)
    for a in N.edges_iter():
        (v, w) = a
        if round(x[v][w] - N.edge[v][w]["upper"], 6) > 0:
            C[v][w] = "u"
        elif round(x[v][w], 6) < 0:
            C[v][w] = "l"
    return C

def coverFromSolutionUndirected(N, x, edges):
    '''Return a cover made of bounds on "edges" violated by x; "edges" can be
    undirected and negative flow in x is given on reverse arc.'''
    C = emptySolutionDict(N)
    for e in edges:
        v,w = netUtil.networkArcFromEdge(N, e)
        assert v in x and w in x[v] or w in x and v in x[w]
        if v in x and w in x[v] and w in x and v in x[w]:
            assert x[v][w] == 0 or x[w][v] == 0

        if v in x and w in x[v] and round(x[v][w] - N.edge[v][w]['upper'],6) > 0:
            C[v][w] = 'u'
        if w in x and v in x[w] and round(x[w][v], 6) != 0:
            C[v][w] = 'l'
    return C

def coverFromSolutionResidual(N, x):
    '''Return a cover made of bounds violated by x; x is supposed to come from a
    residual network, i.e., all original arcs have a value and negative flow in
    x is given on reverse arc.'''
    C = emptySolutionDict(N)
    for v,w,d in N.edges_iter(data=True):
        assert v not in x[w] or x[w][v]==0 or x[v][w]==0
        if round(x[v][w] - d['upper'], 6) > 0:
            C[v][w] = 'u'
        elif v in x[w] and round(x[w][v], 6) > 0:
            C[v][w] = 'l'
    return C

def solutionFromResidualSolution(N, x):
    '''Return a solution on the original graph, build from a solution on the
    residual graph, where negative flow is given on the reversed arc.'''
    # delete all entries for "s" and "t" from x
    del x['s']
    del x['t']
    for v in netUtil.listOfDemandNodes_iter(N):
        del x[v]['t']
    # delete values for reversed arcs
    for v,w in ((v,w) for v in x for w in x[v].keys()):
        if not N.has_edge(v,w):
            if abs(x[v][w]) > 1e-6:
                assert x[w][v] == 0
                x[w][v] = -x[v][w]
            del x[v][w]

def solutionFromResidualGraph(N, R):
    '''Return a solution dict, picked from "flow" values in R.'''
    x = emptySolutionDict(N)
    for v,w in N.edges_iter():
        if R.has_edge(v,w):
            x[v][w] = R[v][w]['flow']
        else:
            x[v][w] = 0
    return x

def joinCovers(C1, C2):
    '''Replace C1 with the union of C1 and C2.'''
    for v in C2:
        if v not in C1:
            C1[v] = {}
        for w in C2[v]:
            if w in C1[v] and C1[v][w] != C2[v][w]:
                raise ValueError("Contradicting bounds for arc ({}, {})."
                                 .format(v,w))
            elif w not in C1[v]:
                C1[v][w] = C2[v][w]

def computePathWeight(N, path):
    '''Return the sum of 'weight' for arcs on path.'''
    return sum(N.edge[v][w]["weight"] for (v,w) in netUtil.arcsOnPath_iter(path))

def computeShortestPath(N, S, T):
    '''Return the shortest path between any node in S and any node in T.
    
    Used approach differs depending on the size of the larger set compared to
    nodes overall. If it is large, compute single-source Dijkstra for all nodes
    in the smaller set. Else, compute pairwise bidirectional Dijkstra. '''
    if max(len(S), len(T)) > math.sqrt(len(N)):
        return computeShortestPathSingleSource(N,S,T)
    else:
        return computeShortestPathBidirectional(N,S,T)

def computeShortestPathFloyd(N, S, T):
    '''Return the shortest path between any node in S and any node in T.
    
    Used approach differs depending on the size of the larger set compared to
    nodes overall. If it is large, compute Floyd-Warshall.
    Else, compute pairwise bidirectional Dijkstra. '''
    if (N.number_of_nodes()**3 < N.number_of_edges() *
        math.log(N.number_of_nodes()) * min(len(S), len(T)) ):
        return computeFloydWarshall(N,S,T)
    elif max(len(S), len(T)) > math.sqrt(len(N)):                       
        return computeShortestPathSingleSource(N,S,T)                           
    else:
        return computeShortestPathBidirectional(N,S,T)


def computeShortestPathSingleSource(N, S, T):
    '''Return the shortest path between any node in S and any node in T, by
    computing single-source Dijkstra for all nodes in the smaller set.'''
    if len(S) < len(T):
        reversed = False
        auxN,auxS,auxT = N,S,T
    else:
        reversed = True
        auxN,auxS,auxT = N.reverse(),T,S

    minCost,minPath = float('inf'), []
    for s in auxS:
        costDict, pathDict = nx.single_source_dijkstra(auxN, s, cutoff=minCost-1)
        for t in ifilter(lambda t: t in costDict and costDict[t] < minCost, auxT):
            minCost, minPath = costDict[t], pathDict[t]
    if reversed:
        minPath.reverse()
    return minCost, minPath

def computeFloydWarshall(N,S,T):
    '''Return the shortest path between any node in S and any node in T, by
    computing Floyd-Warshall.'''
    pred, dist = nx.floyd_warshall_predecessor_and_distance(N, 'upper')
    minCost,s,t = min((dist[v][w],v,w) for (v,w) in list(product(*[S,T])))
    # construct path
    w = t
    path = deque([w])
    while w != s:
        v = pred[s][w]
        assert N.has_edge(v,w)
        path.appendleft(v)
        w = v
    return minCost, list(path)

def computeShortestPathBidirectional(N,S,T):
    '''Return the shortest path between any node in S and any node in T, by
    computing bidirectional Dijkstra for all node pairs in S and T.'''
    minCost,minPath = float('inf'), []
    for (s,t) in list(product(*[S,T])):
        cost, path = nx.bidirectional_dijkstra(N, s, t)
        if minCost > cost:
            minCost, minPath = cost, path
    return minCost, minPath
    
def changeSolutionAlongPath(N, x, path, value):
    '''Change x along path by value; path might use backward arcs.''' 
    for a in netUtil.arcsOnPath_iter(path):
        (v,w) = a
        if N.has_edge(v,w):
            x[v][w] += value
        elif N.has_edge(w, v):
            x[w][v] -= value
        else:
            raise nx.NetworkXNoPath("No arc between {} and {}".format(v, w))

def emptyMap(N):
    '''Return a dictionary with every node as a key and None as value.'''
    eMap = {}
    for v in N:
        eMap[v] = None
    return eMap

def sizeOfCover(C):
    '''Return the number of bounds in C.'''
    return sum(1 for v in C for w in C[v] if C[v][w] != None)




        
