# -*- coding: utf-8 -*-
'''
Functions to ensure data validity.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
from itertools import ifilter, ifilterfalse
import warnings
from netmic.exceptions import DataError

def convertToSimpleGraph(N):
    '''Return a simple digraph, with loops removed and parallel edges combined.
    '''
    removeLoops(N)
    if not isMultiGraph(N):
        return N

    G = nx.DiGraph()
    for v,w,data in N.edges_iter(data=True):
        u = data["upper"]
        l = 0
        if "lower" in data:
            l = data["lower"]
        if G.has_edge(v,w):
            G[v][w]["upper"] += u
            G[v][w]["lower"] += l           
        else:
            G.add_edge(v, w, upper=u, lower=l)
    return G


def areValuesInt(N):
    '''Return True if all given data are integers and $|b| < \inf$.'''
    for v in N.nodes_iter(data=True):
        if abs(v[1]['supply']) >= float('inf'):
            raise DataError("Absolute value 'inf' not allowed for node {}.".
                            format(v))
        if int(v[1]["supply"]) != v[1]["supply"]:
            #warnings.warn("Supply of node {} is not integral.".format(v))
            return False
    for a in N.edges_iter(data=True):
        if (abs(a[2]["upper"]) < float('inf') and
            int(a[2]["upper"]) != a[2]["upper"]):
            #warnings.warn("Upper bound of arc {} is not integral.".format(a))
            return False
        if ('lower' in a[2] and abs(a[2]['lower']) < float('inf') and
            int(a[2]['lower']) != a[2]['lower']):
            #warnings.warn("Lower bound of arc {} is not integral.".format(a))
            return False
    return True

        
def addDefaultValues(N):
    '''If not existent, set node data "supply" = 0, arc data "upper" = 0.'''
    for v in ifilterfalse(lambda v: "supply" in v[1], N.nodes(data=True)):
        v[1]["supply"] = 0
    for v,w,data in ifilterfalse(lambda a: "upper" in a[2],
                                 N.edges_iter(data=True)):
        data["upper"] = 0


def removeLoops(N):
    '''Remove every loop from N.'''
    N.remove_edges_from([(v,v) for v in N.nodes() if N.has_edge(v,v)])

def isMultiGraph(N):
    '''Return True if N has parallel edges.'''
    for v,w in ifilter(lambda a: N.number_of_edges(*a)>1, N.edges_iter()):
        return True
    return False
    

class graphmlConsistency():
    '''wraps up the graphml data checks'''
    def __init__(self, oldN):
        self.oldN = oldN
        self.s = ''
        self.t = ''
        self.checkConsistency()
        self.intValues = areValuesInt(self.newN)

        
    def checkConsistency(self):
        '''Provide simple graph (DiGraph) newN  with default values 0 set.'''
        if not self.oldN.is_directed():
            raise DataError("Graph is not directed, can't work like that.")
        if not nx.is_weakly_connected(self.oldN):
            raise DataError("Graph is not connected, can't work like that.")
        
        addDefaultValues(self.oldN)
        self.newN = convertToSimpleGraph(self.oldN)
        self.nameConvention()
        
    def nameConvention(self):
        '''Replace node names s or t with a suitable integer.'''
        if "s" in self.newN:
            i = len(self.newN)
            while i in self.newN:
                i += 1 
            warnings.warn("Node name s reserved for single source model, renaming to {}"
                          .format(i))
            nx.relabel_nodes(self.newN, {"s": i}, copy=False)
            self.s = i
        if "t" in self.newN:
            i = len(self.newN)
            while i in self.newN:
                i += 1 
            warnings.warn("Node name t reserved for single source model, renaming to {}"
                          .format(i))
            nx.relabel_nodes(self.newN, {"t": i}, copy=False)        
            self.t = i

    def reverseNameConvention(self, C):
        '''If node names "s" or "t" were used, set them back in N and, if
        necessary, in C.'''
        if self.s:
            nx.relabel_nodes(self.newN, {self.s: 's'}, copy=False)
            try: C['s'] = C.pop(self.s)
            except KeyError: pass
            for v in ifilter(lambda v: self.s in C[v], C):
                C[v]['s'] = C[v].pop(self.s)
        if self.t:
            nx.relabel_nodes(self.newN, {self.t: 't'}, copy=False)
            try: C['t'] = C.pop(self.t)
            except KeyError: pass
            for v in ifilter(lambda v: self.t in C[v], C):
                C[v]['t'] = C[v].pop(self.t)

