# -*- coding: utf-8 -*-
'''
IIS Cover Heuristics.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
from networkx.algorithms.flow import preflow_push
from networkx.algorithms.flow import build_residual_network
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil
import netmic.timer as timer
import time

def fixedCover(N, nonredundant=True, timelimit=900):
    '''Return cover obtained by repeated searching for IISs and covering the
    smallest bound.'''
    builder = IISFixedCover(N)
    builder.solve(nonredundant, timelimit)
    if builder.timeOut:
        return (None, None, builder.it, round(builder.timeCover,3),
                round(builder.timeIIS,3))
    C,x = builder.coverFromSolution()
    return C, x, builder.it, round(builder.timeCover,3), round(builder.timeIIS,3)
    
def setCover(N, nonredundant=True, timelimit=900):
    '''Return cover obtained by repeated searching for IISs and solving greedily
    the set cover problem.'''
    builder = IISSetCover(N)
    builder.solve(nonredundant, timelimit)
    if builder.timeOut:
        return (None, None, builder.it, round(builder.timeCover,3),
                round(builder.timeIIS,3))
    C,x = builder.coverFromSolution()
    return C,x, builder.it, round(builder.timeCover,3), round(builder.timeIIS,3)

def hittingSet(N, nonredundant=True, timelimit=900):
    '''Return cover obtained by repeated searching for IISs and solving the
    hitting set problem.'''
    builder = IISHittingSet(N)
    builder.solve(nonredundant, timelimit)
    if builder.timeOut:
        return (None, None, builder.it, round(builder.timeCover,3),
                round(builder.timeIIS,3))
    C,x = builder.coverFromSolution()
    return C, x, builder.it, round(builder.timeCover,3), round(builder.timeIIS,3)

class IISCover():
    '''provides functionality to compute IISs via max-flow and cover them'''
    def __init__(self, N):
        self.N = N
        self.bPlus = netUtil.sumBPlus(self.N)
        self.ssN = netUtil.setUpSingleSourceSinkNetwork(self.N, False)
        netUtil.removeInfiniteUpperBoundsForMaxFlow(self.ssN)
        self._R = build_residual_network(self.ssN, 'upper')
        self.listOfS = []
        self.it = 0
        self.timeOut = False
        self.timeIIS = 0
        self.timeCover = 0

    def findIIS(self, nonredundant=True):
        '''Update (list of) connected set S that violates GH-inequality.'''
        self._R = preflow_push(self.ssN, 's', 't', 'upper', self._R)
        self.it += 1
        # network is feasible
        if abs(self._R.graph['flow_value'] - self.bPlus) < 1e-4:
            self.listOfS = []  
            return
        # check if S is connected
        S = self._minCutFromResidual()
        S.remove('s')
        componentsS = list(nx.weakly_connected_components(self.ssN.subgraph(S)))
        # S is disconnected (or we don't care)
        if not nonredundant or len(componentsS) > 1:  # T is inevitably connected
            self.listOfS = [comp for comp in componentsS if
                            netUtil.GHViolated(self.N, comp)]
        # S is connected, check if T is connected
        else:
            T = set(self.ssN) - S - set('s') - set('t')
            componentsT = list(nx.weakly_connected_components(self.ssN.
                                                              subgraph(T)))
            if len(componentsT) == 1: # S and T are connected
                self.listOfS = componentsS
            else:  # S is connected, T is disconnected -> use components of T
                self.listOfS = [comp for comp in componentsT if
                                netUtil.GHViolated(self.N, comp)]
        assert len(self.listOfS) > 0

    def coverFromSolution(self):
        '''Return cover according to final, feasible solution.'''
        flowVal, q = nx.maximum_flow(self.ssN, 's', 't', 'upper', preflow_push,
                                     residual=self._R)
        assert round(flowVal - self.bPlus, 4) == 0
        alUtil.solutionFromResidualSolution(self.N, q) 
        return alUtil.coverFromSolution(self.N, q), q

    def _minCutFromResidual(self):
        '''Return set S corresponding to the min-cut given by residual _R.'''
        ### code borrowed from nx.min_cut
        cutset = [(u,v,d) for u, v, d in self._R.edges(data=True)
                  if round(d['flow'] - d['capacity'], 6) == 0]
        self._R.remove_edges_from(cutset)
        # cut is defined by nodes that can reach 't'
        non_reachable = set(nx.shortest_path_length(self._R, target='t'))
        S = set(self.ssN) - non_reachable
        # Finaly add again cutset edges to the residual network to make
        # sure that it is reusable.
        if cutset is not None:
            self._R.add_edges_from(cutset)
        ###
        return S
            
    def _setBoundsToInf(self, arcs):
        '''For (v,w) in arcs, set bounds on (v,w) and (w,v) to inf in ssN and _R.'''
        for a in arcs:
            v,w = a
            # in ssN -> remove upper bound, add reverse arc
            del self.ssN[v][w]['upper']
            self.ssN.add_edge(w,v)
            # in R -> set bounds for both arcs to "inf" =  R.graph['inf'],
            # arcs might be non-present if bound was zero (preprocessing)
            if self.N[v][w]['upper'] == 0:
                self._R.add_edge(v,w, flow=0, capacity=self._R.graph['inf'])
                self._R.add_edge(w,v, flow=0, capacity=self._R.graph['inf'])
            else:
                self._R[v][w]['capacity'] = self._R.graph['inf']
                self._R[w][v]['capacity'] = self._R.graph['inf']

    def _resetBounds(self, arcs):
        '''For (v,w) in arcs, set bounds on (v,w) and (w,v) back to 'upper' in ssN and _R.'''
        for v,w in arcs:
            # in ssN: set upper bound, remove reverse arc
            self.ssN[v][w]['upper'] = self.N[v][w]['upper']
            self.ssN.remove_edge(w,v)
            # in R: set bound to u or 0, respectively
            self._R[v][w]['capacity'] = self.N[v][w]['upper']
            self._R[w][v]['capacity'] = 0

    
class IISFixedCover(IISCover):
    '''derived from IISCover, covers IISs by chosing the lowest bounded arc for
    every IIS'''
    def solve(self, nonredundant, timelimit=900):
        '''Compute an IIS in current network, set some bounds to inf and repeat
        until there are no more IISs.'''
        starttime = time.clock()
        with timer.TimerCPU() as t:
            self.findIIS(nonredundant)
        self.timeIIS += t.secs
        # as long as there are uncovered IISs
        while len(self.listOfS) > 0 and not self.timeOut:
            with timer.TimerCPU() as t:
                self._coverIISs()
            self.timeCover += t.secs
            with timer.TimerCPU() as t:
                self.findIIS(nonredundant)
            self.timeIIS += t.secs
            if time.clock() - starttime > timelimit:
                self.timeOut = True
        
    def _coverIISs(self):
        '''For every current IIS, set the smallest bound on delta to inf.'''
        arcs = [min((self.N.edge[a[0]][a[1]]['upper'],a) for a in
                    netUtil.delta(self.N, S))[1] for S in self.listOfS]
        self._setBoundsToInf(arcs)

        
class IISSetCover(IISCover):
    '''derived from IISCover, covers IISs by a greedy solution to the set cover
    problem in every step'''
    def __init__(self, N):
        IISCover.__init__(self, N)
        # data structure of IISList:
        # [{'covered': \# a\in \delta(S) \cap a\in C, 'delta': [a\in \delta(S)]}]
        self.IISList = []
        # data structure of arcDict:
        # { a: [index of S in IISList if a\in \delta(S)] }
        self.arcDict = dict()
        for a in self.N.edges_iter():
            self.arcDict[a] = []
        
    def solve(self, nonredundant, timelimit=900):
        '''Compute an IIS in current network, heuristically solve the set cover
        problem and repeat until there are no more IISs.'''
        starttime = time.clock()
        with timer.TimerCPU() as t:
            self.findIIS(nonredundant)
        self.timeIIS += t.secs
        arcs = []
        # as long as there are uncovered IISs
        while len(self.listOfS) > 0 and not self.timeOut:
            self._resetBounds(arcs)
            self._updateIISs()
            with timer.TimerCPU() as t:
                arcs = self._setCoverHeuristic()
            self.timeCover += t.secs
            self._setBoundsToInf(arcs)
            with timer.TimerCPU() as t:
                self.findIIS(nonredundant)
            self.timeIIS += t.secs
            if time.clock() - starttime > timelimit:
                self.timeOut = True
           
    def _updateIISs(self):
        '''Add data from listOfS to IISList and arcDict, set 'covered' = 0
        for all IISs.'''
        for d in self.IISList:
            d['covered'] = 0
        for S in self.listOfS:
            delta = netUtil.delta(self.N, S)
            self.IISList.append({'delta': delta, 'covered': 0})
            for a in delta:
                self.arcDict[a].append(len(self.IISList)-1)
            
    def _setCoverHeuristic(self):
        '''Return list of arcs corresponding to the set cover solution.
        Solve set cover problem greedily.'''
        arcs = []
        uncovered = range(len(self.IISList))
        while uncovered:
            # a covers largest number of previously uncovered IISs
            a = max( (sum(1 for i in self.arcDict[a]
                          if self.IISList[i]['covered'] == 0), a)
                     for a in self.N.edges_iter())[1]
            arcs.append(a)
            # update number of arcs that cover an IIS
            for i in self.arcDict[a]:
                self.IISList[i]['covered'] += 1
                try:
                    uncovered.remove(i)
                except ValueError: pass
        # remove redundant arcs
        for a in arcs:
            if min(self.IISList[i]['covered'] for i in self.arcDict[a]) > 1:
                arcs.remove(a)
                for i in self.arcDict[a]:
                    self.IISList[i]['covered'] -= 1 
        return arcs


class IISHittingSet(IISCover):
    '''derived from IISCover, covers IISs by chosing the arc that covers the most
    IISs'''
    def __init__(self, N):
        IISCover.__init__(self, N)
        # data structure of IISList: [[a\in \delta(S)]]
        self.IISList = []
        # data structure of arcDict: { a: [index of S in IISList if a\in \delta(S)] }
        self.arcDict = dict()
        for a in self.N.edges_iter():
            self.arcDict[a] = []
            
    def solve(self, nonredundant, timelimit=900):
        '''Compute an IIS in current network, use a hitting set heuristic and 
        repeat until there are no more IISs.'''
        starttime = time.clock()
        with timer.TimerCPU() as t:
            self.findIIS(nonredundant)
        self.timeIIS += t.secs
        arcs = []
        # as long as there are uncovered IISs
        while len(self.listOfS) > 0 and not self.timeOut:  
            self._resetBounds(arcs)
            self._updateIISs()
            with timer.TimerCPU() as t:
                arcs = self._hittingSetHeuristic()
            self.timeCover += t.secs
            self._setBoundsToInf(arcs)
            with timer.TimerCPU() as t:
                self.findIIS(nonredundant)
            self.timeIIS += t.secs
            if time.clock() - starttime > timelimit:
                self.timeOut = True

    def _updateIISs(self):
        '''Add data from listOfS to IISList and arcDict.'''
        for S in self.listOfS:
            delta = netUtil.delta(self.N, S)
            self.IISList.append(delta)
            for a in delta:
                self.arcDict[a].append(len(self.IISList)-1)
                
    def _hittingSetHeuristic(self):
        '''Return list of arcs corresponding to hitting set solution.
        Heuristic: For every IIS, not covered before, choose arc the covers
        the most IISs overall.'''
        arcs = []
        uncovered = range(len(self.IISList))
        while uncovered:
            a = max((len(self.arcDict[a]),a) for a in
                    self.IISList[uncovered[0]])[1]
            arcs.append(a)
            for i in self.arcDict[a]:
                try:
                    uncovered.remove(i)
                except ValueError: pass
        return arcs
    
