# -*- coding: utf-8 -*-
'''
Elementary Cycle Flows Heuristic.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
import time
from networkx.algorithms.flow import preflow_push
from networkx.algorithms.flow import build_residual_network
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil


def enumerateElementaryCycles(N, timelimit=900):
    '''Return cover obtained by elementary cycle heuristic.'''
    builder = elementaryCycleSolver(N)
    builder.solve(timelimit)
    return alUtil.coverFromSolution(N, builder.x), builder.x, builder.it


class elementaryCycleSolver():
    '''Implementation of elementary cycle heuristic.'''
    def __init__(self, N):
        self.N = N
        self.arcPairN = netUtil.arcPairNetwork(N)
        self.x = self.computeStartSolution()
        self.it = 0
        self.timeOut = False

    def solve(self, timelimit=900):
        '''Compute elementary cycle heuristic.
        For all elementary cycles, check if a cycle flow, with value such that
        one bound is met, improves the solution.'''
        starttime = time.clock()
        for cycle in nx.simple_cycles(self.arcPairN):
            # timelimit reached, stop
            if time.clock() - starttime > timelimit:
                break
            # cycle is an arc pair, discard
            if len(cycle) < 3:  
                continue
            self.it += 1
            # cycles are actually given as path, make it a cycle
            cycle.append(cycle[0])
            for a in netUtil.arcsOnPath_iter(cycle):
                v,w = netUtil.networkArcFromEdge(self.N, a)
                if self.x[v][w] > self.N.edge[v][w]['upper']:
                    r = self.N.edge[v][w]['upper'] - self.x[v][w]
                elif self.x[v][w] < 0:
                    r = -self.x[v][w]
                else: # "r = 0", bland
                    continue
                r = r if self.N.has_edge(v,w) else -r
                # if augmentation with value r would improve solution, do it
                if self.checkChangeToBound(r, cycle) > 0:
                    alUtil.changeSolutionAlongPath(self.N, self.x, cycle, r)

    def checkChangeToBound(self, r, cycle):
        '''Return (negative) change in objective function after augmentation
        along cycle with value r.'''
        noImproved = 0
        for y,z in netUtil.arcsOnPath_iter(cycle):
            if self.N.has_edge(y,z):
                noImproved += self.improved(y,z,r)
            else:
                assert self.N.has_edge(z,y)
                noImproved += self.improved(z,y,-r)
        return noImproved

        
    def improved(self, v,w,r):
        '''Return 1 if x+r would improve solution for a=(v,w), -1 if impaired,
        0 otherwise.'''
        if (self.x[v][w] > self.N.edge[v][w]['upper'] and self.x[v][w]+r <=
            self.N.edge[v][w]['upper'] and self.x[v][w]+r >= 0):
            return 1
        if (self.x[v][w] < 0 and self.x[v][w]+r >= 0
            and self.x[v][w]+r <= self.N.edge[v][w]['upper']):
            return 1
        if (self.x[v][w] <= self.N.edge[v][w]['upper'] and self.x[v][w] >= 0 and
            (self.x[v][w]+r < 0 or self.x[v][w]+r > self.N.edge[v][w]['upper'])):
            return -1
        return 0
                
    def computeStartSolution(self):
        '''Return b-feasible solution, computed with all bounds = +/- inf.'''
        ssN = netUtil.setUpInfiniteBoundSingleSourceSinkNetwork(self.N)
        R = build_residual_network(ssN, 'capacity')
        for v,w in self.N.edges_iter():
            R.edge[w][v]['capacity'] = R.graph['inf']
        val, x = nx.maximum_flow(ssN, 's', 't', 'capacity', preflow_push,
                                 residual=R)
        assert abs(val - netUtil.sumBPlus(self.N)) < 1e-4
        alUtil.solutionFromResidualSolution(self.N, x)
        return x
        
        
