# -*- coding: utf-8 -*-
'''
l_1-slack model (heuristic).
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
import pulp
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil
from netmic.exceptions import ModelUnsolvedError
        
def solveEll1(N, timelimit=900):
    '''Return cover computed via ell_1-norm model.'''
    modelBuilder = ell1Model(N)
    return modelBuilder.buildCover(timelimit)

    
class ell1Model():
    '''set up and solve ell_1-norm model of MIAC'''
    def __init__(self, N, write=False):
        self.N = N
        self.lp = pulp.LpProblem("Ell1_MIAC", pulp.LpMinimize)
        self.buildVariables()
        self.buildModel(write)

    def buildVariables(self):
        '''Set up variables and bounds.'''
        self.x = pulp.LpVariable.dicts('x', self.N.edges())
        self.sl = pulp.LpVariable.dicts('s_l', self.N.edges(), 0)
        self.su = pulp.LpVariable.dicts('s_u', self.N.edges(), 0)

    def buildModel(self, write):
        '''Add objective function and constraints to model.'''
        self.lp += (pulp.lpSum(self.sl) + pulp.lpSum(self.su),
                     'minimize slack')
        # add constraints Ax = b
        for v in self.N:
            APlus = pulp.LpAffineExpression([(self.x[a],1) for a in
                                             self.N.out_edges_iter(v)])
            AMinus = pulp.LpAffineExpression([(self.x[a],-1) for a in
                                              self.N.in_edges_iter(v)])
            self.lp += APlus + AMinus == self.N.node[v]['supply'], ''
        # add constraints relaxed bounds
        for a in self.N.edges_iter():
            self.lp += self.x[a] + self.sl[a] >= 0, ''
            self.lp += (self.x[a] - self.su[a] <=
                        self.N.edge[a[0]][a[1]]['upper'], '')
        if write:
            self.lp.writeLP('Ell1.lp')
            
    def buildCover(self, timelimit=900):
        '''Solve model, build and return cover and flow according to solution.'''
        try:
            self.lp.solve(pulp.CPLEX(msg=0,options=['set timeLimit {}'.
                                                    format(timelimit)]))
        except pulp.solvers.PulpSolverError as e:
            if str(e) == "Unknown status returned by CPLEX: time limit exceeded":
                return None, None
            else: raise
            
        if self.lp.status < 1:
            raise ModelUnsolvedError("Model could not be solved, status: {}.".
                                     format(pulp.constants.LpStatus[self.lp.status]))

        # build cover and flow from solution
        C = alUtil.emptySolutionDict(self.N)
        for a in self.N.edges_iter():
            if round(self.sl[a].varValue, 6) > 0:
                C[a[0]][a[1]] = 'l'
            if round(self.su[a].varValue, 6) > 0:
                C[a[0]][a[1]] = 'u'

        y = alUtil.emptySolutionDict(self.N)
        for a in self.N.edges_iter():
            y[a[0]][a[1]] = round(self.x[a].varValue,6)
                
        return C,y
            
        
