# -*- coding: utf-8 -*-
'''
Flow Heuristics 1 and 2.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.


import networkx as nx
from itertools import product, count, ifilter, ifilterfalse
from collections import deque
from heapq import heappush, heappop, heapify
import weight_functions as weightFunc
import netmic.timer as timer
import time

import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil

def heuristic1_1(N, sp='dijkstra', timelimit=900):
    '''Return cover obtained by heuristic 1.1.'''
    builder = heuristic1_1Solver(N)
    builder.solve(sp)
    if builder.timeOut:
        return None, None, builder.it
    C = alUtil.coverFromSolution(N, builder.x)
    return C, builder.x, builder.it 

def heuristic1_2(N, timelimit=900):
    '''Return cover obtained by heuristic 1.2.'''
    builder = heuristic1_2Solver(N)
    builder.solve(timelimit)
    if builder.timeOut:
        return None, None, builder.it
    C = alUtil.coverFromSolution(N, builder.x)
    return C, builder.x, builder.it

def heuristic1_3(N, variant=False, timelimit=900):
    '''Return cover obtained by heuristic 1.3.'''
    builder = heuristic1_3Solver(N, variant)
    builder.solve(timelimit)
    if builder.timeOut:
        return None, None, builder.it
    C = alUtil.coverFromSolution(N, builder.x)
    return C, builder.x, builder.it

def heuristic2_2(N, timelimit=900):
    '''Return cover obtained by heuristic 2.2.'''
    builder = heuristic2_2Solver(N, timelimit)
    builder.solve()
    if builder.timeOut:
        return None, None, builder.it, round(builder.timeStart,2) 
    C = alUtil.coverFromSolution(N, builder.x)
    return C, builder.x, builder.it, round(builder.timeStart,2)

def heuristic2_3(N, variant=False, timelimit=900):
    '''Return cover obtained by heuristic 2.3.'''
    builder = heuristic2_3Solver(N, variant, timelimit)
    builder.solve()
    if builder.timeOut:
        return None, None, builder.it, round(builder.timeStart,2)
    C = alUtil.coverFromSolution(N, builder.x)
    return C, builder.x, builder.it, round(builder.timeStart,2)

class heuristicSolver():
    '''base class for heuristic, provide functionality shared by all heuristics'''
    def __init__(self, N):
        self.N = N
        self.arcPairN = netUtil.arcPairNetwork(N)
        self.it = 0
        self.timeOut = False

    def solve(self):
        raise NotImplementedError('heuristicSolver is a base class, use 1_1, 1_2,'
                                  ' 1_3, 2_2 or 2_3 instead') 
 

class heuristic1Solver(heuristicSolver):
    '''base class for heuristic1, start with l1-opt INC, compute iteratively
    shortest paths between nodes with excess'''
    def __init__(self, N):
        heuristicSolver.__init__(self, N)
        self.x = self.computeStartSolution()
        self._S, self._T = alUtil.unsaturatedSets(self.N, self.x)

    def solve(self):
        raise NotImplementedError('heuristic1Solver is a base class, use 1_1, '
                                  '1_2 or 1_3 instead') 

    def computeStartSolution(self):
        '''Return an initial solution for N, feasible regarding bounds,
        but not flow conservation, by computing the max-flow.'''
        if not self.N.has_node("s") and not self.N.has_node("t"):
            auxN = netUtil.setUpSingleSourceSinkNetwork(self.N, False)
        elif self.N.has_node("s") and self.N.has_node("t"):
            auxN = self.N.copy()
        else:
            raise ValueError("Node s or t in network, but not both")
        value, x = nx.maximum_flow(auxN, 's', 't', 'upper')
        alUtil.solutionFromResidualSolution(self.N,x)
        return x

    def _changeSolutionAndUpdateSandT(self, path):
        '''Update x and remove thereby saturated nodes from S and/or T.
        x is changed along path by the minimum of d_s and d_t, where s and t are
        determined by the start and end node of path. s must be in S, and t in T.
        '''
        s = path[0]
        t = path[len(path)-1]
        assert s in self._S and t in self._T
        ds =  alUtil.flowExcessAtNode(self.N, self.x, s)
        dt = -alUtil.flowExcessAtNode(self.N, self.x, t)
        if round(dt - ds, 6) >= 0:
            alUtil.changeSolutionAlongPath(self.N, self.x, path, ds)
            self._S.remove(s)
        else:
            alUtil.changeSolutionAlongPath(self.N, self.x, path, dt)
            self._T.remove(t)
        if round(dt - ds, 6) == 0:
            self._T.remove(t)

        
class heuristic2Solver(heuristicSolver):
    '''base class for heuristic2, start with heuristic ell_0 solution, compute
    iteratively shortest paths from node with largest excess'''    
    def __init__(self, N, timelimit=900):
        heuristicSolver.__init__(self, N)
        with timer.TimerCPU() as t:
            startSolutionBuilder = heuristicEll0StartSolution(self.N)
            self.x = startSolutionBuilder.solve(timelimit/3)
        self.timeStart = t.secs
        self.timelimit = timelimit - t.secs
        # S will be a min-heap with negative! excess as key 
        self._S, self._T = [], []
        self.excess = dict()
        self._initializeExcessData()
     
    def solve(self):
        raise NotImplementedError('heuristic2Solver is a base class, use 2_2 or '
                                  '2_3 instead')

    def _initializeExcessData(self):
        '''Create dict with value of the excesses of a node, provide lists S and
        T of nodes with positive and negative excess, respectively, where S is
        a heap sorted from lowest to largest excess.'''
        for v in self.N:
            dv = alUtil.flowExcessAtNode(self.N, self.x, v)
            if round(dv, 6) == 0:
                continue
            self.excess[v] = dv
            if round(dv, 6) > 0:
                heappush(self._S, (-self.excess[v], v))
            elif round(dv, 6) < 0:
                self._T.append(v)

    def _changeSolutionAndUpdateSandT(self, path):
        '''Update x and possibly move t to S (maintaining order).
        x is changed along path by d_s, where s and t are determined by the start
        and end node of path. t must be in T.
        '''
        s = path[0]
        t = path[len(path)-1]
        assert t in self._T
        # change solution
        alUtil.changeSolutionAlongPath(self.N, self.x, path, self.excess[s])
        # update T and excess
        self.excess[t] += self.excess[s]
        if round(self.excess[t], 6) >= 0:
            self._T.remove(t)
        if round(self.excess[t], 6) > 0:
            heappush(self._S, (-self.excess[t], t))
        del self.excess[s]
        
class heuristicx_3Solver(heuristicSolver):
    '''base class for heuristics 1.3 and 2.3, compute shortest path with
    forbidden pairs'''
    def __init__(self, variant):
        self.variant = variant

    def solve(self):
        raise NotImplementedError('heuristicx_3Solver is an abstract base class,'
                                  ' use 1_3 or 2_3 instead')
        
    def shortPathWithNegCycles(self, s, t=None):
        '''Return short path between s and t. This is essentially Dijkstra,
        adapted to tolerate negative weights and thereby no longer yielding the
        optimal solution.'''
        if not isinstance(self, (heuristic1_3Solver, heuristic2_3Solver)):
            raise NotImplementedError('heuristicx_3Solver is an abstract base '
                                      'class, use 1_3 or 2_3 instead') 
        ### heap usage copied from nx.single_source_dijkstra
        nodeWeights = {v: float('inf') for v in self.arcPairN.nodes_iter()}
        nodeWeights[s] = 0
        pred = {s: None}
        marked = {v: False for v in self.arcPairN.nodes_iter()}
        minHeap = []
        heappush(minHeap, (0, s))
        while minHeap:
            d,v = heappop(minHeap)
            if marked[v]:  # there can be multiple entries for v in heap
                continue
            marked[v] = True
            if v == t:
                break
            for w in ifilterfalse(lambda w: marked[w],
                                  self.arcPairN.successors_iter(v)):
                if nodeWeights[w] > d + self.arcPairN[v][w]['weight']:
                    nodeWeights[w] = d + self.arcPairN[v][w]['weight']
                    pred[w] = v
                    heappush(minHeap, (nodeWeights[w], w))
                elif (self.variant and
                      nodeWeights[w] == d + self.arcPairN[v][w]['weight']):
                    pred[w] = v  # no need to push on heap, must already be in
        ###
        if t != None:
            return self._BFSPredecessorToPath(s, t, pred)
        else:
            return self._BFSPredecessorToPaths(s, pred)

    def _BFSPredecessorToPath(self, s, t, pred):
        '''Return path from s to t according to "pred".'''
        w = t
        path = deque([w])
        while w != s:
            v = pred[w]
            assert self.arcPairN.has_edge(v,w)
            path.appendleft(v)
            w = v
        assert pred[w] == None
        return list(path)
        
    def _BFSPredecessorToPaths(self, s, pred):
        '''Return dict of paths from s to all v according to "pred".'''
        paths = {v:[] for v in self.arcPairN.nodes_iter()}
        paths[s] = [s]
        for t in ifilter(lambda v: paths[v]==[],
                              self.arcPairN.nodes_iter()):
            try:
                self._constructPath(t, pred, paths)
            except RuntimeError:
                self._constructPathIteratively(s,t,pred,paths)
        return paths

    def _constructPath(self, t, pred, paths):
        '''Recursively construct path from s to t according to pred.'''
        v = pred[t]
        assert self.arcPairN.has_edge(v,t)
        if paths[v] == []:
            self._constructPath(v, pred, paths)
        assert paths[v]
        paths[t] = paths[v] + [t]

    def _constructPathIteratively(self, s, t, pred, paths):
        '''Construct path from s to t according to pred, without storing
        component paths. Used when max recursion depth is exceeded.'''
        v = pred[t]
        assert self.arcPairN.has_edge(v,t)
        path = deque([v,t])
        while v != s:
            assert self.arcPairN.has_edge(pred[v],v)
            if paths[pred[v]] != []:
                paths[t] = paths[pred[v]] + list(path)
                return
            path.appendleft(pred[v])
            v = pred[v]
        assert pred[v] == None
        paths[t] = list(path)
                
        
class heuristic1_1Solver(heuristic1Solver):
    '''Implementation of heuristic 1.1'''
    def solve(self, sp='dijkstra', timelimit=900):
        '''Compute heuristic 1.1. Start flow is given by bound feasible
        max-flow, simple 0/1/2 weights for all nodes at once are used. The SP-
        computations are done using the method indicated by sp.
        '''
        starttime = time.clock()
        # set shortest path function
        if sp == 'dijkstra':
            computeSP = alUtil.computeShortestPath
        elif sp == 'floyd':
            computeSP = alUtil.computeFloydWarshall
        elif sp == 'combined':
            computeSP = alUtil.computeShortestPathFloyd
        else:
            raise NotImplementedError("The chosen method for shortest paths "
                                      "computations does not exist.")
            
        weightFunc.computeArcWeightsSimple(self.N, self.arcPairN, self.x,
                                           self.arcPairN.edges_iter())
        self.it = len(self._S)
        while len(self._S) > 0 and not self.timeOut:
            assert len(self._T) > 0
            _, path = computeSP(self.arcPairN, self._S, self._T)
            self._changeSolutionAndUpdateSandT(path)
            weightFunc.computeArcWeightsSimple(self.N, self.arcPairN, self.x,
                                               netUtil.arcsOnPath_iter(path))
            if time.clock() - starttime > timelimit:
                self.timeOut = True

        if not self.timeOut:
            assert alUtil.xBFlowFeasible(self.N, self.x)

        
class heuristic1_2Solver(heuristic1Solver):
    '''Implementation of heuristic 1.2'''
    def solve(self, timelimit=900):
        '''Compute heuristic 1.2. Start flow is given by bound feasible
        max-flow, 0/1-weights are computed "exact" for every pair (s,t) in (S,T).
        '''
        starttime = time.clock()
        while len(self._S) > 0 and not self.timeOut:
            assert len(self._T) > 0
            minCost = float('inf')
            minPath = []
            self.it += len(self._S) * len(self._T)
            for (s,t) in list(product(*[self._S,self._T])):
                r = min(alUtil.flowExcessAtNode(self.N, self.x, s),
                        -alUtil.flowExcessAtNode(self.N, self.x, t))
                weightFunc.computeArcWeightsDetailedBinary(self.N, self.arcPairN,
                                                           self.x, r)

                cost, path = nx.bidirectional_dijkstra(self.arcPairN, s, t)
                if cost < minCost:
                    minCost, minPath = cost, path
                if time.clock() - starttime > timelimit:
                    self.timeOut = True
                    break
            self._changeSolutionAndUpdateSandT(minPath)

        if not self.timeOut:
            assert alUtil.xBFlowFeasible(self.N, self.x)
        

class heuristic1_3Solver(heuristic1Solver, heuristicx_3Solver):
    '''Implementation of heuristic 1.3'''
    def __init__(self, N, variant=False):
        heuristic1Solver.__init__(self, N)
        heuristicx_3Solver.__init__(self, variant)
        
    def solve(self, timelimit=900):
        '''Compute heuristic 1.3. Start flow is given by bound feasible max-flow,
        -1/0/1-weights are computed "exact" for every pair (s,t) in (S,T),
        then the shortest path is computed heuristically.
        '''
        starttime = time.clock()
        while len(self._S) > 0 and not self.timeOut:
            assert len(self._T) > 0
            minCost = float('inf')
            minPath = []
            self.it += len(self._S) * len(self._T)
            for (s,t) in list(product(*[self._S,self._T])):
                r = min(alUtil.flowExcessAtNode(self.N, self.x, s),
                        -alUtil.flowExcessAtNode(self.N, self.x, t))
                weightFunc.computeArcWeightsDetailedTernary(self.N, self.arcPairN,
                                                            self.x, r)
                path = self.shortPathWithNegCycles(s, t)
                cost = alUtil.computePathWeight(self.arcPairN, path)
                if cost < minCost:
                    minCost, minPath = cost, path
                if time.clock() - starttime > timelimit:
                    self.timeOut = True
                    break
            assert len(minPath) > 0
            self._changeSolutionAndUpdateSandT(minPath)

        if not self.timeOut:
            assert alUtil.xBFlowFeasible(self.N, self.x)
 
        
class heuristic2_2Solver(heuristic2Solver):
    '''Implementation of heuristic 2.2'''
    def solve(self):
        '''Compute heuristic 2.2. Start flow is given by bound-feasible, heuristic
        ell_0 INC, shortest path is computed with binary weights for node with
        largest excess.'''
        starttime = time.clock()
        while len(self._S) > 0 and not self.timeOut:
            self.it += 1
            assert len(self._T) > 0
            # S is sorted by negative excess, ie ds is negative
            ds,s = heappop(self._S)
            weightFunc.computeArcWeightsDetailedBinary(self.N, self.arcPairN,
                                                       self.x, -ds)
            _, path = alUtil.computeShortestPath(self.arcPairN, [s], self._T)
            self._changeSolutionAndUpdateSandT(path)
            if time.clock() - starttime > self.timelimit:
                self.timeOut = True

        if not self.timeOut:
            assert alUtil.xBFlowFeasible(self.N, self.x)


class heuristic2_3Solver(heuristic2Solver, heuristicx_3Solver):
     '''Implementation of heuristic 2.3'''
     def __init__(self, N, variant=False, timelimit=900):
         heuristic2Solver.__init__(self, N, timelimit)
         heuristicx_3Solver.__init__(self, variant)
         
     def solve(self):
         '''Compute heuristic 2.3. Start flow is given by bound-feasible, heuristic
         ell_0 INC, shortest path is computed with ternary weights for node with
         largest excess, then shortest path is computed heuristically.'''
         starttime = time.clock()
         while len(self._S) > 0:
             self.it += 1
             assert len(self._T) > 0
             # S is sorted by negative excess, ie ds is negative
             ds,s = heappop(self._S)
             weightFunc.computeArcWeightsDetailedTernary(self.N, self.arcPairN,
                                                         self.x, -ds)
             paths = self.shortPathWithNegCycles(s)
             minT = min(((alUtil.computePathWeight(self.arcPairN, paths[t]), t)
                         for t in self._T))[1]
             self._changeSolutionAndUpdateSandT(paths[minT])
             if time.clock() - starttime > self.timelimit:
                 self.timeOut = True

         if not self.timeOut:
             assert alUtil.xBFlowFeasible(self.N, self.x)

         
class heuristicEll0StartSolution():
    '''compute INC solution by iteratively satisfying node with smallest excess'''
    def __init__(self, N):
        self.N = N
        self.ssN = netUtil.setUpSingleSourceSinkNetwork(self.N, False)
        self.ssN.graph['origDir'] = True
        self.supplyNodeList = {v: self.N.node[v]['supply'] for v in
                               netUtil.listOfSupplyNodes_iter(self.N)}
        self.demandNodeList = {v: -self.N.node[v]['supply'] for v in
                               netUtil.listOfDemandNodes_iter(self.N)}
        self.S = [(self.N.node[v]['supply'], v)
                  for v in netUtil.listOfSupplyNodes(self.N)]
        heapify(self.S)
        self.T = [(-self.N.node[v]['supply'], v)
                  for v in netUtil.listOfDemandNodes(self.N)]
        heapify(self.T)
        self.x = dict()
        alUtil.completeSolution(self.N, self.x)
        self.timeOut = False
        
    def solve(self, timelimit=300):
        '''Return bound-feasible solution with (heuristically) as much nodes
        satisfied as possible.'''
        starttime = time.clock()
        self._resetSupplyBounds(self.supplyNodeList.keys(),
                                self.demandNodeList.keys())
        while len(self.S) + len(self.T) > 0 and not self.timeOut:
            # sOrT=True if node from S should be satisfied 
            if len(self.S) > 0 and len(self.T) > 0: 
                sOrT = (True if self.S[0][0] < self.T[0][0] else False)
            else:
                sOrT = True if len(self.S) > 0 else False

            if sOrT:
                self._sendFlowFromS()
            else:
                self._sendFlowFromT()
            if time.clock() - starttime > timelimit:
                self.timeOut = True
        return self.x

    def _resetSupplyBounds(self, supplies, demands):
        '''Set all bounds between nodes in supplies and 's', and all bounds
        between nodes in demands and 't' to zero.'''
        if self.ssN.graph['origDir']:
            for v in supplies:
                self.ssN['s'][v]['upper'] = 0
            for v in demands:
                self.ssN[v]['t']['upper'] = 0
        else:
            for v in supplies:
                self.ssN[v]['s']['upper'] = 0
            for v in demands:
                self.ssN['t'][v]['upper'] = 0

    def _sendFlowFromS(self):
        '''Send flow from S[0] to T if whole amount can be sent. Remove S[0].'''
        # get smallest s
        supply,s = heappop(self.S)
        # Check current direction of network
        if not self.ssN.graph['origDir']:
            self.ssN.reverse(copy=False)
            self.ssN.graph['origDir'] = True
        # Set supply and demand bounds
        self.ssN['s'][s]['upper'] = supply
        for v in self.demandNodeList:
            self.ssN[v]['t']['upper'] = self.demandNodeList[v]
        # Compute max-flow
        value, auxX = nx.maximum_flow(self.ssN, 's', 't', 'upper')
        self._resetSupplyBounds([s], self.demandNodeList.keys())
        # Else failed to sent whole flow amount, just discard S[0]
        if round(supply - value, 4) == 0:
            # Update solution, excesses and bounds in network
            self._updateSolutionAndBounds(auxX)
            del self.supplyNodeList[s]
            self._updateDemandList(auxX)        

    def _sendFlowFromT(self):
        '''Send flow from T[0] to S if whole amount can be sent. Remove T[0].'''
        # get smallest t
        demand,t = heappop(self.T)
        # Check current direction of network
        if self.ssN.graph['origDir']:
            self.ssN.reverse(copy=False)
            self.ssN.graph['origDir'] = False
        # Set supply and demand bounds     
        self.ssN['t'][t]['upper'] = demand
        for v in self.supplyNodeList:
            self.ssN[v]['s']['upper'] = self.supplyNodeList[v]
        # Compute max-flow    
        value, auxX = nx.maximum_flow(self.ssN, 't', 's', 'upper')
        self._resetSupplyBounds(self.supplyNodeList.keys(), [t])
        # Else failed to sent whole flow amount, just discard T[0]
        if round(demand - value, 4) == 0:
            # Update solution, excesses and bounds in network    
            self._updateSolutionAndBounds(auxX)
            del self.demandNodeList[t]
            self._updateSupplyList(auxX)

    def _updateSolutionAndBounds(self, auxX):
        '''Add auxX to x, substract auxX from upper bounds.'''
        if self.ssN.graph['origDir']:
            for v,w in self.N.edges_iter():
                value = auxX[v][w]
                self.x[v][w] += value
                self.ssN[v][w]['upper'] -= value
        else:
            for v,w in self.N.edges_iter():
                value = auxX[w][v]
                self.x[v][w] += value
                self.ssN[w][v]['upper'] -= value

    def _updateDemandList(self, auxX):
        '''Update values in demandNodeList and T.'''
        for v,oldDemand in self.demandNodeList.items():
            if round(auxX[v]['t'], 6) == 0:
                continue
            # update value in demandNodeList
            self.demandNodeList[v] -= auxX[v]['t']
            # if node is satisfied, remove it
            if round(self.demandNodeList[v], 6) == 0:
                del self.demandNodeList[v]
                self.T.remove((oldDemand,v))
            # update values in T
            else:
                try:  # v might already be deleted 
                    self.T.remove((oldDemand,v))
                except ValueError: pass
                else:
                    heappush(self.T, (self.demandNodeList[v],v))

    def _updateSupplyList(self, auxX):
        '''Update values in supplyNodeList and S.'''
        for v,oldSupply in self.supplyNodeList.items():
            if round(auxX[v]['s'], 6) == 0:
                continue
            # update value in supplyNodeList    
            self.supplyNodeList[v] -= auxX[v]['s']
            # if node is satisfied, remove it
            if round(self.supplyNodeList[v], 6) == 0:
                del self.supplyNodeList[v]
                self.S.remove((oldSupply,v))
            # update values in S
            else:
                try:  # v might already be deleted
                    self.S.remove((oldSupply,v))
                except ValueError: pass
                else:
                    heappush(self.S, (self.supplyNodeList[v],v))
