# -*- coding: utf-8 -*-
'''
Greedy Addition Heuristic.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
from networkx.algorithms.flow import preflow_push
from networkx.algorithms.flow import build_residual_network
from itertools import ifilter
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil
import time

def maxFSHeuristic(N, blockSize=3, timelimit=900):
    '''Return cover computed via maxFS heuristic;
    start with no bounds, greedily add them.'''
    builder = maxFSSolver(N)
    builder.solve(blockSize, timelimit)
    if builder.timeOut:
        return None, None, builder.it
    return builder.C, builder.x, builder.it


class maxFSSolver():
    '''maxFS-style heuristic (greedily add bounds)'''
    def __init__(self, N):
        self.N = N
        self.ssN = netUtil.setUpSingleSourceSinkNetwork(N, False)
        netUtil.removeInfiniteUpperBoundsForMaxFlow(self.ssN)
        self._R = build_residual_network(self.ssN, 'upper')
        self._addZeroBoundsToR()
        self.bPlus = netUtil.sumBPlus(N)
        self.cleanUp()

    def cleanUp(self):
        '''Set list of arcs to be checked to all arcs, empty cover, remove bounds
        from network, it=0.'''
        self._checkLower = self.N.edges()
        self._checkUpper = self.N.edges()
        self.C = alUtil.emptySolutionDict(self.N)
        self._removeBoundsFromR(self.N.edges())
        self.it = 0
        self.timeOut = False

    def solve(self, blockSize, timelimit=900):
        '''Check greedily (in blocks of size blockSize), which bounds can be
        added to network while maintaining feasibility. All others build cover.'''
        starttime = time.clock()
        while len(self._checkLower) > 0 and not self.timeOut:
            subset = self._checkLower[-blockSize:]
            self._testLowerBounds(subset)
            if time.clock() - starttime > timelimit:
                self.timeOut = True
        while len(self._checkUpper) > 0 and not self.timeOut:
            subset = self._checkUpper[-blockSize:]
            self._testUpperBounds(subset)
            if time.clock() - starttime > timelimit:
                self.timeOut = True

        self._R = preflow_push(self.ssN, 's', 't', 'upper', self._R)
        self.x = alUtil.solutionFromResidualGraph(self.N, self._R)
        

    def _testLowerBounds(self, subset):
        '''Check if all lower bounds in subset can be added without loosing
        feasibility.'''
        self._addLowerBoundsToR(subset)
        self._R = preflow_push(self.ssN, 's', 't', 'upper', self._R,
                               value_only=True)
        self.it += 1
        # feasible
        if round(self._R.graph['flow_value'] - self.bPlus, 4) == 0:
            for a in subset:
                self._checkLower.remove(a)
            self._checkForAdditionallySatisfiedBounds()
        # infeasible
        else:
            self._removeLowerBoundsFromR(subset)
            self._testLowerBoundsIndividually(subset)
                
    def _testLowerBoundsIndividually(self, subset):
        '''Check if the lower bounds in subset can be added individually without
        loosing feasibility.'''
        nrFeasible = 0
        for a in subset:
            # element might be removed in the mean time
            try:
                i = self._checkLower.index(a)
            except ValueError:
                nrFeasible += 1
                continue
            # if all other bounds in the subset were feasible, l_a is infeasible
            if nrFeasible == len(subset)-1:
                del self._checkLower[i]
                assert (a[1] not in self.C[a[0]] or
                        self.C[a[0]][a[1]] == None)
                self.C[a[0]][a[1]] = 'l'
                continue
                
            # add bound and solve max-flow
            self._addLowerBoundsToR([a])
            self._R = preflow_push(self.ssN, 's', 't', 'upper', self._R,
                                   value_only=True)
            self.it += 1
            # feasible
            if round(self._R.graph['flow_value'] - self.bPlus, 4) == 0:
                del self._checkLower[i]
                self._checkForAdditionallySatisfiedBounds()
                nrFeasible += 1
            # infeasible, bound belongs in cover
            else:
                del self._checkLower[i]
                self._removeLowerBoundsFromR([a])
                assert (a[1] not in self.C[a[0]] or
                        self.C[a[0]][a[1]] == None)
                self.C[a[0]][a[1]] = 'l'

                
    def _testUpperBounds(self, subset):
        '''Check if all upper bounds in subset can be added without loosing
        feasibility.'''
        self._addUpperBoundsToR(subset)
        self._R = preflow_push(self.ssN, 's', 't', 'upper', self._R,
                               value_only=True)
        self.it += 1
        # feasible
        if round(self._R.graph['flow_value'] - self.bPlus, 4) == 0:
            for a in subset:
                self._checkUpper.remove(a)
            self._checkForAdditionallySatisfiedBounds()
        # infeasible
        else:
            self._removeUpperBoundsFromR(subset)
            self._testUpperBoundsIndividually(subset)

            
    def _testUpperBoundsIndividually(self, subset):
        '''Check if the upper bounds in subset can be added individually without
        loosing feasibility.'''
        nrFeasible = 0
        for a in subset:
            # element might be removed in the mean time
            try:
                i = self._checkUpper.index(a)
            except ValueError:
                nrFeasible += 1
                continue
            # if all other bounds in the subset were feasible, u_a is infeasible
            if nrFeasible == len(subset)-1:
                del self._checkUpper[i]
                assert (a[1] not in self.C[a[0]] or
                        self.C[a[0]][a[1]] == None)
                self.C[a[0]][a[1]] = 'u'
                continue
            # add bound and solve max-flow
            self._addUpperBoundsToR([a])
            self._R = preflow_push(self.ssN, 's', 't', 'upper', self._R,
                                   value_only=True)
            self.it += 1
            # feasible
            if round(self._R.graph['flow_value'] - self.bPlus, 4) == 0:
                del self._checkUpper[i]
                self._checkForAdditionallySatisfiedBounds()
                nrFeasible += 1
            # infeasible, bound belongs in cover
            else:
                del self._checkUpper[i]
                self._removeUpperBoundsFromR([a])
                assert (a[1] not in self.C[a[0]] or
                        self.C[a[0]][a[1]] == None)
                self.C[a[0]][a[1]] = 'u'

                
    def _checkForAdditionallySatisfiedBounds(self):
        '''Check the current solution and remove all bounds that are satisfied
        from the lists of to be checked arcs.'''
        i = len(self._checkLower)
        while i > 0:
            i -= 1
            v,w = self._checkLower[i]
            if self._R[w][v]['flow'] <= 0:
                del self._checkLower[i]
                self._R[w][v]['capacity'] = 0
        i = len(self._checkUpper)
        while i > 0:
            i -= 1
            v,w = self._checkUpper[i]
            if self._R[v][w]['flow'] <= self.N[v][w]['upper']:
                del self._checkUpper[i]
                self._R[v][w]['capacity'] = self.N[v][w]['upper']
                
                
    def _addLowerBoundsToR(self, arcs):
        '''Add the lower bounds of arcs to R.'''
        for v,w in arcs:
            self._R[w][v]['capacity'] = 0

    def _addUpperBoundsToR(self, arcs):
        '''Add the upper bounds of arcs to R.'''
        for v,w in arcs:
            self._R[v][w]['capacity'] = self.N[v][w]['upper']

    def _removeBoundsFromR(self, arcs):
        '''Remove lower and upper bounds of arcs from R.'''
        self._removeLowerBoundsFromR(arcs)
        self._removeUpperBoundsFromR(arcs)

    def _removeLowerBoundsFromR(self, arcs):
        '''Remove lower bounds of arcs from R.'''
        for v,w in arcs:
            self._R[w][v]['capacity'] = self._R.graph['inf']

    def _removeUpperBoundsFromR(self, arcs):
        '''Remove upper bounds of arcs from R.'''
        for v,w in arcs:
            self._R[v][w]['capacity'] = self._R.graph['inf']

    def _addZeroBoundsToR(self):
        '''Add arcs with 0 upper bounds in N to R (these were ignored by nx).'''
        for v,w,_ in ifilter(lambda a: a[2]['upper'] == 0,
                             self.N.edges_iter(data=True)):
            self._R.add_edge(v,w, capacity=0, flow=0)
            self._R.add_edge(w,v, capacity=0, flow=0)

