#!/usr/bin/python2
# -*- coding: utf-8 -*-

import os
import networkx as nx
import unittest

from netmic.heuristics.IISCover import *


class IISCoverTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=1)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)

    def testMinCut(self):
        IISCoverBuilder = IISCover(self.G)
        for (u,v,d) in IISCoverBuilder._R.edges(data=True):
            d['flow'] = 0
        IISCoverBuilder._R[4]['t']['flow'] = 1
        IISCoverBuilder._R['t'][4]['flow'] = -1
        S = IISCoverBuilder._minCutFromResidual()
        self.assertEqual(S, set([4]))
        
    def testFindIIS(self):
        IISCoverBuilder = IISCover(self.G)
        IISCoverBuilder.findIIS()
        self.assertEqual(IISCoverBuilder.listOfS, [[1,2,3,4]])

    def testFindNonredundantIIS(self):
        IISCoverBuilder = IISCover(self.G)
        IISCoverBuilder.findIIS(True)
        self.assertEqual(IISCoverBuilder.listOfS, [[1,2,3,4]])

    def testFindNonredundantIISRedundant(self):
        self.G.node[4]['supply'] = -2
        self.G.node[1]['supply'] = 4
        IISCoverBuilder = IISCover(self.G)
        IISCoverBuilder.findIIS(True)
        self.assertEqual(IISCoverBuilder.listOfS, [[4], [5,6]])

    def testSetBoundsToInf(self):
        IISCoverBuilder = IISCover(self.G)
        IISCoverBuilder._setBoundsToInf([(3,5),(6,5)])
        self.assertEqual(IISCoverBuilder._R[5][3]['capacity'],
                         IISCoverBuilder._R.graph['inf'])
        self.assertEqual(IISCoverBuilder._R[3][5]['capacity'],
                         IISCoverBuilder._R.graph['inf'])
        self.assertFalse('upper' in IISCoverBuilder.ssN[6][5])
        self.assertTrue(IISCoverBuilder.ssN.has_edge(5,6))

    def testCoverFromResidual(self):
        IISCoverBuilder = IISCover(self.G)
        IISCoverBuilder._R.add_edge(3,5,
                                    capacity=IISCoverBuilder._R.graph['inf'])
        IISCoverBuilder._R.add_edge(5,3,
                                    capacity=IISCoverBuilder._R.graph['inf'])
        IISCoverBuilder._R[5][6]['capacity'] = IISCoverBuilder._R.graph['inf']
        IISCoverBuilder._R[2][1]['capacity'] = IISCoverBuilder._R.graph['inf']
        IISCoverBuilder._R[1][3]['capacity'] = IISCoverBuilder._R.graph['inf']
        C = {1:{2:'l', 3:'u'}, 2:{}, 3:{5:'u'}, 4:{}, 5:{}, 6:{5:'l'}}
        x = {1:{2:-1, 3:4}, 2:{3:1}, 3:{4:1,5:2}, 4:{}, 5:{}, 6:{5:-1}}
        self.assertEqual(IISCoverBuilder.coverFromSolution(),(C,x))
        
    def testIISFixedCover(self):
        C,x,it,_,_ = fixedCover(self.G)
        self.assertEqual(C, {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{}, 6:{5:'l'}})
        self.assertEqual(it, 4)
        self.assertEqual(x, {1:{2:1, 3:2}, 2:{3:3}, 3:{4:1, 5:2}, 4:{}, 5:{},
                             6:{5:-1}})

    def testIISCoverNonredundant(self):
        self.G.node[4]['supply'] = -2
        self.G.node[1]['supply'] = 4        
        C,x,it,_,_ = fixedCover(self.G, True)
        self.assertEqual(C, {1:{}, 2:{3:'u'}, 3:{4: 'u', 5:'u'}, 4:{}, 5:{},
                             6:{5:'l'}})
        self.assertEqual(it, 4)
        self.assertEqual(x, {1:{2:2, 3:2}, 2:{3:4}, 3:{4:2, 5:2}, 4:{}, 5:{},
                             6:{5:-1}})

    def testSetUpdateIIS(self):
        ICB = IISSetCover(self.G)
        ICB.listOfS = [[1,2,3], [1,2,3,4]]
        ICB._updateIISs()
        self.assertEqual(ICB.IISList, [{'delta': [(3,4), (3,5)], 'covered': 0},
                                       {'delta': [(3,5)], 'covered': 0}])
        self.assertEqual(ICB.arcDict[(1,2)], []) 
        self.assertEqual(ICB.arcDict[(3,4)], [0]) 
        self.assertEqual(ICB.arcDict[(3,5)], [0,1])

    def testSetCoverHeuristic(self):
        ICB = IISSetCover(self.G)
        ICB.IISList = [{'delta': [(3,4), (3,5)], 'covered': 0},
                       {'delta': [(6,5)], 'covered': 0}]
        for a in self.G.edges_iter():
            ICB.arcDict[a] = []
        ICB.arcDict[(3,4)] = [0]
        ICB.arcDict[(3,5)] = [0]
        ICB.arcDict[(6,5)] = [1]
        arcs = ICB._setCoverHeuristic()
        self.assertEqual(ICB.IISList[0]['covered'], 1)
        self.assertEqual(ICB.IISList[1]['covered'], 1)
        self.assertEqual(arcs, [(6,5), (3,5)])

    def testResetBounds(self):
        ICB = IISCover(self.G)
        ICB._setBoundsToInf([(3,5),(6,5)])
        ICB._resetBounds([(3,5),(6,5)])
        self.assertFalse(ICB.ssN.has_edge(5,6))
        self.assertEqual(ICB.ssN[6][5]['upper'], 2)
        self.assertEqual(ICB._R[3][5]['capacity'], 0)
        self.assertEqual(ICB._R[5][3]['capacity'], 0)

    def testIISSetCover(self):
        C,x,it,_,_ = setCover(self.G)
        self.assertEqual(C, {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{}, 6:{5:'l'}})
        self.assertEqual(it, 4)
        self.assertEqual(x, {1:{2:1, 3:2}, 2:{3:3}, 3:{4:1, 5:2}, 4:{}, 5:{},
                             6:{5:-1}})
        
    def testHittingSetheuristic(self):
        ICB = IISHittingSet(self.G)
        ICB.IISList = [[(3,4), (3,5)], [(6,5)]]
        for a in self.G.edges_iter():
            ICB.arcDict[a] = []
        ICB.arcDict[(3,4)] = [0]
        ICB.arcDict[(3,5)] = [0]
        ICB.arcDict[(6,5)] = [1]
        arcs = ICB._hittingSetHeuristic()
        self.assertEqual(arcs, [(3,5), (6,5)])

    def testIISHittingSet(self):
        C,x,it,_,_ = hittingSet(self.G)
        self.assertEqual(C, {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{}, 6:{5:'l'}})
        self.assertEqual(it, 4)
        self.assertEqual(x, {1:{2:1, 3:2}, 2:{3:3}, 3:{4:1, 5:2}, 4:{}, 5:{},
                             6:{5:-1}})
        
class TwoIISsTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-5)
        self.G.add_edge(1,3, upper=2)
        self.G.add_edge(2,3, upper=1)

    def testFindIIS(self):
        IISCoverBuilder = IISCover(self.G)
        IISCoverBuilder.findIIS()
        self.assertEqual(IISCoverBuilder.listOfS, [[1],[2]])
        
        
class findIISTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_edges_from([(4,'t'), (5,'t'), (6,'t')], upper=1)
        self.G.add_edge('s', 1, upper=3)
        self.G.add_edge('s', 2, upper=2)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(1, 4, upper=1)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=1)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)

        
if __name__ == '__main__':
    unittest.main()
