#!/usr/bin/python2
# -*- coding: utf-8 -*-

import networkx as nx
import unittest
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil
from netmic.heuristics.flow_heuristics import *

class heuristic1Tests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=2)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)
        self.x1 = {1: {3: 2}, 2: {3: 1}, 3: {4: 1}}
        alUtil.completeSolution(self.G, self.x1)
        self.x2 = {1: {3: 3}, 2: {3: 1}, 3: {4: 1, 5: 1}}
        alUtil.completeSolution(self.G, self.x2)

    def testNotImplemented(self):
        builder = heuristicSolver(self.G)
        with self.assertRaises(NotImplementedError):
            builder.solve()
        builder = heuristic1Solver(self.G)
        with self.assertRaises(NotImplementedError):
            builder.solve()
        builder = heuristicx_3Solver(self.G)
        with self.assertRaises(NotImplementedError):
            builder.solve()

    def testL1StartFlow(self):
        builder = heuristic1Solver(self.G)
        self.assertDictEqual(builder.computeStartSolution(), self.x1)

    def testChangeSolutionAndUpdateSandT(self):
        builder = heuristic1Solver(self.G)
        builder.x = self.x2
        builder._S = [2,3]
        builder._T = [5,6] 
        builder._changeSolutionAndUpdateSandT([2,3,5,6])
        self.assertEqual(builder.x[6][5], -1)
        self.assertListEqual(builder._S, [3])
        self.assertListEqual(builder._T, [5])
        
    def testHeuristic1_1Solve(self):
        builder = heuristic1_1Solver(self.G)
        x = {1: {3: 3}, 2: {3: 2}, 3: {4: 1, 5: 2}, 6: {5: -1}}
        alUtil.completeSolution(self.G, x)
        builder.solve()
        self.assertEqual(builder.x, x)

    def testHeuristic1_1SolveFloyd(self):
        builder = heuristic1_1Solver(self.G)
        x = {1: {3: 3}, 2: {3: 2}, 3: {4: 1, 5: 2}, 6: {5: -1}}
        alUtil.completeSolution(self.G, x)
        builder.solve('floyd')
        self.assertEqual(builder.x, x)

    def testHeuristic1_1SolveCombined(self):
        builder = heuristic1_1Solver(self.G)
        x = {1: {3: 3}, 2: {3: 2}, 3: {4: 1, 5: 2}, 6: {5: -1}}
        alUtil.completeSolution(self.G, x)
        builder.solve('combined')
        self.assertEqual(builder.x, x)

    def testHeuristic1_1(self):
        C = {1:{3:'u'}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{}, 6:{5:'l'}}
        x = {1: {2:0, 3: 3}, 2: {3: 2}, 3: {4: 1, 5: 2}, 4:{}, 5:{}, 6:{5: -1}}
        self.assertEqual(heuristic1_1(self.G), (C,x,2))

    def testHeuristic1_2Solve(self):
        builder = heuristic1_2Solver(self.G)
        x = {1: {3: 3}, 2: {3: 2}, 3: {4: 1, 5: 2}, 6: {5: -1}}
        alUtil.completeSolution(self.G, x)
        builder.solve()
        self.assertEqual(builder.x, x)

    def testHeuristic1_2(self):
        C = {1:{3:'u'}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{}, 6:{5:'l'}}
        x = {1: {2:0, 3: 3}, 2: {3: 2}, 3: {4: 1, 5: 2}, 4:{}, 5:{}, 6: {5: -1}}
        self.assertEqual(heuristic1_2(self.G), (C,x,5))

    def testHeuristic1_3Solve(self):
        builder = heuristic1_3Solver(self.G)
        x = {1: {3: 3}, 2: {3: 2}, 3: {4: 1, 5: 2}, 6: {5: -1}}
        alUtil.completeSolution(self.G, x)
        builder.solve()
        self.assertEqual(builder.x, x)
        
    def testHeuristic1_3VariantSolve(self):
        builder = heuristic1_3Solver(self.G, True)
        x = {1: {2:1, 3:2}, 2: {3: 3}, 3: {4: 1, 5: 2}, 6: {5: -1}}
        alUtil.completeSolution(self.G, x)
        builder.solve()
        self.assertDictEqual(builder.x, x)

    def testHeuristic1_3(self):
        C = {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{}, 6:{5:'l'}}
        x = {1:{2:1, 3:2}, 2:{3: 3}, 3:{4: 1, 5: 2}, 4:{}, 5:{}, 6:{5: -1}}
        self.assertEqual(heuristic1_3(self.G, True), (C,x,5))

class negativeCycleTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(0, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(0, 1, weight=0)
        self.G.add_edge(1, 2, weight=0)
        self.G.add_edge(1, 3, weight=0)
        self.G.add_edge(2, 3, weight=0)
        self.G.add_edge(3, 4, weight=-1)       
        self.G.add_edge(4, 2, weight=1)
        self.G.add_edge(4, 6, weight=1)

    def testBFSPredecessorToPath(self):
        builder = heuristicx_3Solver(False)
        builder.arcPairN = self.G
        pred = {0:None, 1:0, 3:1, 4:3, 6:4}
        self.assertEqual(builder._BFSPredecessorToPath(0,6, pred), [0,1,3,4,6])

    def testBFSPredecessorToPathsIter(self):
        builder = heuristicx_3Solver(False)
        builder.arcPairN = self.G
        pred = {0:None, 1:0, 2:4, 3:1, 4:3, 6:4}
        paths = {0:[0], 1:[], 2:[], 3:[0,1,3], 4:[], 6:[]}
        builder._constructPathIteratively(1,6,pred,paths)
        self.assertEqual(paths, {0:[0], 1:[], 2:[], 3:[0,1,3], 4:[],
                                 6:[0,1,3,4,6]})

    def testBFSPredecessorToPaths(self):
        builder = heuristicx_3Solver(False)
        builder.arcPairN = self.G
        pred = {0:None, 1:0, 2:4, 3:1, 4:3, 6:4}
        self.assertEqual(builder._BFSPredecessorToPaths(0, pred),
                         {0:[0], 1:[0,1], 2:[0,1,3,4,2], 3:[0,1,3], 4:[0,1,3,4],
                          6:[0,1,3,4,6]})
        
    def testShortPathWithNegCycles(self):
        builder = heuristic1_3Solver(self.G)
        self.G.add_edge(2, 1, weight=-1)
        builder.arcPairN = self.G
        self.assertEqual(builder.shortPathWithNegCycles(0,6), [0,1,3,4,6])

    def testShortPathWithNegCyclesVariant(self):
        builder = heuristic1_3Solver(self.G, True)
        self.G.add_edge(2, 1, weight=-1)
        builder.arcPairN = self.G
        self.assertEqual(builder.shortPathWithNegCycles(0,6), [0,1,2,3,4,6])

    def testShortPathWithNegCyclesException(self):
        builder = heuristicx_3Solver(True)
        with self.assertRaises(NotImplementedError):
            builder.shortPathWithNegCycles(0,6)

        
class startSolutionTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=2)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)
        self.builder = heuristicEll0StartSolution(self.G)

    def testResetSupplyBounds(self):
        self.builder._resetSupplyBounds(netUtil.listOfSupplyNodes_iter(self.G),
                                        netUtil.listOfDemandNodes_iter(self.G))
        self.assertEqual(sum(self.builder.ssN['s'][v]['upper'] for v in
                             self.builder.N.nodes_iter() if
                             self.builder.N.node[v]['supply'] > 0), 0)
        self.assertEqual(sum(self.builder.ssN[v]['t']['upper'] for v in
                             self.builder.N.nodes_iter()if
                             self.builder.N.node[v]['supply'] < 0), 0)

    def testUpdateSolutionAndBounds(self):
        auxX = alUtil.emptySolutionDict(self.G)
        auxX[1][2] = 1
        auxX[2][3] = 1
        alUtil.completeSolution(self.G, auxX)
        self.builder._updateSolutionAndBounds(auxX)
        self.assertEqual(sum(self.builder.ssN[v][w]['upper'] for v,w in
                             self.builder.N.edges_iter()), 7)
        self.assertEqual(self.builder.x[1][2], 1)
        self.assertEqual(self.builder.x[2][3], 1)

    def testUpdateDemandList(self):
        self.G.node[5]['supply'] = -3
        SSB = heuristicEll0StartSolution(self.G)
        auxX = {3:{'t':0}, 4:{'t':1}, 5:{'t':2}, 6:{'t':0}}
        SSB._updateDemandList(auxX)
        self.assertEqual(SSB.T, [(1,5),(2,3),(1,6)])
        self.assertEqual(SSB.demandNodeList, {3: 2, 5:1, 6:1})

    def testUpdateSupplyList(self):
        auxX = {1:{'s': 2}, 2:{'s':0}}
        self.builder._updateSupplyList(auxX)
        self.assertEqual(self.builder.S, [(1,1),(2,2)])
        self.assertEqual(self.builder.supplyNodeList, {1:1, 2:2})
    
    def testSendFlowFromS(self):
        self.builder._resetSupplyBounds(netUtil.listOfSupplyNodes_iter(self.G),
                                        netUtil.listOfDemandNodes_iter(self.G))
        self.builder.S = [(3,1),(2,2)]
        self.builder._sendFlowFromS()
        self.assertEqual(self.builder.S, [(2,2)])
        self.assertEqual(self.builder.T, [(1,6),(1,5)])
        self.assertEqual(self.builder.ssN[3][4]['upper'], 1)
        self.assertEqual(sum(self.builder.ssN['s'][v]['upper'] for v in
                             self.builder.N.nodes_iter() if
                             self.builder.N.node[v]['supply'] > 0), 0)
        self.assertEqual(sum(self.builder.ssN[v]['t']['upper'] for v in
                             self.builder.N.nodes_iter()if
                             self.builder.N.node[v]['supply'] < 0), 0)


    def testSendFlowFromT(self):
        self.builder._resetSupplyBounds(netUtil.listOfSupplyNodes_iter(self.G),
                                        netUtil.listOfDemandNodes_iter(self.G))
        self.builder._sendFlowFromT()
        self.assertEqual(self.builder.T, [(1,5),(1,6),(2,3)])
        self.assertEqual(self.builder.ssN[3][1]['upper'], 1)
        self.assertEqual(self.builder.x[3][4], 1)
        self.assertEqual(sum(self.builder.ssN[v]['s']['upper'] for v in
                             self.G.nodes_iter() if
                             self.G.node[v]['supply'] > 0), 0)
        

    def testComputeStartSolution(self):
        x = self.builder.solve()
        self.assertEqual(x, {1:{2:0, 3:2}, 2:{3:1}, 3:{4:1, 5:0}, 4:{}, 5:{},
                             6:{5:0}})

        
class heuristic2Tests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=2)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)
        self.x1 = {1: {3: 0}, 2: {3: 1}, 3: {4: 1}}
        alUtil.completeSolution(self.G, self.x1)
        self.x2 = {1: {3: 3}, 2: {3: 1}, 3: {4: 1, 5: 1}}
        alUtil.completeSolution(self.G, self.x2)

    def testNotImplemented(self):
        builder = heuristic2Solver(self.G)
        with self.assertRaises(NotImplementedError):
            builder.solve()

    def testHeuristic2Init(self):
        builder = heuristic2Solver(self.G)
        self.assertEqual(builder._S, [(-1,1),(-1,2)])
        self.assertEqual(builder._T, [5,6])
        self.assertEqual(builder.excess, {1:1, 2:1, 5:-1, 6:-1})

    def testInitializeExcessData(self):
        builder = heuristic2Solver(self.G)
        x = {6:{5:-2}}
        alUtil.completeSolution(self.G, x)
        builder.x = x
        builder._S, builder._T, builder.excess = [], [], dict()
        builder._initializeExcessData()
        self.assertEqual(builder._S, [(-3,1),(-2,2),(-1,6)])
        self.assertEqual(builder._T, [3,4,5])

    def testChangeSolutionAndUpdateSandT(self):
        builder = heuristic2Solver(self.G)
        builder.excess[2] = 3
        builder._changeSolutionAndUpdateSandT([2,1,3,5,6])
        self.assertEqual(builder._S, [(-2,6),(-1,2),(-1,1)])
        self.assertEqual(builder._T, [5])
        self.assertEqual(builder.excess, {1:1, 5:-1, 6:2})
        self.assertEqual(builder.x, {1:{2:-3, 3:5}, 2:{3:1}, 3:{4:1, 5:3}, 4:{},
                                     5:{}, 6:{5:-3}})

    def testHeuristic2_2Solve(self):
        builder = heuristic2_2Solver(self.G)
        builder.solve()
        self.assertEqual(builder.x, {1:{2:0, 3:3}, 2:{3:2}, 3:{4:1, 5:2}, 4:{},
                                     5:{}, 6:{5:-1}})

    def testHeuristic2_2(self):
        C,x,it,_ = heuristic2_2(self.G)
        self.assertEqual((x,it), ({1:{2:0, 3:3}, 2:{3:2}, 3:{4:1, 5:2}, 4:{},
                                   5:{}, 6:{5:-1}},2))
        self.assertEqual(C, {1:{3:'u'}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{},
                             6:{5:'l'}})


    def testHeuristic2_3Solve(self):
        builder = heuristic2_3Solver(self.G)
        builder.solve()
        self.assertEqual(builder.x, {1:{2:0, 3:3}, 2:{3:2}, 3:{4:1, 5:2}, 4:{},
                                     5:{}, 6:{5:-1}})

    def testHeuristic2_3(self):
        C,x,it,_ = heuristic2_3(self.G)
        self.assertEqual((x,it), ({1:{2:0, 3:3}, 2:{3:2}, 3:{4:1, 5:2}, 4:{},
                                   5:{}, 6:{5:-1}},2))
        self.assertEqual(C, {1:{3:'u'}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{},
                             6:{5:'l'}})
        
class heuristic2LargerTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.read_graphml('netmic/heuristics/tests/20.graphml')

    def testEll0StartSol(self):
        builder = heuristicEll0StartSolution(self.G)
        x = builder.solve()
        self.assertEqual(x['5'], {'13': 2.47, '15': 0, '17': 0.19999999999999973,
                                  '18': 1.19, '1': 0, '9': 0})

    def testHeuristic2_2(self):
        C,x,it,_ = heuristic2_2(self.G)
        self.assertEqual(x['15']['11'], 6.35)
        self.assertEqual(it, 1)
        self.assertEqual(alUtil.sizeOfCover(C), 1)
        
if __name__ == '__main__':
    unittest.main()

