#!/usr/bin/python2
# -*- coding: utf-8 -*-

import networkx as nx
import unittest
import netmic.algorithmic_utility as alUtil

from netmic.heuristics.greedy_maxFS import *

class MaxFSTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=2)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)

    def testMaxFS(self):
        C = {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{}, 6:{5:'l'}}
        x = {1:{2:1, 3:2}, 2:{3:3}, 3:{4:1, 5:2}, 4:{}, 5:{}, 6:{5:-1}}
        self.assertEqual(maxFSHeuristic(self.G), (C,x,6))

    def testSolve(self):
        builder = maxFSSolver(self.G)
        builder.solve(3)
        self.assertEqual(builder.C, {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{},
                                     6:{5:'l'}})

    def testSolutionBuilding(self):
        builder = maxFSSolver(self.G)
        builder.solve(3)
        self.assertEqual(builder.x, {1:{2:1, 3:2}, 2:{3:3}, 3:{4:1, 5:2}, 4:{},
                                     5:{}, 6:{5:-1}})
        
    def testLowerBounds(self):
        builder = maxFSSolver(self.G)
        builder._testLowerBounds([(1,3), (3,4), (3,5)])
        self.assertEqual(builder._checkLower, [(6,5)])
        self.assertEqual(builder._checkUpper, [(1,3), (2,3), (3,5)])
        self.assertEqual(builder.C, {1:{}, 2:{}, 3:{}, 4:{}, 5:{}, 6:{}})
        
        
    def testLowerBoundIndividually(self):
        builder = maxFSSolver(self.G)
        builder._testLowerBoundsIndividually([(1,2), (6,5), (3,4)])
        self.assertEqual(builder._checkLower, [])
        self.assertEqual(builder._checkUpper, [(1,3), (2,3), (3,5)])
        self.assertEqual(builder.C, {1:{}, 2:{}, 3:{}, 4:{}, 5:{}, 6:{5:'l'}})
        
    def testUpperBoundsInfeasible(self):
        builder = maxFSSolver(self.G)
        builder._testUpperBounds([(2,3), (1,2), (3,5)])
        self.assertEqual(builder._checkLower, [(1,2),(6,5)])
        self.assertEqual(builder._checkUpper, [(1,3)])
        self.assertEqual(builder.C, {1:{}, 2:{}, 3:{5:'u'}, 4:{}, 5:{}, 6:{}})

    def testUpperBoundsFeasible(self):
        builder = maxFSSolver(self.G)
        builder._testUpperBounds([(1,2), (1,3), (6,5)])
        self.assertEqual(builder._checkLower, [(6,5)])
        self.assertEqual(builder._checkUpper, [(2,3),(3,5)])

    def testUpperBoundsIndividuallyInfeasible(self):
        builder = maxFSSolver(self.G)
        builder._testUpperBoundsIndividually([(2,3), (3,5)])
        self.assertEqual(builder._checkLower, [(1,2),(6,5)])
        self.assertEqual(builder._checkUpper, [(1,3)])
        self.assertEqual(builder.C, {1:{}, 2:{}, 3:{5:'u'}, 4:{}, 5:{}, 6:{}})
 
    def testUpperBoundsIndividuallyFeasible(self):
        builder = maxFSSolver(self.G)
        builder._testUpperBoundsIndividually([(1,2), (1,3), (6,5)])
        self.assertEqual(builder._checkLower, [(6,5)])
        self.assertEqual(builder._checkUpper, [(2,3),(3,5)])
        self.assertEqual(builder.C, alUtil.emptySolutionDict(self.G))
 
    def testCheckForAdditionallyBounds(self):
        builder = maxFSSolver(self.G)
        for v,w in self.G.edges_iter():
            builder._R[v][w]['flow'] = 0
            builder._R[w][v]['flow'] = 0
        builder._R[1][2]['flow'] = 3
        builder._R[5][6]['flow'] = 1
        builder._checkForAdditionallySatisfiedBounds()
        self.assertEqual(builder._checkLower, [(6,5)])
        self.assertEqual(builder._checkUpper, [(1,2)])
        self.assertEqual(sum(builder._R[v][w]['capacity'] for v,w in
                             self.G.edges_iter()), 7+57)
        self.assertEqual(sum(builder._R[w][v]['capacity'] for v,w in
                             self.G.edges_iter()), 0+57)

    def testInit(self):
        builder = maxFSSolver(self.G)
        self.assertEqual(sum(builder._R[v][w]['capacity'] for v,w in
                             self.G.edges_iter()), 342)
        self.assertEqual(sum(builder._R[w][v]['capacity'] for v,w in
                             self.G.edges_iter()), 342)
        
        
if __name__ == '__main__':
    unittest.main()
