#!/usr/bin/python2
# -*- coding: utf-8 -*-

import networkx as nx
import unittest
import netmic.algorithmic_utility as alUtil

from netmic.heuristics.tree_heuristic import *

class algorithmTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(1, 4, upper=1)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=1)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)

    def testSetWeightsNegative(self):
        H = self.G.to_undirected()
        setWeights(H, 'negative')
        weightSum = sum(a[2]['negative'] for a in H.edges_iter(data=True))
        self.assertEqual(weightSum, -9)

    def testSetWeightsDemand(self):
        H = self.G.to_undirected()
        setWeights(H, 'demand')
        weightSum = sum(a[2]['demand'] for a in H.edges_iter(data=True))
        self.assertEqual(weightSum, -16)
        self.assertEqual(H.edge[1][2]['demand'], -3)

    def testSetWeightsFactor(self):
        H = self.G.to_undirected()
        setWeights(H, 'factor')
        weightSum = sum(a[2]['factor'] for a in H.edges_iter(data=True))
        self.assertAlmostEqual(weightSum, 4.666666666666667)
        self.assertEqual(H.edge[1][2]['factor'], 2/3.)

    def testSetWeightsFactorZero(self):
        self.G.node[5]['supply'] = 0
        self.G.node[6]['supply'] = 0
        H = self.G.to_undirected()
        setWeights(H, 'factor')        
        self.assertEqual(H.edge[6][5]['factor'], float('inf'))
        
    def testSpanningTreeUpper(self):
        edgelist = computeSpanningTree(self.G, 'upper')
        self.assertEqual(sorted(edgelist), [(1,4),(2,3),(3,4),(3,5),(5,6)])

    def testSpanningTreeNegative(self):
        edgelist = computeSpanningTree(self.G, 'negative')
        self.assertEqual(sorted(edgelist), [(1,2),(1,3),(1,4),(3,5),(5,6)])

    def testSpanningTreeDemand(self):
        edgelist = computeSpanningTree(self.G, 'demand')
        self.assertEqual(sorted(edgelist), [(1,2),(1,3),(1,4),(3,5),(5,6)])

    def testSpanningTreeFactor(self):
        edgelist = computeSpanningTree(self.G, 'factor')
        self.assertEqual(sorted(edgelist), [(1,4),(2,3),(3,4),(3,5),(5,6)])
        
    def testTreeHeuristicUpperFactor(self):
        C = alUtil.emptySolutionDict(self.G)
        C[2][3] = 'u'
        C[3][4] = 'u'
        C[3][5] = 'u'
        C[6][5] = 'l'
        C[1][4] = 'l'
        x = {1:{2:2, 3:2, 4:-1}, 2:{3:4}, 3:{4:2, 5:2}, 4:{}, 5:{}, 6:{5:-1}}
        self.assertEqual(treeHeuristic(self.G, 'upper'), (C,x))
        self.assertEqual(treeHeuristic(self.G, 'factor'), (C,x))

    def testTreeHeuristicUniformNegativeDemand(self):
        C = alUtil.emptySolutionDict(self.G)
        C[1][2] = 'l'
        C[1][3] = 'u'
        C[3][5] = 'u'
        C[6][5] = 'l'
        x = {1:{2:-1, 3:4, 4:0}, 2:{3:1}, 3:{4:1, 5:2}, 4:{}, 5:{}, 6:{5:-1}}
        self.assertEqual(treeHeuristic(self.G, 'uniform'), (C,x))
        self.assertEqual(treeHeuristic(self.G, 'negative'), (C,x))
        self.assertEqual(treeHeuristic(self.G, 'demand'), (C,x))

    def testTreeHeuristicCombined(self):
        C = alUtil.emptySolutionDict(self.G)
        C[1][2] = 'l'
        C[1][3] = 'u'
        C[3][5] = 'u'
        C[6][5] = 'l'
        x = {1:{2:-1, 3:4, 4:0}, 2:{3:1}, 3:{4:1, 5:2}, 4:{}, 5:{}, 6:{5:-1}}
        self.assertEqual(treeHeuristicCombined(self.G), (C,x))      
        
if __name__ == '__main__':
    unittest.main()
