# -*- coding: utf-8 -*-
'''
Tree Heuristics.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil

def setWeights(G, weight):
    '''Set dict "weight" on all edges. Supported are for a=(u,v)
    negative: w_a = -u_a; demand: w_a = -max( |b_u|, |b_v|)
    factor: w_a = u_a / max(|b_u|, |b_v|); upper: w_a = u_a;
    uniform: w_a = 1.'''
    if weight == 'upper' or weight == 'uniform':
        return
    elif weight == 'negative':
        for a in G.edges_iter(data=True):
            a[2]['negative'] = -a[2]['upper']
    elif weight == 'factor':
        for a in G.edges_iter(data=True):
            aux = max(abs(G.node[a[0]]['supply']), abs(G.node[a[1]]['supply']))
            if aux == 0:
                a[2]['factor'] = float('inf')
            else:
                a[2]['factor'] = float(a[2]['upper']) / aux
    elif weight == 'demand':
        for a in G.edges_iter(data=True):
            a[2]['demand'] = -max(abs(G.node[a[0]]['supply']),
                                  abs(G.node[a[1]]['supply']))
    else:
        raise NotImplementedError("Weight type unknown.")


def computeSpanningTree(N, weight):
    '''Return list of edges in a min spanning tree of N, with respect to
    "weight".'''
    G = N.to_undirected()
    setWeights(G, weight)
    # returns generator!
    mst = nx.minimum_spanning_edges(G, weight, data=False)
    return list(mst)

def treeHeuristic(N, weight):
    '''Return the cover computed by the tree approximation, where the min
    spanning tree is w.r.t. to weight.
    "weight" can be in {'upper', 'negative', 'uniform', 'demand', 'factor'}.'''
    tree = computeSpanningTree(N, weight)
    sssN = netUtil.setUpSingleSourceSinkNetwork(N)
    netUtil.removeInfiniteUpperBoundsForMaxFlow(sssN)
    netUtil.removeBoundsEdgesForMaxFlow(sssN, tree)
    value, x = nx.maximum_flow(sssN, 's', 't', 'upper')
    assert round(value - netUtil.sumBPlus(N), 4) == 0
    # remove solution values from aux. arcs
    alUtil.solutionFromResidualSolution(N,x)
    return alUtil.coverFromSolution(N,x),x
    
def treeHeuristicCombined(N):
    '''Return the best cover due to treeHeuristic over all weight functions.''' 
    weightFunctions = ['uniform', 'negative', 'demand', 'factor']
    CBest, xBest = treeHeuristic(N, 'upper')
    for weight in weightFunctions:
        C,x = treeHeuristic(N, weight)
        if alUtil.sizeOfCover(C) < alUtil.sizeOfCover(CBest):
            CBest,xBest = C,x
    return CBest, xBest

