# -*- coding: utf-8 -*-
'''
Weight functions as used in Flow Heuristics.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil

def computeArcWeightsSimple(N, revArcsN, x, arcs):
    '''Set 'weight' for arcs in revArcsN according to x, as used for 1.1.

    w_a=0: u_a already violated, w_a=1: x_a < u_a, w_a=2: x_a = u_a.
    '''
    for a in arcs:
        if N.has_edge(*a):
            v,w = a
        else: #backwards arc used
            w,v = a
        if round(x[v][w] - revArcsN.edge[v][w]["upper"], 6) > 0:
            revArcsN.edge[v][w]["weight"] = 0
        elif round(revArcsN.edge[v][w]["upper"] -  x[v][w], 6) > 0:
            revArcsN.edge[v][w]["weight"] = 1
        else:
            revArcsN.edge[v][w]["weight"] = 2
        if round(x[v][w], 6) < 0:
            revArcsN.edge[w][v]["weight"] = 0 
        elif round(x[v][w], 6) > 0:
            revArcsN.edge[w][v]["weight"] = 1
        else:
            revArcsN.edge[w][v]["weight"] = 2

def computeArcWeightsDetailedBinary(N, revArcsN, x, r):
    '''Set 'weight' in revArcsN, according to x and r, as used by heuristic 1.2.
    
    w_a=0 if x_a+r does not changed membership in C or removes a from C,
    w_a=1 if x_a+r adds to C.
    '''
    assert r > 0
    for a in N.edges_iter():
        (v,w) = a
        if round(x[v][w] - N[v][w]['upper'], 6) > 0 or round(x[v][w], 6) < 0:
            revArcsN.edge[v][w]["weight"] = 0
            revArcsN.edge[w][v]["weight"] = 0
        else: # l <= x <= u
            if x[v][w] + r <= N[v][w]['upper']:
                revArcsN.edge[v][w]["weight"] = 0
            else:
                revArcsN.edge[v][w]["weight"] = 1
            if x[v][w] - r >= 0:
                revArcsN.edge[w][v]["weight"] = 0
            else:
                revArcsN.edge[w][v]["weight"] = 1

def computeArcWeightsDetailedTernary(N, revArcsN, x, r):
    '''Set 'weight' in revArcsN, according to x and r as used by heuristic 1.3.
    
    w_a=0 if x_a+r does not changed membership in C, w_a=1 if x_a+r adds to C,
    w_a=-1 if x_a+r removes from C.
    '''
    assert r > 0
    for a in N.edges_iter():
        (v,w) = a
        if round(x[v][w] - N[v][w]['upper'], 6) > 0:
            revArcsN.edge[v][w]["weight"] = 0
            if x[v][w] - r >= 0 and x[v][w] - r <= N[v][w]['upper']:
                revArcsN.edge[w][v]["weight"] = -1
            else:
                revArcsN.edge[w][v]["weight"] = 0
        elif round(x[v][w], 6) < 0:
            if x[v][w] + r >= 0 and x[v][w] + r <= N[v][w]['upper']:
                revArcsN.edge[v][w]["weight"] = -1
            else:
                revArcsN.edge[v][w]["weight"] = 0
            revArcsN.edge[w][v]["weight"] = 0
        else: # l <= x <= u
            if x[v][w] + r <= N[v][w]['upper']:
                revArcsN.edge[v][w]["weight"] = 0
            else:
                revArcsN.edge[v][w]["weight"] = 1
            if x[v][w] - r >= 0:
                revArcsN.edge[w][v]["weight"] = 0
            else:
                revArcsN.edge[w][v]["weight"] = 1

