# -*- coding: utf-8 -*-

#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import os
import argparse
import csv
import networkx as nx
import netmic.timer as timer
from netmic.exceptions import *
import warnings

import netmic.graphml_checker as check
import netmic.network_utility as netUtil
import netmic.algorithmic_utility as alUtil
import netmic.preprocessing as prepro
import netmic.heuristics.flow_heuristics as flowHeu
import netmic.heuristics.greedy_maxFS as maxFS
import netmic.heuristics.ell1_model as ell1
import netmic.heuristics.IISCover as IISHeu
import netmic.heuristics.tree_heuristic as treeHeu
import netmic.heuristics.elementary_cycle_enumeration as FPT
import netmic.optimal.setCover as setCover
import netmic.optimal.mip_model as mip
import netmic.postprocessing.minimalize as minimalize
import netmic.postprocessing.refinement as refine
import netmic.solution_output as solOut


def main(args):
    '''Print a statistic of the IAC solution to stdout for a single network,
    or to a csv-file for a directory of networks.'''
    if os.path.isfile(args.network):
        print computeCover(args.network, args)

    elif os.path.isdir(args.network):
        if (args.func == 'IISFixed' or args.func == 'IISHitting' or
            args.func == 'IISSet' or args.func == "setCover"):
            fieldnames = ['name', 'n', 'm', 'it', 'sol', 'time', 'IISTime',
                          'coverTime']
        elif args.func == 'mip':
            fieldnames = ['name', 'n', 'm', 'it', 'sol', 'time', 'buildTime',
                          'solveTime']
        elif args.func == '2.2' or args.func == '2.3':
            fieldnames = ['name', 'n', 'm', 'it', 'sol', 'time', 'startSolTime']
        elif args.func == 'tree' or args.func == 'treeAll' or args.func == 'ell1':
            fieldnames = ['name', 'n', 'm', 'sol', 'time']
        else:
            fieldnames = ['name', 'n', 'm', 'it', 'sol', 'time']

        if args.bell or args.simple or args.multi or args.mini:    
            fieldnames.extend(['timePostpro', 'solPostpro', 'diff'])

        with open(args.outputName + '.csv', 'wb') as csvfile:
            writer = csv.DictWriter(csvfile, dialect='excel',
                                    fieldnames=fieldnames)
            writer.writerow({e: e for e in fieldnames})

        # save files in a list to sort them alphabetically    
        files = [f for f in os.listdir(args.network)
                 if f.endswith('.graphml')]
        files.sort()
        for net in files:
            with open(args.outputName + '.csv', 'ab') as csvfile:
                writer = csv.DictWriter(csvfile, dialect='excel',
                                        fieldnames=fieldnames)
                writer.writerow(computeCover(os.path.join(args.network,net),
                                             args))
    else:
        raise DataError('parameter network is neither a file nor a directory')


def computeCover(network, args):
    '''Compute an IAC with the given method.'''
    output = {'name': network[network.rfind('/')+1:-8]}
    try:
        dataChecker = check.graphmlConsistency(nx.read_graphml(network))
    except DataError:
        output['sol'] = 'Error'
        return output
    else:
        N = dataChecker.newN

    output['n'] = dataChecker.oldN.number_of_nodes()
    output['m'] = dataChecker.oldN.number_of_edges()
    # check for inconsistent bounds and balanced supply
    if not netUtil.checkPreconditions(N):
        #solOut.drawGraphWithColouredEdges(N, args.outputName + '.png')
        output['sol'] = 'trivial'
        return output
    # check if network is feasible
    if netUtil.feasibilityCheck(N):
        #solOut.drawGraphWithColouredEdges(N, args.outputName + '.png')
        output['sol'] = 'feasible'
        return output
    # remove parallel and anti-parallel arcs, and lower bounds 
    workN = netUtil.ensureNetworkForm(N)

    # preprocessing removes degree one nodes
    prepro.removeDegreeOneInnodes(workN)
    C = prepro.preprocessDegreeOneNodes(workN)
    # if prepro has already solved network, terminate
    if len(workN) <= 1:
        dataChecker.reverseNameConvention(C)
        output['time'] = "prepro"
        return output
     
    # call the actual method given by args to solve IAC
    if args.func == '1.1':
        with timer.TimerCPU() as t:
            COrig,x,output['it'] = flowHeu.heuristic1_1(workN, args.shortestPath,
                                                        args.timelimit)
            if x == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == '1.2':
        with timer.TimerCPU() as t:
            COrig,x,output['it'] = flowHeu.heuristic1_2(workN, args.timelimit)
            if x == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == '1.3':
        with timer.TimerCPU() as t:
            COrig,x,output['it'] = flowHeu.heuristic1_3(workN, args.variant,
                                                        args.timelimit)
            if x == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == '2.2':
        with timer.TimerCPU() as t:
            COrig,x,output['it'],output['startSolTime'] = (
                flowHeu.heuristic2_2(workN, args.timelimit))
            if x == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == '2.3':
        with timer.TimerCPU() as t:
            COrig,x,output['it'],output['startSolTime'] = (
                flowHeu.heuristic2_3(workN, args.variant, args.timelimit))
            if x == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == 'maxfs':
        with timer.TimerCPU() as t:
            COrig,x,output['it'] = maxFS.maxFSHeuristic(workN, args.blockSize,
                                                        args.timelimit)
            if COrig == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == 'ell1':
        with timer.TimerCPU() as t:
            try:
                COrig,x = ell1.solveEll1(workN, args.timelimit)
            except ModelUnsolvedError:
                output['sol'] = 'Error'
                return output
            if COrig == None:
                output['sol'],output['time'] = "None","timelimit"
                return output
    
    if args.func == 'IISFixed':
        with timer.TimerCPU() as t:
            COrig,x,output['it'],output['coverTime'],output['IISTime'] = (
                IISHeu.fixedCover(workN, not args.redundant, args.timelimit))
            if COrig == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == 'IISHitting':
        with timer.TimerCPU() as t:
            COrig,x,output['it'],output['coverTime'],output['IISTime'] = (
                IISHeu.hittingSet(workN, not args.redundant, args.timelimit))
            if COrig == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == 'IISSet':
        with timer.TimerCPU() as t:
            COrig,x,output['it'],output['coverTime'],output['IISTime'] = (
                IISHeu.setCover(workN, not args.redundant, args.timelimit))
            if COrig == None:
                output['sol'],output['time'] = "None","timelimit"
                return output

    if args.func == 'tree':
        with timer.TimerCPU() as t:
            COrig,x = treeHeu.treeHeuristic(workN, args.weight)

    if args.func == 'treeAll':
        with timer.TimerCPU() as t:
            COrig,x = treeHeu.treeHeuristicCombined(workN)

    if args.func == 'FPT':
        with timer.TimerCPU() as t:
            COrig,x,output['it'] = FPT.enumerateElementaryCycles(workN,
                                                                 args.timelimit)

    if args.func == 'setCover':
        with timer.TimerCPU() as t:
            try:
                COrig,output['it'],output['coverTime'],output['IISTime'] = (
                    setCover.solveViaIISSetCover(workN, args.timelimit))
            except ModelUnsolvedError:
                output['sol'] = 'Error'
                return output
            if COrig == None:
                output['sol'],output['time'] = "None", "timelimit"
                return output
            
    if args.func == 'mip':
        with timer.TimerCPU() as t:
            try:
                COrig,output['buildTime'],output['solveTime'] = (
                    mip.solveViaMIP(workN, args.start, args.timelimit))
                output['it'] = 1
            except ModelUnsolvedError:
                output['sol'] = 'Error'
                return output
            except ModelNumericalError:
                output['sol'] = 'Numeric'
                return output
            if COrig == None:
                output['sol'],output['time'] = "None", "timelimit"
                return output

    output['time'] = round(t.secs,2)

    # call to potential postprocessing heuristics    
    if args.bell or args.simple or args.multi or args.mini:    
        output['timePostpro'] = 0
        if args.func == 'mip' or args.func == 'setCover':
            args.bell,args.simple,args.multi,args.mini = False,False,False,False
            warnings.warn("Postprocessing makes no sense for optimal approaches."
                          " Switching off.")
     
    if args.simple:
        with timer.TimerCPU() as t:
            refine.refineSolution(workN, x, 'simpleAss', args.timePost)
            CNew = alUtil.coverFromSolution(workN,x)
        output['timePostpro'] += t.secs
    
    if args.multi:
        with timer.TimerCPU() as t:
            refine.refineSolution(workN, x, 'multiAss', args.timePost)
            CNew = alUtil.coverFromSolution(workN,x)
        output['timePostpro'] += t.secs

    if args.bell:
        with timer.TimerCPU() as t:
            refine.refineSolution(workN, x, 'bellman', args.timePost)
            CNew = alUtil.coverFromSolution(workN,x)
        output['timePostpro'] += t.secs

    if args.mini:
        with timer.TimerCPU() as t:
            CNew = minimalize.minimalize(workN, COrig, args.timePost)
        output['timePostpro'] += t.secs

    # join cover from method and preprocessing            
    alUtil.joinCovers(COrig, C)
    output['sol'] = alUtil.sizeOfCover(COrig)

    if args.bell or args.simple or args.multi or args.mini:
        alUtil.joinCovers(CNew, C)
        output['solPostpro'] = alUtil.sizeOfCover(CNew)
        output['diff'] = alUtil.sizeOfCover(COrig) - alUtil.sizeOfCover(CNew)
        output['timePostpro']  = round(output['timePostpro'], 2)

    return output
    






