# -*- coding: utf-8 -*-
'''
Auxiliary functions concerning graph properties.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import sys        
import networkx as nx
from itertools import ifilter,product
import warnings
from netmic.exceptions import DataError

def listOfInNodes(N):
    '''Return a list of the intermediate nodes in N.'''
    return [v for v,d in N.nodes_iter(data=True) if round(d["supply"], 6) == 0]

def listOfDemandNodes(N):
    '''Return a list of the sinks in N.'''    
    return [v for v,d in N.nodes_iter(data=True) if round(d["supply"], 6) < 0]

def listOfDemandNodes_iter(N):
    '''Return a generator for the sinks in N.'''    
    return (v for v,d in N.nodes_iter(data=True) if round(d["supply"], 6) < 0)
    
def listOfSupplyNodes(N):
    '''Return a list of the sources in N.'''    
    return [v for v,d in N.nodes_iter(data=True) if round(d["supply"], 6) > 0]

def listOfSupplyNodes_iter(N):
    '''Return a generator for the sources in N.'''    
    return (v for v,d in N.nodes_iter(data=True) if round(d["supply"], 6) > 0)

def sumBInS(N, S):
    '''Return b(S).'''
    return sum(N.node[v]["supply"] for v in S)

def sumBPlus(N):
    '''Return overall supply in N.'''
    return sum(d["supply"] for v,d in N.nodes_iter(data=True)
               if round(d["supply"],6) > 0)
    
def areArcBoundsFeasible(N):
    '''Return true if lower bound <= upper bound on all arcs.'''
    for a in N.edges_iter(data=True):
        l = 0.0
        if "lower" in a[2]:
            l = a[2]["lower"]
        if l > a[2]["upper"]:
            warnings.warn("Trivial infeasibility on arc {}".format(a))
            return False
    return True

def isNetworkBalanced(N):
    '''Return true if b(V) = 0.'''
    check = round(sumBInS(N, N.nodes_iter()), 5) == 0
    if not check:
        warnings.warn("Network is not balanced.")
    return check

def removeLowerBoundOnArc(N, a, bigM=sys.maxint):
    '''Remove non-null lower bound on arc a. 
    If l = -inf, use bigM for supply and mark node in network correspondingly.'''
    if "lower" in a[2]:
        if a[2]['lower'] > float('-inf'):
            lowerBound = a[2]["lower"]
            N.node[a[0]]["supply"] -= lowerBound  # b_v = b_v - l_a
            N.node[a[1]]["supply"] += lowerBound  # b_w = b_w + l_a
            a[2]["upper"] -= lowerBound           # u_a = u_a - l_a
            del a[2]["lower"]                     # l_a = 0
        elif a[2]["upper"] >= float('inf'):
            N.node[a[0]]["supply"] += bigM        # b_v = b_v - (-M)
            N.node[a[0]]['bigM'] = bigM
            N.node[a[1]]["supply"] -= bigM        # b_w = b_w + (-M)
            N.node[a[1]]['bigM'] = bigM
            del a[2]["lower"]                     # l_a = 0
        else: #for some reason the arc was not reversed before
            reverseArcWithBounds(N, a)
            removeLowerBoundOnArc(N, (a[1],a[0], N.edge[a[1]][a[0]]), bigM)
            assert 'lower' not in N.edge[a[1]][a[0]]

def removeLowerBounds(N):
    '''Remove all nonzero lower bounds, direct arcs to larger bound
    if necessary (l_a=-inf, u_a < inf).'''
    bigM = computeBigM(N)
    for a in N.edges_iter(data=True):
       removeLowerBoundOnArc(N, a, bigM)
       
def computeBigM(N):
    '''Return $\max \abs(u_a) + \max \abs(\ell_a) +
    \sum_{v\in V | -\inf < b_v < \inf} b_v$.'''
    bigM = sumAbsWithoutInf([d["supply"] for v,d in N.nodes_iter(data=True)])
    try: 
        bigM += maxAbsWithoutInf([d['upper'] for v,w,d in N.edges_iter(data=True)])
    except ValueError: # list might be empty, in this case "add zero"
        pass
    try:
        bigM += maxAbsWithoutInf([d['lower'] for v,w,d in N.edges_iter(data=True)
                                  if 'lower' in d])
    except ValueError:
        pass
    return bigM + 1

def maxAbsWithoutInf(L):
    '''Return the maximum absolute value in L without 'inf'.
    Raise ValueError if there is no value.'''
    return max(abs(el) for el in L if abs(el) < float('inf'))

def sumAbsWithoutInf(L):
    '''Return the sum of absolute values in L without 'inf'.
    Raise ValueError if there is no value.'''    
    return sum(abs(el) for el in L if abs(el) < float('inf'))

def removeAntiParallelArcs(N):
    '''Check for pairs of oppositely directed arcs and remove one of them.'''
    antiPar = ifilter(lambda a: N.has_edge(a[1],a[0]), N.edges_iter())
    for v,w in antiPar:
        if N.has_edge(w,v):
            if 'lower' in N.edge[v][w]:
                N.edge[v][w]["lower"] -= N.edge[w][v]["upper"]
            else:
                N.edge[v][w]["lower"] = -N.edge[w][v]["upper"]
            if "lower" in N.edge[w][v]:
                N.edge[v][w]["upper"] -= N.edge[w][v]["lower"]
            N.remove_edge(w,v)

def ensureArcDirectionToLargerBound(N):
    '''Reverse arcs for which abs(l) > abs(u).'''
    for a in N.edges_iter(data=True):
        if 'lower' in a[2] and abs(a[2]['lower']) > abs(a[2]["upper"]):
            reverseArcWithBounds(N,a)

def reverseArcWithBounds(N, a):
    '''Reverse arc a.'''
    assert 'lower' in a[2]
    N.add_edge(a[1], a[0], lower = -a[2]['upper'], upper = -a[2]['lower'])
    N.remove_edge(a[0], a[1])
        
def ensureNetworkForm(N):
    '''Return network without reversed arcs, without lower bounds, and u>=0.'''
    newN = N.copy()
    removeAntiParallelArcs(newN)
    assert areArcBoundsFeasible(newN)
    ensureArcDirectionToLargerBound(newN)
    removeLowerBounds(newN)
    return newN

def setUpInfiniteBoundSingleSourceSinkNetwork(N):
    '''Return auxiliary network with a single source s and sink t, use keyword
    capacity for bounds from s and to t.'''
    # since s and t are the keys for identifying source and sink,
    # make sure that these names do not exist
    assert not N.has_node("s") and not N.has_node("t")
    sssN = N.copy()
    sssN.add_node("s")
    sssN.add_node("t")
    
    for v in ifilter(lambda v: "supply" in v[1], sssN.nodes_iter(data=True)):
        if round(v[1]["supply"], 6) > 0:   # for sources
            # add an arc from s to v with u = b_v
            sssN.add_edge("s", v[0], capacity=v[1]["supply"])
            del sssN.node[v[0]]["supply"]
        elif round(v[1]["supply"], 6) < 0: # for sinks
            # add an arc from v to t with u = -b_v
            sssN.add_edge(v[0], "t", capacity=-v[1]["supply"])
            del sssN.node[v[0]]["supply"]                
    return sssN

def setUpSingleSourceSinkNetwork(N, checkLowerBounds=True):
    '''Return auxiliary network with a single source s and sink t,
    remove lower bounds if necessary.'''
    # since s and t are the keys for identifying source and sink,
    # make sure that these names do not exist
    assert not N.has_node("s") and not N.has_node("t")
    sssN = N.copy()
    if checkLowerBounds:
        removeLowerBounds(sssN)
    sssN.add_node("s")
    sssN.add_node("t")
    
    for v in ifilter(lambda v: "supply" in v[1], sssN.nodes_iter(data=True)):
        if round(v[1]["supply"], 6) > 0:   # for sources
            # add an arc from s to v with u = b_v
            sssN.add_edge("s", v[0], upper=v[1]["supply"])
            del sssN.node[v[0]]["supply"]
        elif round(v[1]["supply"], 6) < 0: # for sinks
            # add an arc from v to t with u = -b_v
            sssN.add_edge(v[0], "t", upper=-v[1]["supply"])
            del sssN.node[v[0]]["supply"]                
    return sssN

def deltaMinus(N, S):
    '''Return $\delta^-(S)$ as a list.'''
    return [a for a in N.in_edges_iter(S) if a[0] not in S]

def deltaPlus(N, S):
    '''Return $\delta^+(S)$ as a list.'''
    A = N.out_edges(S)
    return [a for a in N.out_edges_iter(S) if a[1] not in S]

def delta(N, S):
    '''Return $\delta(S)$ as a list.'''
    return deltaPlus(N, S) + deltaMinus(N, S)

def deltaMinusIfLowerBound(N, S):
    '''Return $\delta^-(S)\setminus \{a: l_a=0\}$ as a list.'''
    return [a for a in N.in_edges_iter(S) if a[0] not in S and 'lower' in
            N.edge[a[0]][a[1]] and round(N.edge[a[0]][a[1]]['lower'], 6) != 0]

def deltaPlusIfLowerBound(N, S):
    '''Return $\delta^+(S)\setminus \{a: l_a=0\}$ as a list.'''
    return [a for a in N.out_edges_iter(S) if a[1] not in S and 'lower' in
            N.edge[a[0]][a[1]] and round(N.edge[a[0]][a[1]]['lower'], 6) != 0]    
       
def sumUInA(N, A):
    '''Return u(A).'''
    return sum(N.edge[v][w]['upper'] for (v,w) in A)

def sumLInA(N, A):
    ''' Return l(A).'''
    return sum(N.edge[v][w]['lower'] for (v,w) in A if 'lower' in N.edge[v][w])

# expects no lower bounds
def GHSupplyViolated(N, S):
    '''Return true if the GH-inequality in supply-form is violated for S.'''
    return round(sumBInS(N, S) - sumUInA(N, deltaPlus(N, S)), 6) > 0

# expects no lower bounds
def GHDemandViolated(N, S):
    '''Return true if the GH-inequality in demand-form is violated for S.'''
    return round(sumBInS(N, S) + sumUInA(N, deltaMinus(N, S)), 6) < 0

# expects no lower bounds
def GHViolated(N, S):
    '''Return true if one of the GH-inequalities is violated for S.'''
    return GHSupplyViolated(N, S) or GHDemandViolated(N, S)

def removeInfiniteUpperBoundsForMaxFlow(N):
    '''Remove all upper bounds with value "inf", including the key.'''
    for a in N.edges_iter(data=True):
        if a[2]['upper'] == float('inf'):
            del a[2]['upper']
        elif a[2]['upper'] == float('-inf'):
            raise DataError("u({},{}) = - inf".format(*a))
    
def feasibilityCheck(N):
    '''Return true if the network problem is feasible by computing the max-flow
    value in the single-source-sink-network.
    '''
    if not N.has_node("s") and not N.has_node("t"):
        auxN = setUpSingleSourceSinkNetwork(N)
    elif N.has_node("s") and N.has_node("t"):
        auxN = N.copy()
    else:
        raise DataError("Node s or t in network, but not both")

    # we cannot check against b for N since the removal of lower bounds changes
    # the values, but the upper bounds in the auxiliary network for $\delta(s)$
    # equal the supply
    bPlus = sumUInA(auxN, deltaPlus(auxN, ["s"]))
    assert bPlus < float('inf')
    # Networkx cannot handle "inf" and expects to have non-present capacities
    # instead. Cannot switch place with the previous step!
    removeInfiniteUpperBoundsForMaxFlow(auxN)
    flow = nx.maximum_flow_value(auxN, "s", "t", "upper")
    return abs(flow - bPlus) < 1e-4

def checkPreconditions(N):
    '''Return true if the network is balanced and arc bounds are consistent.'''
    return isNetworkBalanced(N) and areArcBoundsFeasible(N) 

def allNeighbours(N, v):
    '''Return list of all neighbours of a node,
    necessary because networkx.all_neighbors is buggy.
    '''
    auxList = N.successors(v) + N.predecessors(v)
    return list(set(auxList[:]))  # remove duplicate entries
    
def listDegreeOneNodes(N):
    '''Return list of all v with deg(v) = 1.'''
    return [v for v in N if N.degree(v) == 1]

def arcsOnPath(path):
    '''Return list of all arcs on path.'''
    return [(path[i], path[i+1]) for i in xrange(len(path)-1)]

def arcsOnPath_iter(path):
    '''Return a generator for all arcs on path.'''
    return ((path[i], path[i+1]) for i in xrange(len(path)-1))

def arcPairNetwork(N):
    '''Return N with additional arcs (w,v) for all (v,w) in N, u(w,v)=0'''
    revArcsN = N.copy()
    for v,w in N.edges_iter():
        assert not N.has_edge(w,v)
        revArcsN.add_edge(w, v, upper=0)
    return revArcsN

# def delNodeDicts(N, *args):
#     '''Delete for all nodes all dictionaries whose names are given in args.'''
#     for s,v in product(args,N):
#         if s in N.node[v]:
#             del N.node[v][s]

# def delArcDicts(N, *args):
#     '''Delete for all arcs all dictionaries whose names are given in args.'''
#     for s,a in product(args, N.edges_iter()):
#         v,w = a
#         if s in N[v][w]:
#             del N[v][w][s]

def removeBoundsEdgesForMaxFlow(N, edges):
    '''For arcs in "N" corresponding to "edges", remove upper bound, ie, set to
    infty, and add reversed arc without bound, ie, set lower bound to -infty.'''
    for e in edges:
        v,w = networkArcFromEdge(N, e)
        N.add_edge(w,v)
        del N.edge[v][w]['upper']

def networkArcFromEdge(N, e):
    '''Return the arc in N corresponding to the "undirected" e.'''
    if N.has_edge(e[0],e[1]):
        return e
    elif N.has_edge(e[1],e[0]):
        return (e[1],e[0])
    else:
        raise KeyError('Neither {} nor ({},{}) is an arc of the network.'.
                       format(e, e[1], e[0]))
        return 
        

