# -*- coding: utf-8 -*-
'''
Big-M MIP formulation for MIAC.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
import pulp
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil
from netmic.exceptions import ModelUnsolvedError, ModelNumericalError
import netmic.timer as timer
import netmic.heuristics.IISCover as IISC
import time
        
def solveViaMIP(N, lowBound=False, timelimit=1800):
    '''Return cover computed via exakt big-M-MIP.'''
    modelBuilder = bigMMip(N, lowBound, timelimit)
    try:
        C = modelBuilder.buildCover()
    except AssertionError:
        raise ModelNumericalError("MIP produced conflicting variable values")
    return C, round(modelBuilder.timeBuild,3), round(modelBuilder.timeSol,3)

    
class bigMMip():
    '''set up big M formulation of MIAC and keep data at hand'''
    def __init__(self, N, lowBound=False, timelimit=1800):
        self.N = N
        self.timelimit = timelimit
        self.timeSol = 0
        with timer.TimerCPU() as t:
            self.mip = pulp.LpProblem("MIAC", pulp.LpMinimize)
            self.buildVariables()
            self.buildModel()
            if lowBound:
                self.addLowerBound()
        self.timeBuild = t.secs
        self.timelimit -= self.timeBuild
        #self.mip.writeLP('mipStart.lp')

    def buildVariables(self):
        '''Set up variables and bounds'''
        self.x = pulp.LpVariable.dicts('x', self.N.edges())
        self.indL = pulp.LpVariable.dicts('ind_lower', self.N.edges(),
                                          cat="Binary")
        self.indU = pulp.LpVariable.dicts('ind_upper', self.N.edges(),
                                          cat="Binary")

    def buildModel(self):
        '''Add objective function and constraints to model'''
        self.mip += (pulp.lpSum(self.indL) + pulp.lpSum(self.indU),
                     'minimize indicator usage')
        # add constraints Ax = b
        for v in self.N:
            APlus = pulp.LpAffineExpression([(self.x[a],1) for a in
                                             self.N.out_edges_iter(v)])
            AMinus = pulp.LpAffineExpression([(self.x[a],-1) for a in
                                              self.N.in_edges_iter(v)])
            self.mip += APlus + AMinus == self.N.node[v]['supply'], ''
        # add constraints relaxed bounds
        bigM = netUtil.sumBPlus(self.N)
        for a in self.N.edges_iter():
            self.mip += self.x[a] + bigM * self.indL[a] >= 0, ''
            self.mip += (self.x[a] - (bigM - self.N.edge[a[0]][a[1]]['upper'])
                         * self.indU[a] <= self.N.edge[a[0]][a[1]]['upper'], '')

    def addLowerBound(self):
        lbBuilder = lowerBoundForSetCover(self.N)
        lb = lbBuilder.solve(self.timelimit/6)
        # add lower bound
        self.mip += pulp.lpSum(self.indL) + pulp.lpSum(self.indU) >= lb, ''
        # add known IISs
        for S in lbBuilder.IISListS:
            if netUtil.sumBInS(self.N, S) > 0:
                exprU = pulp.LpAffineExpression([(self.indU[a],1) for a in
                                                 netUtil.deltaPlus(self.N,S)])
                exprL = pulp.LpAffineExpression([(self.indL[a],1) for a in
                                                 netUtil.deltaMinus(self.N,S)])
            else:
                exprU = pulp.LpAffineExpression([(self.indU[a],1) for a in
                                                 netUtil.deltaMinus(self.N,S)])
                exprL = pulp.LpAffineExpression([(self.indL[a],1) for a in
                                                 netUtil.deltaPlus(self.N,S)])
            self.mip += exprU + exprL >= 1, ''
            
    def buildCover(self):
        '''Solve mip, build and return cover according to solution.'''
        try:
            with timer.TimerCPU() as t:
                self.mip.solve(pulp.CPLEX(msg=0,options=['set threads 1',
                                                         'set timeLimit {}'.
                                                         format(self.timelimit)]))
            self.timeSol = t.secs
        except pulp.solvers.PulpSolverError as e:
            if str(e) == "Unknown status returned by CPLEX: time limit exceeded":
                return None
            elif str(e) == ("Unknown status returned by CPLEX: optimal with "
                            "unscaled infeasibilities"):
                raise ModelUnsolvedError("MIP has unscaled infeasibilities.")
            else: raise 
    
        if self.mip.status < 1:
            raise ModelUnsolvedError("MIP could not be solved, status: {}.".
                                     format(pulp.constants.LpStatus[self.mip.status]))

        C = alUtil.emptySolutionDict(self.N)
        for a in self.N.edges():
            assert self.indL[a].varValue < 1e-4 or self.indL[a].varValue > 1-1e-4
            if self.indL[a].varValue > 1e-4:
                C[a[0]][a[1]] = 'l'
                assert self.x[a].varValue < 0
                
            assert self.indU[a].varValue < 1e-4 or self.indU[a].varValue > 1-1e-4    
            if self.indU[a].varValue > 1e-4:
                C[a[0]][a[1]] = 'u'
                assert self.x[a].varValue > self.N[a[0]][a[1]]['upper']

        assert abs(alUtil.sizeOfCover(C) - pulp.value(self.mip.objective)) < 1e-4
        return C
            
class lowerBoundForSetCover(IISC.IISCover):
    '''compute a lower bound for MIAC and a start set by computing disjunct IISs'''
    def __init__(self, N):
        IISC.IISCover.__init__(self, N)
        self.IISListS = []
        
    def solve(self, timelimit=300):
        '''Compute an IIS in current network, set some bounds to inf and repeat
        until there are no more IISs.'''
        starttime = time.clock()
        self.findIIS(True)
        while len(self.listOfS) > 0:  # as long as there are IISs
            for S in self.listOfS:
                self.IISListS.append(S)
                self._setBoundsToInf(netUtil.delta(self.N, S))
            if time.clock() - starttime > timelimit:
                break
            self.findIIS(True)

        return len(self.IISListS)
        
