# -*- coding: utf-8 -*-
'''
Optimal IIS Cover computation.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
from itertools import ifilterfalse
import pulp
import netmic.network_utility as netUtil
import netmic.heuristics.IISCover as IISC
from netmic.exceptions import ModelUnsolvedError
import netmic.timer as timer
import time

def solveViaIISSetCover(N, timelimit=1800):
    '''Return cover computed via exakt set cover algorithm.'''
    ICB = setCover(N, timelimit)
    ICB.solve()
    if ICB.timeOut:
        return None, ICB.solvedMips, round(ICB.timeCover,3), round(ICB.timeIIS,3)
    C,_ = ICB.coverFromSolution()    
    return C, ICB.solvedMips, round(ICB.timeCover,3), round(ICB.timeIIS,3)
    
class setCover(IISC.IISSetCover):
    '''solve the set cover formulation, first greedily, then with a mip'''
    def __init__(self, N, timelimit=1800):
        IISC.IISSetCover.__init__(self, N)
        self.timelimit = timelimit
        self.solvedMips = 0
        
    def solve(self):
        with timer.TimerCPU() as t:
            IISC.IISSetCover.solve(self, True, self.timelimit)
            if self.timeOut:
                return
        self.timelimit -= t.secs
        self._resetBoundsInitially()
        self._setUpMip()
        self._solveOptimally()

    def _resetBoundsInitially(self):
        '''Set all upper bounds in ssN and _R back to their original value.'''
        for v,w,d in ifilterfalse(lambda a: 'upper' in self.ssN[a[0]][a[1]],
                                  self.N.edges_iter(data=True)):
            # in ssN: set upper bound, remove reverse arc
            self.ssN[v][w]['upper'] = d['upper']
            self.ssN.remove_edge(w,v)
            # in R: set bound to u or 0, respectively
            self._R[v][w]['capacity'] = d['upper']
            self._R[w][v]['capacity'] = 0
                

    def _setUpMip(self):
        '''Initialize mip with a binary variable for every arc and the set cover
        constraints for the known IISs.'''
        self.mip = pulp.LpProblem("SetCover", pulp.LpMinimize)
        self.y = pulp.LpVariable.dicts('y', self.N.edges(), cat="Binary")
        self.mip += pulp.lpSum(self.y), 'minimize variables'
        for i in xrange(len(self.IISList)):
            expr = pulp.LpAffineExpression([(self.y[a],1) for a in
                                            self.IISList[i]['delta']])
            self.mip += expr >= 1, ''
            
    def _solveOptimally(self):
        '''Solve mip, build and return cover according to solution.'''
        starttime = time.clock()
        with timer.TimerCPU() as t:
            self.mip.solve(pulp.CPLEX(msg=0, options=['set timeLimit {}'.
                                                      format(self.timelimit/2)]))
        self.timeCover += t.secs
        if self.mip.status < 1:
            raise ModelUnsolvedError("Mip could not be solved, status: {}.".
                                     format(pulp.constants.LpStatus[self.mip.status]))
        self.solvedMips = 1
        arcs = [a for a in self.N.edges_iter() if self.y[a].varValue > 0]
        self._setBoundsToInf(arcs)
        with timer.TimerCPU() as t:
            self.findIIS(True)
        self.timeIIS += t.secs
        
        # as long as there are uncovered IISs
        while len(self.listOfS) > 0 and not self.timeOut:  
            self._resetBounds(arcs)
            self._updateMip()
            with timer.TimerCPU() as t:
                self.mip.resolve()
            self.timeCover += t.secs
            self.solvedMips += 1
            if self.mip.status < 1:
                raise ModelUnsolvedError("Mip could not be solved, status: {}.".
                                         format(pulp.constants.LpStatus[self.mip.status]))
            arcs = [a for a in self.N.edges_iter() if self.y[a].varValue > 0]
            self._setBoundsToInf(arcs)
            with timer.TimerCPU() as t:
                self.findIIS(True)
            self.timeIIS += t.secs
            if time.clock() - starttime > self.timelimit:
                self.timeOut = True
 
    def _updateMip(self):
        '''Add set cover constraint to mip for IISs in listOfS.'''
        for S in self.listOfS:
            delta = netUtil.delta(self.N, S)
            expr = pulp.LpAffineExpression([(self.y[a],1) for a in delta])
            self.mip += expr >= 1, ''
            
        
