#!/usr/bin/python2
# -*- coding: utf-8 -*-

import networkx as nx
import unittest
import pulp

from netmic.optimal.mip_model import *

class MIPModelTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=2)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)

    @unittest.skipUnless(pulp.CPLEX.available(pulp.CPLEX()),
                         "buildCover() requires CPLEX")
    def testSolveViaMipWithLB(self):
        cover,_,_ = solveViaMIP(self.G, True)
        self.assertEqual(cover, {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{},
                                 6:{5:'l'}})
        
    def testBuildModelWithLB(self):
        modelBuilder = bigMMip(self.G, True)
        self.assertEqual(modelBuilder.mip.numConstraints(), 22)

    def testBuildModel(self):
        modelBuilder = bigMMip(self.G)
        self.assertEqual(modelBuilder.mip.numVariables(), 18)
        self.assertEqual(modelBuilder.mip.numConstraints(), 18)
        self.assertEqual(modelBuilder.mip.getSense(), 1)
        self.assertEqual(sum(1 for v in modelBuilder.mip.variables() if
                             v.cat==pulp.LpInteger), 12)
        
    def testBuildModelWithSolve(self):
        modelBuilder = bigMMip(self.G)
        modelBuilder.mip.solve()
        self.assertEqual(modelBuilder.x[(6,5)].varValue, -1)
        self.assertEqual(modelBuilder.x[(3,5)].varValue, 2)
        self.assertEqual(modelBuilder.x[(3,4)].varValue, 1)
        self.assertEqual(pulp.value(modelBuilder.mip.objective), 3)

    @unittest.skipUnless(pulp.CPLEX.available(pulp.CPLEX()),
                         "buildCover() requires CPLEX")        
    def testBuildCover(self):
        modelBuilder = bigMMip(self.G)
        cover = modelBuilder.buildCover()
        self.assertEqual(cover, {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{},
                                 6:{5:'l'}})

    @unittest.skipUnless(pulp.CPLEX.available(pulp.CPLEX()),
                         "buildCover() requires CPLEX")        
    def testSolveViaMip(self):
        cover,_,_ = solveViaMIP(self.G)
        self.assertEqual(cover, {1:{}, 2:{3:'u'}, 3:{5:'u'}, 4:{}, 5:{},
                                 6:{5:'l'}})

class lowerBoundTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=1)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)

    def testLowerBound(self):
        builder = lowerBoundForSetCover(self.G)
        nr = builder.solve()
        self.assertEqual(nr, 3)
        self.assertEqual(builder.IISListS, [[1,2,3,4], [1,2], [1,2,3,4,5]])
        
        
if __name__ == '__main__':
    unittest.main()
