# -*- coding: utf-8 -*-
'''
Postprocessing heuristic Minimalizer.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
from networkx.algorithms.flow import preflow_push
from networkx.algorithms.flow import build_residual_network
from itertools import ifilter
from copy import deepcopy
import netmic.network_utility as netUtil
import time

def minimalize(N, C, timelimit=900):
    '''Return a minimal subcover of C.'''
    builder = minimalizer(N)
    builder.check(C, timelimit)
    return builder.C

class minimalizer():
    '''take a cover and return a minimal subcover'''
    def __init__(self, N):
        self.ssN = netUtil.setUpSingleSourceSinkNetwork(N, False)
        netUtil.removeInfiniteUpperBoundsForMaxFlow(self.ssN)
        self._R = build_residual_network(self.ssN, 'upper')
        self._addZeroBoundsToR()
        self.bPlus = netUtil.sumBPlus(N)
            
    def cleanUp(self):
        '''Set bounds in R back to original values.'''
        for v,w in ifilter(lambda a: a[0]!='s' and a[1]!='t', self.ssN.edges_iter()):
            self._R[w][v]['capacity'] = 0
            if 'upper' in self.ssN[v][w]:
                self._R[v][w]['capacity'] = self.ssN[v][w]['upper']

    def check(self, C, timelimit=900):
        '''Check greedily which constraints can be dropped without
        loosing feasibility.'''
        starttime = time.clock()
        self.C = deepcopy(C)
        self._removeCoverBoundsInitially()
        red = 0
        for v,w in ((v,w) for v in C for w in C[v]):
            # 'remove' bound from cover
            if C[v][w] == 'u':
                self._R[v][w]['capacity'] = self.ssN[v][w]['upper']
            elif C[v][w] == 'l':
                self._R[w][v]['capacity'] = 0
            # still feasible?
            self._R = preflow_push(self.ssN, 's', 't', 'upper', self._R,
                                   value_only=True)
            # infeasible, set bounds back to inf
            if self.bPlus - self._R.graph['flow_value'] > 1e-4:
                if C[v][w] == 'u':
                    self._R[v][w]['capacity'] = self._R.graph['inf']
                elif C[v][w] == 'l':
                    self._R[w][v]['capacity'] = self._R.graph['inf']
            # feasible
            else:
                del self.C[v][w]
                red += 1
            if time.clock() - starttime > timelimit:
                break
        return red

    def _addZeroBoundsToR(self):
        '''Add arcs with 0 upper bounds in ssN to R (these were ignored by nx).'''
        for v,w,_ in ifilter(lambda a: a[2]['upper'] == 0,
                             self.ssN.edges_iter(data=True)):
            self._R.add_edge(v,w, capacity=0)
            self._R.add_edge(w,v, capacity=0)

    def _removeCoverBoundsInitially(self):
        '''For all bounds in C, set bound in ssN to inf.'''
        for v,w in ((v,w) for v in self.C for w in self.C[v]):
            if self.C[v][w] == 'u':
                self._R[v][w]['capacity'] = self._R.graph['inf']
            elif self.C[v][w] == 'l':
                self._R[w][v]['capacity'] = self._R.graph['inf']



