# -*- coding: utf-8 -*-
'''
Negative Cycle with Forbidden Pairs heuristics.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
from collections import deque
from itertools import ifilter
import netmic.network_utility as netUtil
import time

def setUpAssignmentProblem(N):
    '''Return bipartite redraft of N, with a demand of -1 for every node in the
    set 0 and 1 for every node in set 1. All upper bounds are 1.'''
    Nbip = bipartizeAuxGraph(N)
    for v in N:
        Nbip.node[(v,0)]['demand'] = -1
        Nbip.node[(v,1)]['demand'] = 1
        for a in Nbip.edges_iter(data=True):  # add upper bounds, not necessary,
            a[2]['upper'] = 1                 # but helps speed up MCF algorithm
    return Nbip


def bipartizeAuxGraph(N, s=None, t=None):
    '''Return bipartite redraft of N. Nodes have the form (v,0) and (v,1).
    If s and t are given, (s,1) and (t,0) do not exist.'''
    Nbip = nx.DiGraph()
    # node copies have the form (v,0), (v,1)
    for v,w,data in ((v,w,data) for v,w,data in N.edges_iter(data=True)
                     if v!=t and w!=s):
        if 'weight' in data:
            Nbip.add_edge((v,0), (w,1), weight=data['weight'])
        else:
            Nbip.add_edge((v,0), (w,1))
    for v in ifilter(lambda v: v!=s and v!=t, N):
        Nbip.add_edge((v,0), (v,1), weight=0)
    return Nbip

def MCFForDisconnectedGraphs(N, demand='demand', capacity='capacity',
                             weight='weight'):
    '''Return a min cost flow solution. If N is disconnected, solve components
    independently and assemble the solutions.'''
    if nx.is_weakly_connected(N):
        return nx.capacity_scaling(N, demand, capacity, weight)
    
    x = dict()
    c = 0
    connectedGraphs = nx.weakly_connected_component_subgraphs(N, copy=True)
    for N_i in connectedGraphs:
        c_i,x_i = nx.capacity_scaling(N_i, demand, capacity, weight)
        c += c_i
        for v in x_i:
            assert v not in x
            x[v] = dict()
            for w in x_i[v]:
                x[v][w] = x_i[v][w]
    return c,x

class negativeCycleAssignmentProblem():
    '''Set up and solve the negative cycle with forbidden pairs problem as an
    assignment problem.'''
    def __init__(self, N):
        self.N = N
        self.Nbip = setUpAssignmentProblem(self.N)
        self.solMap = dict()

    def updateWeightsFromN(self):
        '''Copy the weights from N to the corresponding arc in Nbip.'''
        for v,w,d in self.N.edges_iter(data=True):
            self.Nbip[(v,0)][(w,1)]['weight'] = d['weight']
        
    def solveViaSimpleMCF(self):
        '''Return list of lists of all (found) negative cycles in N. Solved via MCF
        on the corresponding assignment problem, where potential pairs are broken in
        the trivial way, by taking them out.'''
        c,x = nx.capacity_scaling(self.Nbip, capacity='upper')
        if c >= 0:
            return []
        self._solMapFromBipartiteSolution(x)
        self._breakPairsInSolMap()
        cost = self._costFromSolMap()
        if cost >= 0:
            return []
        return self._cyclesFromSolMap()

    def solveViaMultipleMCF(self, timelimit=900):
        '''Return list of lists of all (found) negative cycles in N. Solved via
        MCF on the corresponding assignment problem, where potential pairs are
        broken by removing one arc and solving MCF again, until no pairs exist.'''
        starttime = time.clock()
        containsPairs = True
        while containsPairs:
            c,x = MCFForDisconnectedGraphs(self.Nbip, capacity='upper')
            if c >= 0:
                return []
            if time.clock() - starttime > timelimit:
                return []
            containsPairs = self._checkForPairs(x)

        self._solMapFromBipartiteSolution(x)
        return self._cyclesFromSolMap()
        

    def _solMapFromBipartiteSolution(self, x):
        '''Return a map of nodes to nodes, where v->w if x[v][w]=1.'''
        for v in self.N:
            for w,_ in x[(v,0)]:
                if x[(v,0)][(w,1)] == 1:
                    break
            else:
                raise ValueError('Node ({},0) has no adjacent node in solution.'.
                                 format(v))
            self.solMap[v] = w

    def _breakPairsInSolMap(self):
        '''If there is v->w, w->v in solMap, set v->v, w->w.'''
        for v in self.N:
            w = self.solMap[v]
            if self.solMap[w] == v:
                self.solMap[v] = v
                self.solMap[w] = w

    def _costFromSolMap(self):
        '''Return the cost of the mapping according to weight in Nbip.'''
        assert len(self.Nbip.nodes()) == 2*len(self.solMap)
        return sum(self.Nbip.edge[(v,0)][(self.solMap[v],1)]['weight']
                   for v in self.solMap) 

    def _costFromCycleList(self, cycle):
        '''Return cost of a cycle according to weight in N.'''
        return sum(self.N.edge[cycle[i]][cycle[i+1]]['weight']
                    for i in range(len(cycle)-1))


    def _cyclesFromSolMap(self):
        '''Return list of lists of all negative cycles in solMap. Assumes solMap
        is a perfect matching.'''
        cycles = []
        while self.solMap != {}:
            v = self.solMap.keys()[0]
            if v == self.solMap[v]:
                del self.solMap[v]
                continue
            w = self.solMap.pop(v)
            cycle = [v,w]
            while w != v:
                w = self.solMap[w]
                del self.solMap[cycle[len(cycle)-1]]
                cycle.append(w)
            assert cycle[0] == cycle[len(cycle)-1]
            if self._costFromCycleList(cycle) < 0:
                cycles.append(cycle)
        return cycles


    def _checkForPairs(self, x):
        '''Return False if there is no mapping v->w, w->v according to x.
        Otherwise return True and for the first found pair remove the more
        expensive corresponding arc in Nbip.'''
        for v in ifilter(lambda v: v[1]==0, self.Nbip.nodes_iter()):
            for w in x[v]:
                if x[v][w] == 1:
                    break
            else:
                raise ValueError('Node ({},0) has no adjacent node in solution.'.
                                 format(v))
            if v[0] == w[0]:  # v is self assigned, all good
                continue
            if (v[0],1) in x[(w[0],0)] and x[(w[0],0)][(v[0],1)] == 1:
                if (self.Nbip.edge[v][w]['weight'] >
                    self.Nbip.edge[(w[0],0)][(v[0],1)]['weight']):
                    self.Nbip.remove_edge(v,w)
                else:
                    self.Nbip.remove_edge((w[0],0),(v[0],1))
                return True
        return False


class negativeCycleBellmanFordProblem():
    '''Set up and solve the negative cycle with forbidden pairs problem with a
    modified Bellman-Ford algorithm.
    assignment problem.'''
    def __init__(self, N):
        self.N = N

    def solve(self, timelimit=900):
        '''Return list of list of (found) negative cycle in N. Solved via the
        (heuristic) Bellman-Ford with Forbidden Pairs, test all nodes as start node,
        break if a cycle is found.'''
        starttime = time.clock()
        for v in self.N:
            cycle = self.bellmanFordWithForbiddenPairs(v)
            if len(cycle) > 0:
                return [cycle]
            if time.clock() - starttime > timelimit:
                break
        return []


    def bellmanFordWithForbiddenPairs(self, s):
        '''Return found negative cycle in N as list. Solved via the
        (heuristic) Bellman-Ford with Forbidden Pairs, with start node s.'''
        nodeWeights = {v: float('inf') for v in self.N.nodes_iter()}
        nodeWeights[s] = 0
        pred = {v: None for v in self.N.nodes_iter()}
        blocked = {a: False for a in self.N.edges_iter()}
        count = 1
        while nodeWeights[s] >= 0 and count < len(self.N): 
            for v,w,d in self.N.edges_iter(data=True):
                if blocked[(w,v)]:   # reversed arc was already used
                    continue
                if (nodeWeights[w] > nodeWeights[v] + d['weight']):
                    if pred[w] != None:
                        assert blocked[(pred[w],w)]
                        blocked[(pred[w],w)] = False
                    pred[w] = v
                    blocked[(v,w)] = True
                    nodeWeights[w] = nodeWeights[v] + d['weight']
            count += 1
        if nodeWeights[s] < 0:
            return self._constructCycle(s, pred)

        # n-th call of Bellman; if nodeWeights[w] can be improved, w lays on
        # a negative cycle
        for v,w,d in self.N.edges_iter(data=True):
            if blocked[(w,v)]:   # reversed arc was already used
                continue
            if (nodeWeights[w] > nodeWeights[v] + d['weight']):
                assert pred[w] != None
                pred[w] = v
                return self._constructCycle(w, pred)
        return []
        
    
    def _constructCycle(self, s, pred):
        '''Return cycle constructed from 'pred'. Assumes weight of s < 0. Cycle
        either contains s or lays on the path to s.'''
        w = pred[s]
        path = deque([w,s])
        while path.count(path[0]) == 1:
            v = pred[w]
            assert self.N.has_edge(v,w)
            path.appendleft(v)
            w = v
        # cycle does not necessarily close at s, remove path from cycle to s
        path = list(path)
        j = path.index(path[0], 1)
        return path[:j+1]

                   
