# -*- coding: utf-8 -*-
'''
Postprocessing heuristic Refinement.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx

import netmic.postprocessing.negative_cycle_heuristics as negCycle
from netmic.heuristics.weight_functions import computeArcWeightsDetailedTernary
import netmic.algorithmic_utility as alUtil
import netmic.network_utility as netUtil
import time


def refineSolution(N, x, func='simpleAss', timelimit=900):
    '''Implementation of Negative-Cycle-Refinement-Heuristic.
    
    Check for cycle flows such that afterwards fewer arc bounds are violated.
    Utilizes ternary weights as in heuristic 1c. The checked flow amounts are
    r = -x_a, u_a - x_a for every arc a.

    Return True if solution was improved
    '''
    starttime = time.clock()
    revArcsN = netUtil.arcPairNetwork(N)
    refined = False
    checkedValues = [0]
    # algorithm runs once for every arc, independently from intermediate changes 
    for v,w in N.edges_iter():
        for r in [abs(x[v][w]), abs(x[v][w] - N.edge[v][w]['upper'])]:
            if r in checkedValues:
                continue
            checkedValues.append(r)
            computeArcWeightsDetailedTernary(N, revArcsN, x, r)
            cycles = negCycles(revArcsN, func, timelimit-time.clock()+starttime)
            for cycle in cycles:
                alUtil.changeSolutionAlongPath(N, x, cycle, r)
            if cycles:
                refined = True
                checkedValues = [0]  # delete list of checked values
        if time.clock() - starttime > timelimit:
            break

    return refined


def negCycles(N, func, timelimit=900):
    if func == 'simpleAss':
        assignProblem = negCycle.negativeCycleAssignmentProblem(N)
        cycles = assignProblem.solveViaSimpleMCF()
    elif func == 'multiAss':
        assignProblem = negCycle.negativeCycleAssignmentProblem(N)
        cycles = assignProblem.solveViaMultipleMCF(timelimit)
    elif func == 'bellman':
        BFProblem = negCycle.negativeCycleBellmanFordProblem(N)
        cycles = BFProblem.solve(timelimit)
    else:
        raise NotImplementedError('Requested method {} does not exist'.
                                  format(func))
    return cycles
