#!/usr/bin/python2
# -*- coding: utf-8 -*-

import networkx as nx
import unittest
import netmic.algorithmic_utility as alUtil

from netmic.postprocessing.negative_cycle_heuristics import *

class bipartizeTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_nodes_from(range(5))
        self.G.add_edge(1, 2, weight=0)
        self.G.add_edge(2,1, weight=1)
        self.G.add_edge(1, 3, weight=0)
        self.G.add_edge(3,1, weight=-1)
        self.G.add_edge(2, 3, weight=1)
        self.G.add_edge(3,2, weight=0)
        self.G.add_edge(3, 4, weight=0)
        self.G.add_edge(4,3, weight=0)
        self.G.add_edge(3, 5, weight=0)
        self.G.add_edge(5,3, weight=-1)
        self.G.add_edge(0, 5, weight=0)
        self.G.add_edge(5,0, weight=1)

    def testBipartizeAuxGraphWithSAndT(self):
        Gbip = bipartizeAuxGraph(self.G, 2, 0)
        sumWeights = sum(a[2]["weight"] for a in Gbip.edges(data=True))
        self.assertEqual(sumWeights, 1)
        self.assertEqual(Gbip[(5,0)][(3,1)]['weight'], -1)
        self.assertFalse(Gbip.has_node((2,1)) or Gbip.has_node((0,0)))
        self.assertEqual(len(Gbip.edges()), 13)

    def testBipartizeAuxGraph(self):
        Gbip = bipartizeAuxGraph(self.G)
        for v in self.G:
            self.assertTrue(Gbip.has_node((v,0)) and Gbip.has_node((v,1)))

    def testSetUpAssignmentProblem(self):
        Gbip = setUpAssignmentProblem(self.G)
        self.assertEqual(sum(Gbip.node[(v,0)]['demand'] for v in
                             self.G.nodes_iter()), -6)
        self.assertEqual(sum(Gbip.edge[v][w]['upper'] for v,w in Gbip.edges()),
                         18)

    def testCostFromSolMap(self):
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        negCycleBuilder.solMap = {0:5, 5:3, 3:4, 2:3, 1:2, 4:3}
        self.assertEqual(negCycleBuilder._costFromSolMap(), 0)

    def testCostFromCycleList(self):
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        self.assertEqual(negCycleBuilder._costFromCycleList([1,3,2,1]), 1)

    def testAssignmentViaMCFSimple(self):
        self.G.edge[1][2]['weight'] = -1
        self.G.edge[2][3]['weight'] = 0
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        self.assertEqual(negCycleBuilder.solveViaSimpleMCF(), [[1,2,3,1]])

    def testAssignmentViaMultipleMCF(self):
        self.G.edge[1][2]['weight'] = -1
        self.G.edge[2][3]['weight'] = 0
        self.G.edge[0][5]['weight'] = -1
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        self.assertEqual(negCycleBuilder.solveViaMultipleMCF(), [[1,2,3,1]])

    def testUpdateWeights(self):
        del self.G[2][3]['weight']
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        for v in self.G:
            self.assertEqual(negCycleBuilder.Nbip[(v,0)][(v,1)]['weight'], 0)
        self.assertFalse('weight' in negCycleBuilder.Nbip[(2,0)][(3,1)])
        negCycleBuilder.N[2][3]['weight'] = 5
        negCycleBuilder.updateWeightsFromN()
        self.assertEqual(sum(d['weight'] for v,w,d in
                             negCycleBuilder.Nbip.edges_iter(data=True)), 5)
        

class breakSolutionTests2(unittest.TestCase):
    def setUp(self):        
        self.G = nx.DiGraph()
        self.G.add_edges_from([(0,1), (1,0), (0,2), (2,1), (2,3), (3,2)], weight=0)
        self.x = dict()
        for v in self.G:
            self.x[(v,0)] = dict()
            self.x[(v,1)] = dict()
        for v,w in self.G.edges_iter():
            self.x[(v,0)][(w,1)] = 0
        self.x[(0,0)][(1,1)] = 1
        self.x[(1,0)][(0,1)] = 1
        self.x[(2,0)][(3,1)] = 1
        self.x[(3,0)][(2,1)] = 1

    def testSolMapFromSolution(self):
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        negCycleBuilder._solMapFromBipartiteSolution(self.x)
        self.assertEqual(negCycleBuilder.solMap, {0:1, 1:0, 2:3, 3:2})
        
    def testBreakPairs(self):
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        negCycleBuilder.solMap = {0:1, 1:0, 2:3, 3:2}
        negCycleBuilder._breakPairsInSolMap()
        self.assertEqual(negCycleBuilder.solMap, {0:0, 1:1, 2:2, 3:3})

    def testCheckForPairsTrue(self):
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        negCycleBuilder.Nbip.edge[(0,0)][(1,1)]['weight'] = 0
        negCycleBuilder.Nbip.edge[(1,0)][(0,1)]['weight'] = -1
        self.assertTrue(negCycleBuilder._checkForPairs(self.x))
        self.assertFalse(negCycleBuilder.Nbip.has_edge((0,0),(1,1)))

    def testCheckForPairsError(self):
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        self.x[(0,0)][(1,1)] = 0
        with self.assertRaises(ValueError):
            negCycleBuilder._checkForPairs(self.x)

    def testCheckForPairsFalse(self):
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        self.x[(0,0)][(1,1)] = 0
        self.x[(0,0)][(2,1)] = 1
        self.x[(3,0)][(2,1)] = 0        
        self.x[(3,0)][(1,1)] = 1
        self.assertFalse(negCycleBuilder._checkForPairs(self.x))
        self.assertTrue(len(negCycleBuilder.Nbip.edges()), 11)
        
    def testCheckForPairsSelfAssigned(self):
        negCycleBuilder = negativeCycleAssignmentProblem(self.G)
        x = dict()
        for v in self.G:
            x[(v,0)] = dict()
            x[(v,1)] = dict()
        for v,w in self.G.edges_iter():
            x[(v,0)][(w,1)] = 0
        x[(0,0)][(0,1)] = 1
        x[(1,0)][(1,1)] = 1
        x[(2,0)][(2,1)] = 1
        x[(3,0)][(3,1)] = 1
        self.assertFalse(negCycleBuilder._checkForPairs(x))
        self.assertTrue(len(negCycleBuilder.Nbip.edges()), 11)
        
class negativeCycleAuxFunctionsTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_edge(1,2, weight=1)
        self.G.add_edge(2,3, weight=0)
        self.G.add_edge(4,5, weight=1)
        self.G.add_edge(5,6, weight=-1)
        self.G.add_edge(3,1, weight=-1)
        self.G.add_edge(6,4, weight=-1)
        self.builder = negativeCycleAssignmentProblem(self.G)

    def testCyclesFromSolMap1(self):
        self.builder.solMap = {0:0, 1:2, 2:3, 4:5, 5:6, 3:1, 6:4}
        cycles = self.builder._cyclesFromSolMap()
        self.assertEqual(cycles, [[4,5,6,4]])

    def testCyclesFromSolMap2(self):
        self.builder.N.edge[1][2]['weight'] = 0
        self.builder.solMap = {0:0, 1:2, 2:3, 4:5, 5:6, 3:1, 6:4}
        cycles = self.builder._cyclesFromSolMap()
        self.assertEqual(cycles, [[1,2,3,1], [4,5,6,4]])
         
class MCFTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, demand=-3)
        self.G.add_node(2, demand=2)
        self.G.add_node(3, demand=1)
        self.G.add_node(4, demand=-1)
        self.G.add_node(0, demand=-1)
        self.G.add_node(6, demand=2)
        self.G.add_edge(1, 2, weight=0, upper=3)
        self.G.add_edge(1, 3, weight=1, upper=1)
        self.G.add_edge(2, 3, weight=-1, upper=1)
        self.G.add_edge(2, 1, weight=1)
        self.G.add_edge(4,6, weight=1, upper=2)
        self.G.add_edge(0,6, weight=1, upper=2)

    def testMCFForDisconnectedGraphs(self):
        x = alUtil.emptySolutionDict(self.G)
        x[1][2] = 3
        x[2][3] = 1
        x[4][6] = 1
        x[0][6] = 1
        alUtil.completeSolution(self.G, x)
        self.assertEqual(MCFForDisconnectedGraphs(self.G), (1,x))

class negativeCycleTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(0, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_edge(0, 1, weight=0)
        self.G.add_edge(1, 2, weight=0)
        self.G.add_edge(1, 3, weight=0)
        self.G.add_edge(2, 3, weight=0)
        self.G.add_edge(3, 4, weight=-1)       
        self.G.add_edge(4, 2, weight=0)
        self.G.add_edge(4, 5, weight=1)
        self.G.add_edge(2, 1, weight=-1)
        self.G.add_edges_from([(1,0), (3,1), (4,3), (5,4), (2,4), (3,2)],
                              weight=0)
        self.builder = negativeCycleBellmanFordProblem(self.G)

    def testConstructCycleWithS(self):
        pred = {2:4, 3:2, 4:3}
        self.assertEqual(self.builder._constructCycle(4,pred), [4,2,3,4])
        
    def testConstructCycleWithoutS(self):
        pred = {1:3, 2:4, 3:2, 4:3}
        self.assertEqual(self.builder._constructCycle(1,pred), [3,4,2,3])

    def testBellmanFord(self):
        self.assertEqual(self.builder.bellmanFordWithForbiddenPairs(1),
                         [2,3,4,2])
        
    def testBellmanFord2(self):
        self.builder.N.edge[2][3]['weight'] = 1
        self.assertEqual(self.builder.bellmanFordWithForbiddenPairs(1),
                         [1,3,4,2,1])

    def testBellmanFordWrapper(self):
        self.assertEqual(self.builder.solve(), [[1,3,4,2,1]])
                                 
if __name__ == '__main__':
    unittest.main()
