# -*- coding: utf-8 -*-
'''
Different method to display a cover.
'''
#   Copyright (C) 2015 by
#   Imke Joormann <joormann@mathematik.tu-darmstadt.de>
#   All rights reserved.
#   BSD license.

import networkx as nx
import matplotlib.pyplot as plt

def listOfUpperViolatedArcs(C):
    '''Return list of arc for which the upper bound is in the cover.'''
    return [(v,w) for v in C for w in C[v] if C[v][w] == 'u']

def listOfLowerViolatedArcs(C):
    '''Return list of arc for which the lower bound is in the cover.'''
    return [(v,w) for v in C for w in C[v] if C[v][w] == 'l']
    
        
def terminalPrint(uList, lList):
    '''Print list of contained bounds to terminal.'''
    size = len(uList) + len(lList)
    print "Cover size: {}".format(size)
    for (v,w) in uList:
        print "u({},{})".format(v,w)
    for (v,w) in lList:    
        print "l({},{})".format(v,w)

def filePrintInternal(uList, lList, filename):
    '''Print list of contained bounds to file.'''
    with open(filename, 'w') as f:
        for (v,w) in uList:
            f.write("u({},{})\n".format(v,w))
        for (v,w) in lList:    
            f.write("l({},{})\n".format(v,w))
                
def filePrint(C, filename):
    '''Print all contained bounds to file.'''
    uList = listOfUpperViolatedArcs(C)
    lList = listOfLowerViolatedArcs(C)
    print "Cover size: {}".format(len(uList) + len(lList))
    filePrintInternal(uList, lList, filename)
    
def coverToFeasibleNetwork(N, uList, lList):
    '''Set bounds in C to infty in N.'''
    for (v,w) in uList:
        N[v][w]['upper'] = float('inf')
    for (v,w) in lList:
        N[v][w]['lower'] = float('-inf')

def writeCoveredNetwork(N, uList, lList, filename):
    '''Write the feasible network, originated from N by covering bounds, to
    filename.'''
    feasN = N.copy()
    coverToFeasibleNetwork(feasN, uList, lList)
    nx.write_graphml(feasN, filename)
    
def drawGraphWithColouredEdges(N, filename, uList=[], lList=[]):
    '''Draw network N, with the arcs in C marked red and blue, respectively.'''
    pos = nx.spring_layout(N)
    nx.draw_networkx_nodes(N, pos, node_color='0.8',
                           label="nodes with supply value")
    nx.draw_networkx_labels(N, pos)
    nx.draw_networkx_edges(N, pos, label="arcs with bounds")
    nx.draw_networkx_edges(N, pos, uList, width=3, alpha=0.4, edge_color='r',
                           label="upper bound in cover")
    nx.draw_networkx_edges(N, pos, lList, width=3, alpha=0.4, edge_color='b',
                           label="lower bound in cover")
    nx.draw_networkx_edge_labels(N, pos, alpha=0.2, font_size=7)
    # node labels
    for v in N.nodes():
        x,y = pos[v]
        plt.text(x,y+0.05,s=N.node[v]['supply'], horizontalalignment='center',
                 fontsize=11)
    plt.axis('off')
    plt.legend(fancybox=True, prop={'size':'small'}, loc='upper right',
               markerscale=0.5)
    plt.savefig(filename)
    plt.show()

    
def outputAll(N, C, filename):
    '''Call all output functions, png and terminalPrint only if small enough.'''
    uList = listOfUpperViolatedArcs(C)
    lList = listOfLowerViolatedArcs(C)
    filePrintInternal(uList, lList, filename)
    writeCoveredNetwork(N, uList, lList, filename + '.graphml')
    if len(uList) + len(lList) <= 50:
        terminalPrint(uList, lList)
    if N.number_of_nodes() <= 50:
        drawGraphWithColouredEdges(N, filename + '.png', uList, lList)

    


