import networkx as nx
from networkx.algorithms.flow import preflow_push
import unittest
import netmic.network_utility as netUtil

from netmic.algorithmic_utility import *

class solutionTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=2)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)
        self.x1 = {1: {2: 2, 3: 1}, 3: {4: 0}}
        self.x2 = {1: {3: 3}, 2: {3: 1}, 3: {4: 1, 5: 1}}
        self.x3 = {1: {2: 2, 3: 1}, 2: {3: 4}, 3: {4: 1, 5: 2}, 6: {5: -1}}
        completeSolution(self.G, self.x1)
        completeSolution(self.G, self.x2)
        completeSolution(self.G, self.x3)

        
    def testCompleteSolution(self):
        sumX = sum(self.x1[a[0]][a[1]] for a in self.G.edges())
        self.assertEqual(sumX, 3)
        self.assertFalse(6 in self.x3[5])

    def testBoundFeasible(self):
        self.assertTrue(xBoundFeasible(self.G, self.x1))
        self.assertFalse(xBoundFeasible(self.G, self.x3))

    def testflowExcessAtNode(self):
        self.assertEqual(flowExcessAtNode(self.G, self.x1, 2), 4)
        self.assertEqual(flowExcessAtNode(self.G, self.x1, 3), -1)

    def testBFlowFeasible(self):
        self.assertFalse(xBFlowFeasible(self.G, self.x1))        
        self.assertTrue(xBFlowFeasible(self.G, self.x3))

    def testUnsaturatedSets(self):
        self.assertListEqual(unsaturatedSets(self.G, self.x1), [[2],[3,4,5,6]])
        self.assertListEqual(unsaturatedSets(self.G, self.x3), [[],[]])

    def testEmptySolutionDict(self):
        self.assertDictEqual(emptySolutionDict(self.G),
                             {1:{}, 2:{}, 3:{}, 4:{}, 5:{}, 6:{}})

    def testCoverFromSolution(self):
        C = emptySolutionDict(self.G)
        C[2][3] = "u"
        C[3][5] = "u"
        C[6][5] = "l"
        self.assertDictEqual(coverFromSolution(self.G, self.x3), C)

    def testCoverFromSolutionUndirected(self):
        x = {3: {1:1, 4:3}, 2: {3:1}}
        C = emptySolutionDict(self.G)
        C[1][3] = 'l'
        C[3][4] = 'u'
        C2 = coverFromSolutionUndirected(self.G, x, [(4,3), (1,3), (2,3)])
        self.assertEqual(C, C2)

    def testCoverFromSolutionResidual(self):
        x = {3: {1:1, 4:3}, 2: {3:1}}
        completeSolution(self.G, x)
        C = emptySolutionDict(self.G)
        C[1][3] = 'l'
        C[3][4] = 'u'
        self.assertEqual(coverFromSolutionResidual(self.G,x), C)

    def testSolutionFromResidualSolution(self):
        x = {3: {1:1, 4:3, 't':10}, 2: {3:1}}
        ssN = netUtil.setUpSingleSourceSinkNetwork(self.G, False)
        completeSolution(ssN, x)
        y = emptySolutionDict(self.G)
        y[1][3] = -1
        y[3][4] = 3
        y[2][3] = 1
        completeSolution(self.G, y)
        solutionFromResidualSolution(self.G,x)
        self.assertEqual(x, y)

    def testSolutionFromResidualGraph(self):
        R = preflow_push(netUtil.setUpSingleSourceSinkNetwork(self.G, False),
                         's', 't', 'upper', value_only=True)
        x = {1:{3:2}, 2:{3:1}, 3:{4:1}}
        completeSolution(self.G, x)
        self.assertEqual(solutionFromResidualGraph(self.G,R), x)
        

    def testComputePathWeight(self):
        auxG = netUtil.arcPairNetwork(self.G)
        auxG.edge[3][5]['weight'] = 0
        auxG.edge[5][6]['weight'] = 1
        self.assertEqual(computePathWeight(auxG, [3, 5, 6]), 1)

    def testChangeSolutionAlongPath(self):
        changeSolutionAlongPath(self.G, self.x2, [2, 1, 3, 5], 2)
        flow = sum(self.x2[a[0]][a[1]] for a in self.G.edges())
        self.assertEqual(flow, 8)
        with self.assertRaises(nx.NetworkXNoPath):
            changeSolutionAlongPath(self.G, self.x2, [2, 4], 1)            

    def testEmptyMap(self):
        solMap = {1: None, 2: None, 3: None, 4: None, 5: None, 6: None}
        self.assertDictEqual(emptyMap(self.G), solMap)

class shortestPathTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_edge(1, 2, weight=1)
        self.G.add_edge(1, 3, weight=2)
        self.G.add_edge(2, 3, weight=2)
        self.G.add_edge(3, 4, weight=1)
        self.G.add_edge(3, 5, weight=1)
        self.G.add_edge(6, 5, weight=1)
        self.G.add_edge(2, 1, weight=2)
        self.G.add_edge(3, 1, weight=1)
        self.G.add_edge(3, 2, weight=1)
        self.G.add_edge(4, 3, weight=1)
        self.G.add_edge(5, 3, weight=1)
        self.G.add_edge(5, 6, weight=2)

    def testComputeShortestPathFloyd(self):
        self.assertEqual(computeShortestPathFloyd(self.G, [1,2], [5,6]), (3, [1,3,5]))

    def testComputeFloyd(self):
        self.assertEqual(computeFloydWarshall(self.G, [1,2,3], [4,6]),
                         (1, [3,4]))
        
    def testComputeShortestPathSingleSource(self):
        self.assertEqual(computeShortestPathSingleSource(self.G, [1,2,3], [4,6]),
                         (1, [3,4]))

    def testComputeShortestPathBidirectional(self):
        self.assertEqual(computeShortestPathBidirectional(self.G, [6,1], [4]),
                         (3, [6,5,3,4]))
        
    def testComputeShortestPath(self):
        self.assertEqual(computeShortestPath(self.G, [1,2], [5,6]), (3, [1,3,5]))

class CoverManipulationTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)        
        self.C1 = emptySolutionDict(self.G)
        self.C1[1][2] = "l"
        self.C1[2][3] = "u"
        self.C2 = emptySolutionDict(self.G)
        self.C2[2][3] = "u"
        self.C2[3][4] = "u"

    def testJoiningCovers(self):
        joinCovers(self.C1, self.C2)
        self.assertDictEqual(self.C1, {1: {2: "l"}, 2: {3: "u"},
                                       3: {4: "u"}, 4: {}})

    def testJoinContradictingCovers(self):
        self.C2[1][2] = "u"
        with self.assertRaises(ValueError):
            joinCovers(self.C1, self.C2)

    def testJoinCoversOnDifferentNetworks(self):
        #due to preprocessing, it may be possible that not all v in C2 are in C1
        self.C2[0] = {3: 'l', 1: 'u'}
        joinCovers(self.C1, self.C2)
        self.assertDictEqual(self.C1, {0: {3: 'l', 1: 'u'}, 1: {2: "l"},
                                       2: {3: "u"}, 3: {4: "u"}, 4: {}})

    def testSizeOfCover(self):
        self.assertEqual(sizeOfCover(self.C1), 2)
        self.assertEqual(sizeOfCover(emptySolutionDict(self.G)), 0)

    def testSizeOfCoverNone(self):
        C = {'x': {}, 2: {5: 3}, 1: {'d': None}}
        self.assertEqual(sizeOfCover(C), 1)
        
class UndirectedCoverTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(1, 4, upper=1)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=1)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)
        self.tree = [(3, 5), (1, 4), (2, 3), (3, 4), (5, 6)]
        self.x = {1: {2: 2, 3: 2, 4: 0}, 2: {3: 4}, 3: {2: 0, 4: 2, 5: 2, 't': 2},
                  4: {1: 1, 3: 0, 't': 1}, 5: {3: 0, 't': 1, 6: 1},
                  6: {5: 0, 't': 1}, 's': {1: 3, 2: 2}, 't': {}}

    def testTreeHeuristicFailure(self):
        C = emptySolutionDict(self.G)
        C[2][3] = 'u'
        C[3][4] = 'u'
        C[3][5] = 'u'
        C[6][5] = 'l'
        C[1][4] = 'l'
        self.assertEqual(coverFromSolutionUndirected(self.G,self.x,self.tree), C)
        
if __name__ == '__main__':
    unittest.main()

