#!/usr/bin/python2
# -*- coding: utf-8 -*-

import networkx as nx
import unittest
from netmic.exceptions import DataError

from netmic.graphml_checker import *

class DefaultValueTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.MultiDiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3)
        self.G.add_edge(1, 1, lower=1)
        self.G.add_edge(1, 2, lower=1, upper=2)
        self.G.add_edge(1, 2)
        self.G.add_edge(3, 2, lower=0, upper=3)
        addDefaultValues(self.G)
        
    def testDefaultValues(self):
        sumSupply = sum(v[1]["supply"] for v in self.G.nodes(data=True))
        self.assertEqual(sumSupply, 5)
        sumBounds = sum(data["upper"] for v,w,data in self.G.edges(data=True))
        self.assertEqual(sumBounds, 5)

    def testCheckIntTrue(self):
        self.assertTrue(areValuesInt(self.G))

    def testCheckIntNode(self):
        self.G.node[3]['supply'] = 5.3
        self.assertFalse(areValuesInt(self.G))

    def testCheckIntNodeInf(self):
        self.G.node[3]['supply'] = float('-inf')
        with self.assertRaises(DataError):
            areValuesInt(self.G)

    def testCheckIntArc(self):
        self.G[3][2][0]['upper'] = 5.3
        self.assertFalse(areValuesInt(self.G))

    def testInfinityArcs(self):
        self.G[3][2][0]['upper'] = float('-inf')
        self.assertTrue(areValuesInt(self.G))

class SimpleGraphTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.MultiDiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3)
        self.G.add_edge(1, 1, lower=1, upper=2)
        self.G.add_edge(1, 2, lower=1, upper=2)
        self.G.add_edge(1, 2, upper=3)
        self.G.add_edge(2, 2, lower=1, upper=2)
        self.G.add_edge(3, 2, lower=0, upper=3)
        self.G.add_edge(2, 3, lower=0, upper=3)

    def testIsMultiGraphTrue(self):
        self.assertTrue(isMultiGraph(self.G))
        
    def testRemoveLoops(self):
        removeLoops(self.G)
        self.assertEqual(len(self.G.edges()), 4)
        self.assertTrue(not self.G.has_edge(1,1) and not self.G.has_edge(2,2))

    def testParallelEdges(self):
        diG = convertToSimpleGraph(self.G)
        self.assertEqual(len(diG.edges()), 3)
        self.assertEqual(diG.edge[1][2]["upper"], 5)

class ConsistencyTests(unittest.TestCase):
    def testExitUnconnected(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3)
        self.G.add_edge(1, 2, upper=3)
        
        with self.assertRaises(DataError):
            graphmlConsistency(self.G)

    def testExitUndirected(self):
        self.G = nx.Graph()
        with self.assertRaises(DataError):
            graphmlConsistency(self.G)
        
class RenamingTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node("s", supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3)
        self.G.add_node("t")        
        self.G.add_edge("s", 2, upper=3)
        self.G.add_edge(3, 2, lower=0, upper=3)
        self.G.add_edge(2, 3, lower=0, upper=3)
        self.G.add_edge("t", "s", upper=3)        

    def testNameConvention(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            checker = graphmlConsistency(self.G)
        self.assertTrue(checker.newN.has_edge(4,2))
        self.assertEqual(checker.newN[5][4]["upper"], 3)

    def testIsMultiGraphFalse(self):
        self.assertFalse(isMultiGraph(self.G))

    def testReverseNaming(self):
        checker = graphmlConsistency(self.G)
        N = checker.newN
        C = {4:{5:'l', 2:'u'}, 2:{4:'parrot'}, 3:{5:'10000 Volt'},
             5:{4:'vroom', 2:'l'}}
        checker.reverseNameConvention(C)
        self.assertEqual(C, {'s':{'t':'l', 2:'u'}, 2:{'s':'parrot'},
                             3:{'t':'10000 Volt'}, 't':{'s':'vroom', 2:'l'}})
        self.assertTrue(N.has_node('s') and N.has_node('t'))

###############################################################################        
if __name__ == '__main__':
    unittest.main()
