#!/usr/bin/python2
# -*- coding: utf-8 -*-

import networkx as nx
import unittest
from netmic.exceptions import DataError
from netmic.network_utility import *

class GHViolatedTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-2)
        self.G.add_node(4, supply=-1)
        self.G.add_node(5, supply=-1)
        self.G.add_node(6, supply=-1)
        self.G.add_edge(1, 2, upper=2)
        self.G.add_edge(1, 3, upper=2)
        self.G.add_edge(2, 3, upper=1)
        self.G.add_edge(3, 4, upper=2)
        self.G.add_edge(3, 5, upper=0)
        self.G.add_edge(6, 5, upper=2)

    def testGHSupplyTrue(self):
        self.assertTrue(GHSupplyViolated(self.G, [1,2,4]))
    def testGHSupplyFalse(self):
        self.assertFalse(GHSupplyViolated(self.G, [2,3,4]))
    def testGHDemandTrue(self):
        self.assertTrue(GHDemandViolated(self.G, [3,4,5,6]))
    def testGHDemandFalse(self):
        self.assertFalse(GHDemandViolated(self.G, [5]))
    def testGHFalse(self):
        self.assertFalse(GHViolated(self.G, [2,3,5,6]))

    def testArcPairNetwork(self):
        auxG = arcPairNetwork(self.G)
        self.assertListEqual(auxG.edges(),
                             [(1,2), (1,3), (2,1), (2,3), (3,1), (3,2), (3,4),
                              (3,5), (4,3), (5,3), (5,6), (6,5)])

    def testRemoveBoundsEdges(self):
        removeBoundsEdgesForMaxFlow(self.G, [(1,2), (2,3), (5,6)])
        self.assertTrue(self.G.has_edge(5,6) and self.G.has_edge(2,1))
        self.assertFalse('upper' in self.G.edge[2][3] or 'upper' in
                         self.G.edge[3][2])

    def testRemoveBoundsEdgesError(self):
        with self.assertRaises(KeyError):
            removeBoundsEdgesForMaxFlow(self.G, [(1,2), (2,3), (3,6)])

    def testArcFromEdgeError(self):
        with self.assertRaises(KeyError):
            networkArcFromEdge(self.G, (3,6))

    def testArcFromEdge(self):
        self.assertEqual(networkArcFromEdge(self.G, (2,3)), (2,3))
        self.assertEqual(networkArcFromEdge(self.G, (5,3)), (3,5))

    def testDelta(self):
        self.assertEqual(delta(self.G, [2,3]), [(3,4), (3,5), (1,2), (1,3)])
        

class allNeighboursTests(unittest.TestCase):
    def setUp(self):     
        self.G = nx.DiGraph()        
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-3)
        self.G.add_node(4, supply=-2)        
        self.G.add_edge(1, 2)
        self.G.add_edge(3, 1)
        self.G.add_edge(1, 4)
        self.G.add_edge(4, 1)
        
    def testNeighbourhood(self):
        self.assertEqual(len(allNeighbours(self.G, 1)), 3)

    def testListOneNodes(self):
        self.assertEqual(len(listDegreeOneNodes(self.G)), 2)

class feasibilityTests(unittest.TestCase):
    def setUp(self): 
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-3)
        self.G.add_node(4, supply=-2)
        self.G.add_edge(1, 2, lower=1, upper=2)
        self.G.add_edge(2, 3, lower=2, upper=2)
        self.G.add_edge(1, 4, lower=1, upper=4)

    def testInfeasibilityError(self):
        self.G.add_node("s")        
        with self.assertRaises(DataError):
            feasibilityCheck(self.G)

    def testInfeasibilityFalse(self):
        self.assertFalse(feasibilityCheck(self.G))

    def testInfeasibilityTrue(self):
        self.G.edge[2][3]["upper"] = 3
        self.assertTrue(feasibilityCheck(self.G))

    def testFeasibilityInfinityArcs(self):
        self.G.edge[2][3]["upper"] = float('inf')
        self.G.edge[1][2]["upper"] = float('inf')
        self.assertTrue(feasibilityCheck(self.G))
        
class convertNetworkTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=2)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-3)
        self.G.add_node(4, supply=-1)
        self.G.add_edge(1, 2, lower=1, upper=2)
        self.G.add_edge(2, 3, lower=2, upper=2)
        self.G.add_edge(1, 4, lower=1, upper=4)

    def testRemovalLowerBoundOnArc(self):
        removeLowerBoundOnArc(self.G, (1,2, self.G.edge[1][2]))
        self.assertFalse('lower' in self.G.edge[1][2])
        self.assertEqual(self.G.node[1]["supply"], 1)
        self.assertEqual(self.G.node[2]["supply"], 3)

    def testRemovalLowerBoundOnArcInf(self):
        self.G.edge[1][2]["lower"] = float('-inf')
        self.G.edge[1][2]["upper"] = float('inf')
        removeLowerBoundOnArc(self.G, (1,2, self.G.edge[1][2]))
        self.assertEqual(self.G.edge[1][2]['upper'], float('inf'))
        self.assertEqual(self.G.node[1]['supply'], sys.maxint + 2)
        self.assertEqual(self.G.node[2]['bigM'], sys.maxint)

    def testRemovalLowerBoundOnArcReverseArc(self):
        self.G.edge[1][2]["lower"] = float('-inf')
        removeLowerBoundOnArc(self.G, (1,2, self.G.edge[1][2]))
        self.assertFalse(self.G.has_edge(1,2))
        self.assertEqual(self.G.edge[2][1]['upper'], float('inf'))
        self.assertEqual(self.G.node[1]['supply'], 0)

    def testComputeBigM(self):
        self.assertEqual(computeBigM(self.G), 15)

    def testRemovalLowerBounds(self):    
        removeLowerBounds(self.G)
        self.assertTrue(isNetworkBalanced(self.G))
        sumBounds = sum([a[2]["upper"] for a in self.G.edges(data=True)])
        self.assertEqual(sumBounds, 4)

    def testRemovalLowerBoundsInf(self):
        self.G.edge[2][3]["lower"] = float('-inf')
        removeLowerBounds(self.G)
        self.assertEqual(self.G.node[2]['supply'], 1)
        self.assertFalse('bigM' in self.G.node[3])
        
    def testSingleSourceSinkConversion(self):
        newG = setUpSingleSourceSinkNetwork(self.G)
        self.assertEqual(len(newG.edges(["s"])), 1)
        self.assertEqual(newG.in_edges(["t"], data=True)[0][2]["upper"], 1)

    def testSingleSourceSinkWithoutLower(self):
        newG = setUpSingleSourceSinkNetwork(self.G, False)
        self.assertEqual(len(newG.edges(["s"])), 2)
        self.assertEqual(newG.edge[1][4]['lower'], 1)

    def testSingleSourceSinkInfinite(self):
        newG = setUpInfiniteBoundSingleSourceSinkNetwork(self.G)
        self.assertEqual(len(newG.edges(["s"])), 2)
        self.assertEqual(newG.edge['s'][1]['capacity'], 2)
        self.assertEqual(newG.edge[4]['t']['capacity'], 1)

    def testDeltaPlus(self):
        self.assertEqual(len(deltaPlus(self.G, [1,4])), 1)

    def testDeltaPlusIfLower(self):
        self.assertEqual(len(deltaPlusIfLowerBound(self.G, [1])), 2)
        del self.G.edge[1][2]['lower']
        self.G.edge[1][4]['lower'] = 0
        self.assertEqual(deltaPlusIfLowerBound(self.G, [1]), [])

    def testDeltaMinus(self):
        self.assertEqual(len(deltaMinus(self.G, [2,3,4])), 2)

    def testDeltaMinusIfLower(self):
        self.assertEqual(len(deltaMinusIfLowerBound(self.G, [2,3,4])), 2)
        del self.G.edge[1][2]['lower']
        self.G.edge[1][4]['lower'] = 0
        self.assertEqual(deltaMinusIfLowerBound(self.G, [2,3,4]), [])

    def testSumU(self):
        self.assertEqual(sumUInA(self.G, [(1,4), (2,3)]), 6)

    def testSumL(self):
        self.assertEqual(sumLInA(self.G, self.G.edges()), 4)

    def testSumLEmpty(self):
        del self.G.edge[1][2]['lower']
        del self.G.edge[2][3]['lower']
        del self.G.edge[1][4]['lower']
        self.assertEqual(sumLInA(self.G, self.G.edges()), 0)

    def testSumBPlus(self):
        self.assertEqual(sumBPlus(self.G), 4)
        
class ListingNodeTests(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(1, supply=1)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=3)
        self.G.add_node(4, supply=-4)
        self.G.add_node(5, supply=-6)
        self.G.add_node(6, supply=-0)

    def testListOfSupply(self):
        self.assertEqual(listOfSupplyNodes(self.G), [1,2,3])
    def testGeneratorOfSupplyNodes(self):
        self.assertEqual(list(listOfSupplyNodes_iter(self.G)), [1,2,3])

    def testListOfDemand(self):
        self.assertEqual(listOfDemandNodes(self.G), [4,5])
    def testGeneratorOfDemandNodes(self):
        self.assertEqual(list(listOfDemandNodes_iter(self.G)), [4,5])

    def testListOfInnodes(self):
        self.assertEqual(listOfInNodes(self.G), [6])
        
class ConnectedAndBalancedFailTests(unittest.TestCase):
    def setUp(self):
        auxG=nx.DiGraph()
        auxG.add_node("n1", supply=2)
        auxG.add_node("n2", supply=2)
        auxG.add_node("n3", supply=0)
        auxG.add_edge("n1", "n3", lower=3, upper=2 )           
        self.G = auxG

    def testBalancedFail(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.assertFalse(isNetworkBalanced(self.G))
    def testArcBoundFeasibleFail(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.assertFalse(areArcBoundsFeasible(self.G))

        
class ConnectedAndBalancedPassTests(unittest.TestCase):
    def setUp(self):
        auxG=nx.DiGraph()
        auxG.add_node(1, supply=2)
        auxG.add_node(2, supply=-1)
        auxG.add_node(3, supply=-1)
        auxG.add_edge(1, 2, lower=-1, upper=1 )
        auxG.add_edge(1, 3, lower=0, upper=2 )        
        self.G = auxG

    def testBalancedPass(self):
        self.assertTrue(isNetworkBalanced(self.G))
    def testSumB(self):
        self.assertEqual(sumBInS(self.G, [1,2]), 1)
    def testArcBoundFeasiblePass(self):
        self.assertTrue(areArcBoundsFeasible(self.G))

class pathTest(unittest.TestCase):
    def testArcsOnPath(self):
        self.assertEqual(arcsOnPath([1,3,6,5]), [(1,3), (3,6), (6,5)])

    def testArcsOnPathIter(self):
        gen = arcsOnPath_iter([1,3,6,5])
        for a in gen:
            self.assertTrue(a in [(1,3), (3,6), (6,5)])

class reversedEdgesTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()        
        self.G.add_node(1, supply=3)
        self.G.add_node(2, supply=2)
        self.G.add_node(3, supply=-3)
        self.G.add_node(4, supply=-2)        
        self.G.add_edge(1, 2, lower=0, upper=3)
        self.G.add_edge(3, 1, lower=0, upper=3)
        self.G.add_edge(1, 4, lower=1, upper=3)
        self.G.add_edge(4, 1, lower=2, upper=3)        

    def testAntiParallelEdges(self):
        removeAntiParallelArcs(self.G)
        self.assertFalse(self.G.has_edge(4,1))
        self.assertEqual(self.G.edge[1][4]["lower"], -2)
        self.assertEqual(self.G.edge[1][4]["upper"], 1)

    def testAntiParallelEdgesWithoutLower(self):
        del self.G.edge[1][4]['lower']
        removeAntiParallelArcs(self.G)
        self.assertFalse(self.G.has_edge(4,1))
        self.assertEqual(self.G.edge[1][4]["lower"], -3)
        self.assertEqual(self.G.edge[1][4]["upper"], 1)

    def testReverseArc(self):
        reverseArcWithBounds(self.G, (3,1, self.G.edge[3][1]))
        self.assertFalse(self.G.has_edge(3,1))
        self.assertEqual(self.G.edge[1][3]['lower'], -3)

    def testEnsureArcDirectionToLargerBound(self):
        removeAntiParallelArcs(self.G)
        self.G.add_edge(3,2, upper=1, lower=float('-inf'))
        ensureArcDirectionToLargerBound(self.G)
        self.assertEqual(self.G.edge[4][1]["upper"], 2)
        self.assertFalse(self.G.has_edge(1,4))
        self.assertEqual(self.G.edge[2][3]["upper"], float('inf'))
        self.assertFalse(self.G.has_edge(3,2))

    def testEnsureNetworkForm(self):
        auxG = ensureNetworkForm(self.G)
        self.assertEqual(auxG.edge[4][1]["upper"], 3)
        self.assertEqual(auxG.node[1]['supply'], 2)

    # def testDelNodeDicts(self):
    #     for v in self.G:
    #         self.G.node[v]['random'] = 0
    #     delNodeDicts(self.G, 'supply', 'random', 'aux')
    #     self.assertEqual(sum(len(v[1].keys()) for v in self.G.nodes(data=True)),0)
    #     with self.assertRaises(KeyError):
    #         self.G.node[2]['supply']

    # def testDelArcDicts(self):
    #     delArcDicts(self.G, 'upper')
    #     self.assertEqual(sum(d['lower'] for v,w,d in self.G.edges(data=True)), 3)
    #     with self.assertRaises(KeyError):
    #         self.G[4][1]['upper']
        
class infiniteMaxFlowTest(unittest.TestCase):
    def setUp(self):
        self.G = nx.DiGraph()
        self.G.add_node(0, supply=2)
        self.G.add_node(1, supply=0)
        self.G.add_node(2, supply=-2)        
        self.G.add_edge(1, 0, lower=float("-inf"), upper=1)
        self.G.add_edge(1, 2, upper=float("inf"))

    def testRemoveInfiniteUpperBounds(self):
        removeInfiniteUpperBoundsForMaxFlow(self.G)
        self.assertFalse('upper' in self.G[1][2])
        self.assertEqual(self.G[1][0]['upper'], 1)
        
    def testFeasibility(self):
        self.assertTrue(feasibilityCheck(self.G))

    def testFeasibilityInfinity(self):
        self.G[1][2]['upper'] = 1
        self.assertFalse(feasibilityCheck(self.G))

class simpleFunctionTests(unittest.TestCase):
    def setUp(self):
        self.L = [float('-inf'), 5, 7, 8.3, -10, float('inf')]

    def testMaxAbsWithoutInf(self):
        self.assertEqual(maxAbsWithoutInf(self.L), 10)
        with self.assertRaises(ValueError):
            maxAbsWithoutInf([])

    def testSumAbsWithoutInf(self):
        self.assertEqual(sumAbsWithoutInf(self.L), 30.3)
        
        
if __name__ == '__main__':
    unittest.main()
